(* ::Package:: *)

(* ::Input::Initialization:: *)
(*Coordinates and metric for the relevant 3dim.subspace in the Poincare patch*)
xmu={z,t,y};
d=Length[xmu];
(*smooth step function determining the location of the shell*)
M[t_]:=m (1.+Tanh[a t])/2;
(*conformal factor*)
omega=z^-(dim-2);
(*effective metric for the extremal surfaces*)
metric=omega^2{{0,-1/z^2,0},{-1/z^2,-(1-M[t] z^dim)/z^2,0},{0,0,1/z^2}};
metricinv=Inverse[metric];
g=Det[metric];
(*Christoffel symbol and contraction*)
Gudd=Table[Sum[1/2metricinv[[i,l]](D[metric[[k,l]],xmu[[j]]]+D[metric[[j,l]],xmu[[k]]]-D[metric[[j,k]],xmu[[l]]]),{l,1,d}],{i,1,d},{j,1,d},{k,1,d}];
(*building the Christoffels used in the relaxation code*)
gamma[z_,t_,y_]=Table[Gudd[[i,j,k]],{i,d},{j,d},{k,d}];
Dgamma[z_,t_,y_]=Table[Table[D[Gudd[[i,j,k]],xmu[[l]]],{i,d},{j,d},{k,d}],{l,d}]//Simplify;


(* ::Input::Initialization:: *)
(*initial guess for the geodesic*)makeGeoAnsatz[gridsize_,zCUT_,l0_,t0_]:=Module[{tt,tav,f,tau,Jf,tte,tts,tCUTOFF,lambda,dlambda,h,Jav,Y,Z,T,dY,dZ,dT,x,Zmax,lambdap,dlambdap,Yp,Ym,dYp,dYm},(*turning point in the bulk*)
Zmax=l0/2Gamma[1/(2(dim-1))]/(Sqrt[Pi]*Gamma[dim/(2(dim-1))]);
(*parameter value at the z-cutoff*)
tCUTOFF=1-Sqrt[1-zCUT/Zmax];
tts=-1+tCUTOFF;
tte=1-tCUTOFF;
h=(tte-tts)/(gridsize-1);
tt=Table[i,{i,tts,tte,h}];
tav=Table[(i+i+h)/2,{i,tts,tte-h,h}];
f[x_]=x;
tau[x_]=ArcTanh[f[x]*Sqrt[2-f[x]^2]];
Jf[x_]=Which[dim==2,-Derivative[2][tau][x]/Derivative[1][tau][x],dim==3,-1*(Derivative[2][f][x]/Derivative[1][f][x])+(Derivative[1][f][x]*f[x]*(-22+38*f[x]^2-27*f[x]^4+7*f[x]^6))/(4-10f[x]^2+10*f[x]^4-5*f[x]^6+f[x]^8),dim==4,-1*(Derivative[2][f][x]/Derivative[1][f][x])+(Derivative[1][f][x]*f[x]*(-51+145*f[x]^2-205*f[x]^4+159*f[x]^6-65*f[x]^8+11*f[x]^10))/((-2+f[x]^2)*(-1+f[x]^2)*(3-3*f[x]^2+f[x]^4)*(1-f[x]^2+f[x]^4))];
Jav=Jf[tav];
lambda=f[tt];
dlambda=Derivative[1][f][tt];
lambdap=lambda[[1;;Length[tt]/2]];
dlambdap=dlambda;
Yp=-l0/2+(Zmax*(1-lambdap^2))^dim/(dim*Zmax^(dim-1))*Hypergeometric2F1[1/2,dim/(2(dim-1)),(3dim-2)/(2dim-2),(-1+lambdap^2)^(2(dim-1))];
Ym=Reverse[-Yp];
Y=Join[Yp,Ym];
dYp=dlambdap l0 lambdap (1-lambdap^2)^dim Gamma[1/(2(dim-1))]/((lambdap^2-1)Sqrt[Pi]Sqrt[1-(lambdap^2-1)^(2(dim-1))]*Gamma[dim/(2(dim-1))]);
dYm=Reverse[dYp];
dY=Join[dYp,dYm];
Z=Zmax*(1-lambda^2);
dZ=-2*Zmax*dlambda*lambda;
T=t0-Z;
dT=-dZ;
{{Z,T,Y},{dZ,dT,dY},Jav,lambda,dlambda,h}];


(* ::Input::Initialization:: *)
(*evaluating the numerical values of the Christoffel symbols*)evaluateChristoffels[gridsize_,X_,dX_]:=Module[{Xav,dXav,gammaALL,DgammaALL},Xav=Table[1/2(X[[i,j+1]]+X[[i,j]]),{i,Length[X]},{j,1,gridsize-1}];
dXav=Table[1/2(dX[[i,j+1]]+dX[[i,j]]),{i,Length[X]},{j,1,gridsize-1}];
gammaALL=gamma[Xav[[1]],Xav[[2]],Xav[[3]]];
DgammaALL=Table[Dgamma[Xav[[1]],Xav[[2]],Xav[[3]]][[i]],{i,Length[X]}];
For[i=1,i<4,i++,For[j=1,j<4,j++,For[k=1,k<4,k++,If[Length[gammaALL[[i,j,k]]]==0,gammaALL[[i,j,k]]=Table[0,{gridsize-1}]];
Do[If[Length[DgammaALL[[l,i,j,k]]]==0,DgammaALL[[l,i,j,k]]=Table[0,{gridsize-1}]],{l,Length[X]}];]]];
{Xav,dXav,gammaALL,DgammaALL}];


(* ::Input::Initialization:: *)
(*building the finite differences equations*)writeFDEs[gridsize_,X_,dX_,dXav_,Jav_,h_,gammaALL_]:=Module[{i,j,FDEs},FDEs=Table[0,{6*gridsize}];
j=1;
For[i=1,i<6*(gridsize-1),i+=6,Do[FDEs[[i+2m+1]]=X[[m,j+1]]-X[[m,j]]-h*dXav[[m,j]];
FDEs[[i+2m+2]]=dX[[m,j+1]]-dX[[m,j]]+h*Jav[[j]]*dXav[[m,j]]+h*Sum[gammaALL[[m,k,l,j]]*dXav[[k,j]]*dXav[[l,j]],{k,Length[X]},{l,Length[X]}];,{m,1,3}];
j+=1;];
FDEs];


(* ::Input::Initialization:: *)
(*computing the numerical entries of the S-matrix*)findSElements[gridsize_,dXav_,Jav_,h_,gammaALL_,DgammaALL_]:=Module[{j,Sp,Sm},Sp=Sm=Table[Table[Table[0,{gridsize-1}],{6}],{6}];
For[j=1,j<(gridsize),j+=1,Do[If[m==n,Sp[[2m-1,2n-1,j]]=1;
Sm[[2m-1,2n-1,j]]=-1;
Sp[[2m-1,2n,j]]=-h/2;
Sm[[2m-1,2n,j]]=-h/2;
Sp[[2m,2n,j]]=1+h/2Jav[[j]]+h Sum[dXav[[k,j]]*gammaALL[[m,n,k,j]],{k,3}];
Sm[[2m,2n,j]]=-1+h/2Jav[[j]]+h Sum[dXav[[k,j]]*gammaALL[[m,n,k,j]],{k,3}];,Sp[[2m,2n,j]]=h Sum[dXav[[k,j]]*gammaALL[[m,n,k,j]],{k,3}];
Sm[[2m,2n,j]]=Sp[[2m,2n,j]];];
Sp[[2m,2n-1,j]]=h/2Sum[dXav[[k,j]]*dXav[[l,j]]*DgammaALL[[n,m,k,l,j]],{k,3},{l,3}];
Sm[[2m,2n-1,j]]=Sp[[2m,2n-1,j]];,{m,3},{n,3}];];
{Sp,Sm}];


(* ::Input::Initialization:: *)
(*building the sparse matrix*)createSmatrix[gridsize_,Sp_,Sm_,h_]:=Module[{i,j,dvS,dvSh,dvM,rvS,rvM,cvS,cvSh,cvM,sMatrix,k,n,l},{rvS,cvSh,dvS,dvSh}=Table[Table[0,{24*(gridsize-1)}],{4}];
j=0;
For[i=1,i<24*(gridsize-1),i+=24,k=0;n=1;
While[k<23,Do[rvS[[i+k]]=If[l<3,2(n+1),2n+3]+6j;
cvSh[[i+k]]=If[l<3,2n-2+l,l-2]+6j;
dvS[[i+k]]=Sp[[If[l<3,2n-1,2n],If[l<3,2n-2+l,l-2],j+1]];
dvSh[[i+k]]=Sm[[If[l<3,2n-1,2n],If[l<3,2n-2+l,l-2],j+1]];
k++;,{l,1,8}];
n++;];
j+=1;];
cvS=cvSh+6;
rvM=Flatten[{1,2,3,rvS,rvS,(6*gridsize-2),(6*gridsize-1),(6*gridsize)}];
cvM=Flatten[{1,3,5,cvSh,cvS,(6*gridsize-5),(6*gridsize-3),(6*gridsize-1)}];
dvM=Flatten[{1,1,1,dvSh,dvS,1,1,1}];
sMatrix=SparseArray[Table[{rvM[[k]],cvM[[k]]}->dvM[[k]],{k,1,Length[cvM]}]];
sMatrix];


(* ::Input::Initialization:: *)
(*Newton iteration*)
NewtonStep[gridsize_,errorFDE_,deltaX_,X_,dX_]:=Module[{i,j,eps,Xnew,dXnew},Xnew=Table[0,{Length[X]},{gridsize}];
dXnew=Table[0,{Length[X]},{gridsize}];
If[errorFDE<1*10^-3,eps=1.,eps=5*10^-1];
j=1;
For[i=1,i<6*gridsize,i+=6,Do[Xnew[[k,j]]=X[[k,j]]+eps*deltaX[[i+2k-2]];
dXnew[[k,j]]=dX[[k,j]]+eps*deltaX[[i+2k-1]];,{k,1,3}];
j+=1;];
{Xnew,dXnew}];


(* ::Input::Initialization:: *)
(*relaxation routine*)relax[gridsize_,X_,dX_,h_,Jav_]:=Module[{Xav,dXav,gammaALL,DgammaALL,FDEs,errorFDE,Sp,Sm,sMatrix,deltaX,Xnew,dXnew},{Xav,dXav,gammaALL,DgammaALL}=evaluateChristoffels[gridsize,X,dX];
FDEs=writeFDEs[gridsize,X,dX,dXav,Jav,h,gammaALL];
errorFDE=Sum[Abs[FDEs[[i]]],{i,Length[FDEs]}]/(6*gridsize);
{Sp,Sm}=findSElements[gridsize,dXav,Jav,h,gammaALL,DgammaALL];
sMatrix=createSmatrix[gridsize,Sp,Sm,h];
deltaX=-LinearSolve[sMatrix,FDEs];
{Xnew,dXnew}=NewtonStep[gridsize,errorFDE,deltaX,X,dX];
{Xnew,dXnew,errorFDE}];


(* ::Input::Initialization:: *)
(*computing the geodesic length*)findGeoLength[X0_,dX0_,X_,dX_,lambda_,zCUT_]:=Module[{Xint,X0int,dsint,ds0int,L,L0},Xint=Table[Interpolation[Thread[{lambda,X[[i,;;]]}]],{i,Length[X]}];
X0int=Table[Interpolation[Thread[{lambda,X0[[i,;;]]}]],{i,Length[X0]}];
dsint=Xint[[1]][x]^-(dim-1)*Sqrt[-(1-M[Xint[[2]][x]]Xint[[1]][x]^dim)D[Xint[[2]][x],x]^2-2 D[Xint[[2]][x],x] D[Xint[[1]][x],x]+D[Xint[[3]][x],x]^2];
ds0int=X0int[[1]][x]^-(dim-1) Sqrt[-D[X0int[[2]][x],x]^2-2 D[X0int[[2]][x],x] D[X0int[[1]][x],x]+D[X0int[[3]][x],x]^2];
L=NIntegrate[dsint,{x,First[lambda],Last[lambda]}];
L0=NIntegrate[ds0int,{x,First[lambda],Last[lambda]}];
{L,L0}];


(* ::Input::Initialization:: *)
(*routine that relaxes a single geodesic*)relaxOneGeodesic[gridsize_,zCUT_,l0_,t0_,errMAX_,itMAX_]:=Module[{L,L0,ansatz,geodesic,i,X0,dX0,Jav,lambda,dlambda,h,X,dX,errorFDE},{X0,dX0,Jav,lambda,dlambda,h}=makeGeoAnsatz[gridsize,zCUT,l0,t0];
{X,dX}={X0,dX0};
{X[[2,1]],X[[2,-1]],X[[3,1]],X[[3,-1]]}={t0-zCUT,t0-zCUT,-l0/2,l0/2};
errorFDE=10.^5;
i=1;
While[errorFDE>errMAX,{X,dX,errorFDE}=relax[gridsize,X,dX,h,Jav];
If[i==itMAX,Break[]];
i+=1;];
Print["Iteration: ",i," with errorFDE: ",errorFDE];
{L,L0}=findGeoLength[X0,dX0,X,dX,lambda,zCUT];
ansatz={t0,l0,gridsize,X0,dX0,lambda,0,0,L0};
geodesic={t0,l0,gridsize,X,dX,lambda,errorFDE,i-1,L};
{ansatz,geodesic}];


(* ::Input::Initialization:: *)
(*Compute Entanglement entropy as function of time*)
relaxTevolution[gridsize_,zCUT_,l0_,t0_,tMax_,Nt_,errMAX_,itMAX_]:=Module[{i,j,t,deltat,X0,dX0,Jav,lambda,dlambda,h,X,dX,errorFDE,L,L0,ansatz,geodesic},deltat=(tMax-t0)/(Nt-1);
t=t0;ansatz={};geodesic={};
{X0,dX0,Jav,lambda,dlambda,h}=makeGeoAnsatz[gridsize,zCUT,l0,t0];
For[j=1,j<=Nt,j++,Print["step: ",j," @ t = ",t];
If[j==1,{X,dX}={X0,dX0}];
{X[[2,1]],X[[2,-1]],X[[3,1]],X[[3,-1]]}={t-zCUT,t-zCUT,-l0/2,l0/2};
errorFDE=10^5;i=1;
While[errorFDE>errMAX,{X,dX,errorFDE}=relax[gridsize,X,dX,h,Jav];
(*Print["Iteration: ",i," with errorFDE: ",errorFDE];*)If[i==itMAX,Break[]];
i+=1;];
Print[i-1," iterations with error= ",errorFDE];
{L,L0}=findGeoLength[X0,dX0,X,dX,lambda,zCUT];
AppendTo[ansatz,{t,l0,gridsize,X0,dX0,lambda,0,0,L0}];
AppendTo[geodesic,{t,l0,gridsize,X,dX,lambda,errorFDE,i-1,L}];
t=t+deltat;];
{ansatz,geodesic}];


(* ::Input::Initialization:: *)
(*compute entanglement entropy as function of the separation*)relaxLevolution[gridsize_,zCUT_,l0_,t0_,lMax_,Nl_,errMAX_,itMAX_]:=Module[{i,j,l,deltal,X0,dX0,Jav,lambda,dlambda,h,X,dX,errorFDE,L,L0,ansatz,geodesic},deltal=(lMax-l0)/(Nl-1);
l=l0;ansatz={};geodesic={};
For[j=1,j<=Nl,j++,Print["step: ",j," @ L = ",l];
{X0,dX0,Jav,lambda,dlambda,h}=makeGeoAnsatz[gridsize,zCUT,l,t0];
If[j==1,{X,dX}={X0,dX0}];
{X[[2,1]],X[[2,-1]],X[[3,1]],X[[3,-1]]}={t0-zCUT,t0-zCUT,-l/2,l/2};
errorFDE=10^5;i=1;
While[errorFDE>errMAX,{X,dX,errorFDE}=relax[gridsize,X,dX,h,Jav];
(*Print["Iteration: ",i," with errorFDE: ",errorFDE];*)If[i==itMAX,Break[]];
i+=1;];
Print[i-1," iterations with error= ",errorFDE];
{L,L0}=findGeoLength[X0,dX0,X,dX,lambda,zCUT];
AppendTo[ansatz,{t0,l,gridsize,X0,dX0,lambda,0,0,L0}];
AppendTo[geodesic,{t0,l,gridsize,X,dX,lambda,errorFDE,i-1,L}];
l=l+deltal;];
{ansatz,geodesic}];
