(* ::Package:: *)

(* ::Section::Initialization:: *)
(*Einstein equations in Eddington-Finkelstein gauge*)


(* ::Input::Initialization:: *)
(* Coordinates and metric *)
xmu={r,t,y,x1,x2};
d=Length[xmu];
metric={{0,1,0,0,0},{1,-A[r,t],0,0,0},{0,0,E^(-2 B[r,t]) S[r,t]^2,0,0},{0,0,0,E^B[r,t] S[r,t]^2,0},{0,0,0,0,E^B[r,t] S[r,t]^2}};
metricinv=Inverse[metric];
g=Det[metric];
Print["Bulk metric:"]
Print["\!\(\*SubscriptBox[\(g\), \(\[Mu]\[Nu]\)]\)[r,t]\[Equal]",MatrixForm[metric]]


(* ::Input::Initialization:: *)
(* Christoffel symbol and contraction *)
Gudd=Table[Sum[1/2metricinv[[i,l]](D[metric[[k,l]],xmu[[j]]]+D[metric[[j,l]],xmu[[k]]]-D[metric[[j,k]],xmu[[l]]]),{l,1,d}],{i,1,d},{j,1,d},{k,1,d}];
GGuddd=Table[Sum[Gudd[[i,j,m]]*Gudd[[m,k,l]],{m,1,d}],{i,1,d},{j,1,d},{k,1,d},{l,1,d}];


(* ::Input::Initialization:: *)
(* Riemann tensor, Ricci tensor and Ricci scalar *)
Ruddd=Table[D[Gudd[[i,l,j]],xmu[[k]]]-D[Gudd[[i,k,j]],xmu[[l]]]+GGuddd[[i,k,l,j]]-GGuddd[[i,l,k,j]],{i,1,d},{j,1,d},{k,1,d},{l,1,d}];
Rdd=Table[Sum[Ruddd[[m,i,m,j]],{m,1,d}],{i,1,d},{j,1,d}]//Simplify;
Rud=Table[Sum[metricinv[[i,m]]*Rdd[[m,j]],{m,1,d}],{i,1,d},{j,1,d}]//Simplify;
Rscal=Sum[Rud[[m,m]],{m,1,d}]//Simplify;


(* ::Input::Initialization:: *)
(* Vacuum Einstein equations + cosmological constant *)
replL={L->1}; (* we set the AdS radius to 1 *)
Lambda=-(d-1)(d-2)/(2L^2)/.replL;
EinsteinEqns=Rdd-1/2metric*Rscal+Lambda*metric//Simplify;
(* note that the order in which the equations appear in the list can depend on the Mathematica version used, this notebook assumes Mathematica11 *)
EE=Drop[Union[Flatten[EinsteinEqns]],1];


(* ::Input:: *)
(*TableForm[EEeq=#==0&/@EE]*)


(* ::Section::Initialization:: *)
(*Coordinate transformation r-> z=1/r *)


(* ::Input::Initialization:: *)
Dr[func_, times_Integer] := D[func[1/r, t], {r, times}] /. r -> 1/z;
replrtoz = {\!\(\*
TagBox[
StyleBox[
RowBox[{
RowBox[{
RowBox[{"Derivative", "[", 
RowBox[{"ddr_", ",", "ddt_"}], "]"}], "[", "f_", "]"}], "[", 
RowBox[{"r", ",", "t"}], "]"}],
ShowSpecialCharacters->False,
ShowStringCharacters->True,
NumberMarks->True],
FullForm]\) -> Dr[\!\(\*
TagBox[
StyleBox[
RowBox[{
RowBox[{"Derivative", "[", 
RowBox[{"0", ",", "ddt"}], "]"}], "[", "f", "]"}],
ShowSpecialCharacters->False,
ShowStringCharacters->True,
NumberMarks->True],
FullForm]\), ddr], r -> z};
EOMz=EE/.replrtoz ;


(* ::Section::Initialization:: *)
(*Near boundary analysis*)


(* ::Input::Initialization:: *)
(* ansatz for the near boundary expansion, for order>7 the series ansatz must be extended with higher power log terms. *)
Asb[z_,t_,ord_]:=1/(z \[Epsilon])^2Sum[(Subscript[a, n][t]+Subscript[\[Alpha]1, n][t]Log[z]+Subscript[\[Alpha]2, n][t]Log[z]^2)(z \[Epsilon])^n,{n,0,ord}];
Ssb[z_,t_,ord_]:=1/(z \[Epsilon]) Sum[(Subscript[s, n][t]+Subscript[\[Sigma]1, n][t]Log[z]+Subscript[\[Sigma]2, n][t]Log[z]^2)(z \[Epsilon])^n,{n,0,ord}];
Bsb[z_,t_,ord_]:=Sum[(Subscript[b, n][t]+Subscript[\[Beta]1, n][t]Log[z]+Subscript[\[Beta]2, n][t]Log[z]^2)(z \[Epsilon])^n,{n,0,ord}];


(* ::Input::Initialization:: *)
(* boundary conditions *)
BCs={Subscript[a, 0][t]->1,Subscript[a, 1]->(0&),Subscript[s, 0][t]->1,(*Subscript[b, 0][t]\[Rule]b0[t],*)Subscript[b, 0]->(0&)};


(* ::Input::Initialization:: *)
(* check if BCs give the correct boundary metric *)
Print["Boundary metric:"]
Print["\!\(\*SubsuperscriptBox[\(ds\), \(bdry\), \(2\)]\)=",z^2(-Asb[z,t,0]dt^2+Ssb[z,t,0]^2(E^(-2 Bsb[z,t,0]) dy^2+E^Bsb[z,t,0] dx1^2+E^Bsb[z,t,0] dx2^2))/.{Log[z]->0,\[Epsilon]->1}/.BCs//Simplify]


(* ::Input::Initialization:: *)
(* replacements for putting the expansion parameter into the EOMs *)
ztoZ={f_[z,t]->f[Z,t]};
putEps={\!\(\*SuperscriptBox[\(f_\), 
TagBox[
RowBox[{"(", 
RowBox[{"n_", ",", "m_"}], ")"}],
Derivative],
MultilineFunction->None]\)[Z,t]->\[Epsilon]^-n \!\(\*SuperscriptBox[\(f\), 
TagBox[
RowBox[{"(", 
RowBox[{"n", ",", "m"}], ")"}],
Derivative],
MultilineFunction->None]\)[Z,t],z->\[Epsilon] z};
Ztoz={f_[Z,t]->f[z,t]};


(* ::Input::Initialization:: *)
(* overall factors make the EOMs have the same leading order in z *)
EOMzEps={1/z^2 EOMz[[1]],z^2EOMz[[2]],z^2EOMz[[3]],z^2EOMz[[4]],z^2EOMz[[5]]}/.ztoZ/.putEps/.Ztoz;


(* ::Input::Initialization:: *)
(* solving the Einstein equations order by order in z and save the result to a file *)
order=5;resFile="NBsolAdS5Ord5.res";
sol=BCs;putDel={Log[z]->\[Delta]};
(*Print["BCs: ",Simplify[sol]/.Rule\[Rule]Equal]//TableForm;
Print["Total time: ",AbsoluteTiming[*)
Do[
(*Print["Time: ",AbsoluteTiming[*)
Ab[z_,t_]=Asb[z,t,i]/.sol;
Sb[z_,t_]=Ssb[z,t,i]/.sol;
Bb[z_,t_]=Bsb[z,t,i]/.sol;
replBdry={A->Ab,S->Sb,B->Bb};
EOMnow=SeriesCoefficient[EOMzEps[[;;]]/.replBdry/.putDel,{\[Epsilon],0,i},{\[Delta],0,0}];
EOMnowLog=SeriesCoefficient[EOMzEps[[;;]]/.replBdry/.putDel,{\[Epsilon],0,i},{\[Delta],0,1}];
EOMnowLog2=SeriesCoefficient[EOMzEps[[;;]]/.replBdry/.putDel,{\[Epsilon],0,i},{\[Delta],0,2}];
solnow=Solve[Thread[{EOMnow[[;;]]==0,EOMnowLog[[;;]]==0,EOMnowLog2[[;;]]==0}],{Subscript[a, 4]'[t],Subscript[a, i][t],Subscript[b, i][t],Subscript[s, i][t],Subscript[\[Alpha]1, i][t],Subscript[\[Sigma]1, i][t],Subscript[\[Beta]1, i][t],Subscript[\[Alpha]2, i][t],Subscript[\[Sigma]2, i][t],Subscript[\[Beta]2, i][t]}]//Flatten//Quiet;
sol=Flatten[{sol,solnow}];
checknow=Thread[SeriesCoefficient[EOMzEps[[;;]]/.replBdry//.sol,{\[Epsilon],0,i}]==0]//Simplify;
(*Print["Order=",i,"; Check equations: ",checknow];
Print[solnow];
][[1]]];*)
,{i,0,order}];
(*][[1]]];*)
nbSol=sol;
Clear[order];
SetDirectory[NotebookDirectory[]];
Put[nbSol//Simplify,resFile];


(* ::Input:: *)
(*(* conservation of the stress tensor *)*)
(*Derivative[1][Subscript[a, 4]][t]==(Derivative[1][Subscript[a, 4]][t]/.nbSol)//Expand*)


(* ::Input:: *)
(*(* conformal Ward identity *)*)
(*Subscript[s, 4][t]==(Subscript[s, 4][t]/.nbSol)*)


(* ::Input::Initialization:: *)
(* solution of the near boundary expansion *)
Print["Near boundary solution:"]
Print["A[z,t]=",Collect[Asb[z,t,5]/.nbSol/.\[Epsilon]->1,z,Expand]]
Print["S[z,t]=",Collect[Ssb[z,t,5]/.nbSol/.\[Epsilon]->1,z,Simplify]]
Print["B[z,t]=",Collect[Bsb[z,t,5]/.nbSol/.\[Epsilon]->1,z,Simplify]]


(* ::Input::Initialization:: *)
(* residual gauge freedom of the line element z\[Rule]zp=z/(1+\[Xi][t]z),A[z,t]\[Rule]A[zp,t]-2\[Xi]'[t],S[z]->S[zp],B[z]\[Rule]B[zp] *)
ds2old=-A[z,t]dt^2-2dz dt/z^2+S[z,t]^2(Exp[-2B[z,t]]dy^2+Exp[B[z,t]](dx1^2+dx2^2));
zp[z_,t_]:=z/(1+\[Xi][t]z);
repldzp={dz->D[zp[z,t],z]dz+D[zp[z,t],t]dt,z->zp[z,t]};
Collect[ds2old/.repldzp,{dz,dt,dx1,dx2,dy},Simplify];


(* ::Input:: *)
(*(* solution of the near boundary expansion including the gauge function \[Xi][t]; note that the +2\[Xi]'[t] term comes from transforming the drdt-term in the line element *)*)
(*A[z,t]==Collect[Series[Asb[z/(1+z \[Xi][t]),t,5]-2\[Xi]'[t]/.nbSol/.\[Epsilon]->1,{z,0,5}],{z,Log[z]}]*)
(*S[z,t]==Collect[Series[Ssb[z/(1+z \[Xi][t]),t,5]/.nbSol/.\[Epsilon]->1,{z,0,5}],z,Simplify]*)
(*B[z,t]==Collect[Series[Bsb[z/(1+z \[Xi][t]),t,5]/.nbSol/.\[Epsilon]->1,{z,0,5}],z,Simplify]*)


(* ::Section::Initialization:: *)
(*Transformation from EF  to FG coordinates*)


(* ::Input::Initialization:: *)
(* metric in EF coordinates *)
metricEFz=({
 {0, -1/z^2, 0, 0, 0},
 {-1/z^2, -A[z,t], 0, 0, 0},
 {0, 0, S[z,t]^2 E^(-2 B[z,t]), 0, 0},
 {0, 0, 0, S[z,t]^2 E^B[z,t], 0},
 {0, 0, 0, 0, S[z,t]^2 E^B[z,t]}
});


(* ::Input::Initialization:: *)
(* loading the near-boundary solution *)
SetDirectory[NotebookDirectory[]];
nbSol=Get["NBsolAdS5Ord5.res"];order=5;
Clear[Ab,Sb,Bb];


(* ::Input::Initialization:: *)
(* near-boundary series in EF coordinates *)
Ab[z_,t_]=Asb[z,t,order]/.nbSol/.\[Epsilon]->1;
Sb[z_,t_]=Ssb[z,t,order]/.nbSol/.\[Epsilon]->1;
Bb[z_,t_]=Bsb[z,t,order]/.nbSol/.\[Epsilon]->1;


(* ::Input::Initialization:: *)
(* line element in EF-coordinates; we need to transform only the dudv and dv^2 parts, the rest is already diagonal *)
ds2EF=-(2/z^2)dz dt-Ab[z,t]dt^2;


(* ::Input::Initialization:: *)
(* series ansatz for the EF-coords in terms of FG-coords; the log-terms are necessary only from 5th order on *)
ZEFs[v_,u_,ord_]:=Sum[Subscript[zEF, n][v]u^n,{n,1,ord}]+Subscript[\[Zeta]EF, 5][v]Log[u]u^5+Subscript[\[Zeta]EF, 6][v]Log[u]u^6;
TEFs[v_,u_,ord_]:=v+Sum[Subscript[tEF, n][v]u^n,{n,1,ord}]+Subscript[\[Tau]EF, 5][v]Log[u]u^5+Subscript[\[Tau]EF, 6][v]Log[u]u^6;


(* ::Input::Initialization:: *)
(* solving the near boundary expansion of the coordinate transformation order by order *)
BCs={Subscript[tEF, 1][v]->-1};(* this follows from Subscript[g, tt]=-1 *)
sol=BCs;
Do[
ZEF=ZEFs[v,u,i]/.sol;
TEF=TEFs[v,u,i]/.sol;
replFG={z->ZEF,t->TEF,dz->D[ZEF,u] du+D[ZEF,v] dv,dt->D[TEF,u] du+D[TEF,v] dv};
(* derivative expansion of the line element *)
ds2FG=ds2EF//.replFG;
ds2FGnow=Coefficient[Series[ds2FG,{u,0,i-3}]/.Log[u]->0,{du^2,du dv}];
ds2FGnowLog=Coefficient[D[Series[ds2FG,{u,0,i-3}],Log[u]],{du^2,du dv}];
solNow=Solve[{ds2FGnow[[1]]==1/u^2,ds2FGnow[[2]]==0,
ds2FGnowLog[[1]]==0,ds2FGnowLog[[2]]==0},{Subscript[tEF, i][v],Subscript[zEF, i][v],Subscript[\[Zeta]EF, i][v],Subscript[\[Tau]EF, i][v]}]//Quiet;
(*Print[solNow];*)
sol=Flatten[{sol,solNow}];
,{i,1,6}];


(* ::Input::Initialization:: *)
(* FG coordinates in terms of EF series coefficients *)
ZEF=ZEFs[v,u,6]/.sol;
TEF=TEFs[v,u,6]/.sol;
(* expansion of the v-coordinate *)
Series[TEF,{u,0,5}]
(* expansion of the u-coordinate *)
Series[ZEF,{u,0,5}]


(* ::Input::Initialization:: *)
Print["Solution of the EF to FG transformation:"]
(* expansion of the v-coordinate *)
Print["t[u,v]=",Series[TEF,{u,0,5}]]
(* expansion of the u-coordinate *)
Print["z[u,v]=",Series[ZEF,{u,0,5}]]


(* ::Input::Initialization:: *)
(* metric in FG coordinates *)
replFG={z->ZEF,t->TEF,dz->D[ZEF,u] du+D[ZEF,v] dv,dt->D[TEF,u] du+D[TEF,v] dv};
ds2FG=ds2EF+Sb[z,t]^2(E^(-2 Bb[z,t]) dy^2+E^ Bb[z,t] dx1^2+E^ Bb[z,t] dx2^2)//.replFG;
(* leading order expansion of the metric in FG coordinates *)
Collect[Series[ds2FG,{u,0,-2}],{du,dv,dy,dx1,dx2}];


(* ::Input::Initialization:: *)
Print["Line element in FG coordinates:"]
(* line element in u-coordinate *)
ds2u=Collect[Series[ds2FG,{u,0,2}],{du,dv,dx,u,Log[u]}];
Print["\!\(\*SubsuperscriptBox[\(ds\), \(FG\), \(2\)]\)=",ds2u]


(* ::Input::Initialization:: *)
(* metric in \[Rho]-coordinate (ds^2=d\[Rho]^2/(4\[Rho])+1/\[Rho]*Subscript[g, ij](x)dx^idx^j) such as used in (arXiv:hep-th/0002230) *)
replrho={du^2->1/(4\[Rho])d\[Rho]^2,u->Sqrt[\[Rho]]};
g\[Rho]\[Rho]=Coefficient[Series[ds2u/.replrho,{\[Rho],0,-1}],d\[Rho]^2];
gvv0=Coefficient[SeriesCoefficient[ds2u/.replrho,{\[Rho],0,-1}],dv^2];
gvv2=Coefficient[SeriesCoefficient[ds2u/.replrho,{\[Rho],0,0}],dv^2];
gvv4=Coefficient[SeriesCoefficient[ds2u/.replrho,{\[Rho],0,1}],dv^2]/.Log[\[Rho]]->0;
hvv4=D[Coefficient[SeriesCoefficient[ds2u/.replrho,{\[Rho],0,1}],dv^2],Log[\[Rho]]];
gyy0=Coefficient[SeriesCoefficient[ds2u/.replrho,{\[Rho],0,-1}],dy^2];
gyy2=Coefficient[SeriesCoefficient[ds2u/.replrho,{\[Rho],0,0}],dy^2];
gyy4=Coefficient[SeriesCoefficient[ds2u/.replrho,{\[Rho],0,1}],dy^2]/.Log[\[Rho]]->0;
hyy4=D[Coefficient[SeriesCoefficient[ds2u/.replrho,{\[Rho],0,1}],dy^2],Log[\[Rho]]];
gxx0=Coefficient[SeriesCoefficient[ds2u/.replrho,{\[Rho],0,-1}],dx1^2];
gxx2=Coefficient[SeriesCoefficient[ds2u/.replrho,{\[Rho],0,0}],dx1^2];
gxx4=Coefficient[SeriesCoefficient[ds2u/.replrho,{\[Rho],0,1}],dx1^2]/.Log[\[Rho]]->0;
hxx4=D[Coefficient[SeriesCoefficient[ds2u/.replrho,{\[Rho],0,1}],dx1^2],Log[\[Rho]]];
g0=DiagonalMatrix[{gvv0,gyy0,gxx0,gxx0}];
g0inv=Inverse[g0];
g2=DiagonalMatrix[{gvv2,gyy2,gxx2,gxx2}];
g4=DiagonalMatrix[{gvv4,gyy4,gxx4,gxx4}];
h14=DiagonalMatrix[{hvv4,hyy4,hxx4,hxx4}];
G\[Rho]\[Rho]=g\[Rho]\[Rho];
Gvv=1/\[Rho](gvv0+\[Rho] gvv2+\[Rho]^2 gvv4+\[Rho]^2 Log[\[Rho]]hvv4);
Gyy=1/\[Rho](gyy0+\[Rho] gyy2+\[Rho]^2 gyy4+\[Rho]^2 Log[\[Rho]]hyy4);
Gxx=1/\[Rho](gxx0+\[Rho] gxx2+\[Rho]^2 gxx4+\[Rho]^2 Log[\[Rho]]hxx4);
Gdd=DiagonalMatrix[{G\[Rho]\[Rho],Gvv,Gyy,Gxx,Gxx}];


(* ::Input::Initialization:: *)
(* A,B and S field in FG coordinates *)
AFG[\[Rho]_,v_]:=Collect[Series[Ab[z,t]/.replFG/.replrho,{\[Rho],0,3}],{\[Rho],Log[\[Rho]]}];
BFG[\[Rho]_,v_]:=Collect[Series[Bb[z,t]/.replFG/.replrho,{\[Rho],0,3}],{\[Rho],Log[\[Rho]]}];
SFG[\[Rho]_,v_]:=Collect[Series[Sb[z,t]/.replFG/.replrho,{\[Rho],0,3}],{\[Rho],Log[\[Rho]]}];


(* ::Section::Initialization:: *)
(*Holographic EMT*)


(* ::Input::Initialization:: *)
Print["Holographic EMT:"]
(* Holographic energy-momentum tensor from Eq.(3.15) of (0002230) *)
EMTdd=g4-1/8g0(Tr[g0inv.g2]^2-Tr[g0inv.g2.g0inv.g2])-1/2g2.g0inv.g2+g2 Tr[g0inv.g2]//Simplify;
Print["\!\(\*SubscriptBox[\(T\), \(\[Mu]\[Nu]\)]\)[v]=",MatrixForm[EMTdd]]


(* ::Input::Initialization:: *)
Print["Trace of the EMT:"]
(* the EMT is traceless *)
EMTud=g0inv.EMTdd;
Print["\!\(\*SubsuperscriptBox[\(T\), \(\[Mu]\), \(\[Mu]\)]\)[v]=",Tr[EMTud]//Expand]


(* ::Input::Initialization:: *)
(* Coordinates and boundary metric *)
xmuB={v,y,x1,x2};
dB=Length[xmuB];
metricB=g0;
metricBinv=Inverse[g0];


(* ::Input::Initialization:: *)
(* Christoffel symbol of the boundary metric *)
GuddB=Table[Sum[1/2metricBinv[[i,l]](D[metricB[[k,l]],xmuB[[j]]]+D[metricB[[j,l]],xmuB[[k]]]-D[metricB[[j,k]],xmuB[[l]]]),{l,1,dB}],{i,1,dB},{j,1,dB},{k,1,dB}];


(* ::Input::Initialization:: *)
(* the EMT is covariantly conserved *)
EMTuu=g0inv.EMTud;
(* relation obtained in the near boundary analysis *)
a4dv=Derivative[1][Subscript[a, 4]][t]/.nbSol/.t->v;
repla4dv={Derivative[1][Subscript[a, 4]][v]->a4dv};
Print["Covariant derivative of the EMT:"]
(* covariant derivative: Subscript[D, \[Mu]]T^\[Mu]\[Nu]=\!\(
\*SubscriptBox[\(\[PartialD]\), \(\[Mu]\)]
\*SuperscriptBox[\(T\), \(\[Mu]\[Nu]\)]\)+\!\(
\*SubsuperscriptBox[\(\[CapitalGamma]\), \(\[Sigma]\[Mu]\), \(\[Mu]\)]
\*SuperscriptBox[\(T\), \(\[Sigma]\[Nu]\)]\)+\!\(
\*SubsuperscriptBox[\(\[CapitalGamma]\), \(\[Sigma]\[Mu]\), \(\[Nu]\)]
\*SuperscriptBox[\(T\), \(\[Mu]\[Sigma]\)]\) *)
Print["\!\(\*SubscriptBox[\(D\), \(\[Mu]\)]\)\!\(\*SuperscriptBox[\(T\), \(\[Mu]\[Nu]\)]\)==\!\(\*SubscriptBox[\(\[PartialD]\), \(\[Mu]\)]\)\!\(\*SuperscriptBox[\(T\), \(\[Mu]\[Nu]\)]\)+\!\(\*SubsuperscriptBox[\(\[CapitalGamma]\), \(\[Sigma]\[Mu]\), \(\[Mu]\)]\)\!\(\*SuperscriptBox[\(T\), \(\[Sigma]\[Nu]\)]\)+\!\(\*SubsuperscriptBox[\(\[CapitalGamma]\), \(\[Sigma]\[Mu]\), \(\[Nu]\)]\)\!\(\*SuperscriptBox[\(T\), \(\[Mu]\[Sigma]\)]\)=",MatrixForm[Table[Sum[D[EMTuu[[i,j]],xmuB[[i]]],{i,1,dB}],{j,1,dB}]+
Table[Sum[GuddB[[k,i,j]]EMTuu[[i,j]],{i,1,dB},{j,1,dB}],{k,1,dB}]+
Table[Sum[GuddB[[i,j,i]]EMTuu[[j,k]],{i,1,dB},{j,1,dB}],{k,1,dB}]/.repla4dv//Simplify]]


(* ::Section::Initialization:: *)
(*Equations of motion in characteristic form*)


(* ::Input::Initialization:: *)
(* replacements for the modified derivatives: \!\(
\*SubscriptBox[\(\[PartialD]\), \(t\)]f\)\[Rule]Overscript[f, .]-1/2A*\!\(
\*SubscriptBox[\(\[PartialD]\), \(r\)]f\) *)
rule1t=Flatten[Table[{
\!\(\*SuperscriptBox[\(S\), 
TagBox[
RowBox[{"(", 
RowBox[{"ddr", ",", "1"}], ")"}],
Derivative],
MultilineFunction->None]\)[r,t]->D[\!\(\*SuperscriptBox[\(S\), 
TagBox[
RowBox[{"(", 
RowBox[{"0", ",", "1"}], ")"}],
Derivative],
MultilineFunction->None]\)[r,t]-1/2 A[r,t] \!\(\*SuperscriptBox[\(S\), 
TagBox[
RowBox[{"(", 
RowBox[{"1", ",", "0"}], ")"}],
Derivative],
MultilineFunction->None]\)[r,t],{r,ddr}],
\!\(\*SuperscriptBox[\(A\), 
TagBox[
RowBox[{"(", 
RowBox[{"ddr", ",", "1"}], ")"}],
Derivative],
MultilineFunction->None]\)[r,t]->D[\!\(\*SuperscriptBox[\(A\), 
TagBox[
RowBox[{"(", 
RowBox[{"0", ",", "1"}], ")"}],
Derivative],
MultilineFunction->None]\)[r,t]-1/2 A[r,t] \!\(\*SuperscriptBox[\(A\), 
TagBox[
RowBox[{"(", 
RowBox[{"1", ",", "0"}], ")"}],
Derivative],
MultilineFunction->None]\)[r,t],{r,ddr}],
\!\(\*SuperscriptBox[\(B\), 
TagBox[
RowBox[{"(", 
RowBox[{"ddr", ",", "1"}], ")"}],
Derivative],
MultilineFunction->None]\)[r,t]->D[\!\(\*SuperscriptBox[\(B\), 
TagBox[
RowBox[{"(", 
RowBox[{"0", ",", "1"}], ")"}],
Derivative],
MultilineFunction->None]\)[r,t]-1/2 A[r,t] \!\(\*SuperscriptBox[\(B\), 
TagBox[
RowBox[{"(", 
RowBox[{"1", ",", "0"}], ")"}],
Derivative],
MultilineFunction->None]\)[r,t],{r,ddr}],
\!\(\*SuperscriptBox[\(S\), 
TagBox[
RowBox[{"(", 
RowBox[{"ddr", ",", "2"}], ")"}],
Derivative],
MultilineFunction->None]\)[r,t]->D[\!\(\*SuperscriptBox[\(S\), 
TagBox[
RowBox[{"(", 
RowBox[{"0", ",", "2"}], ")"}],
Derivative],
MultilineFunction->None]\)[r,t]-1/2 A[r,t] \!\(\*SuperscriptBox[\(S\), 
TagBox[
RowBox[{"(", 
RowBox[{"1", ",", "1"}], ")"}],
Derivative],
MultilineFunction->None]\)[r,t],{r,ddr}],
\!\(\*SuperscriptBox[\(A\), 
TagBox[
RowBox[{"(", 
RowBox[{"ddr", ",", "2"}], ")"}],
Derivative],
MultilineFunction->None]\)[r,t]->D[\!\(\*SuperscriptBox[\(A\), 
TagBox[
RowBox[{"(", 
RowBox[{"0", ",", "2"}], ")"}],
Derivative],
MultilineFunction->None]\)[r,t]-1/2 A[r,t] \!\(\*SuperscriptBox[\(A\), 
TagBox[
RowBox[{"(", 
RowBox[{"1", ",", "1"}], ")"}],
Derivative],
MultilineFunction->None]\)[r,t],{r,ddr}],
\!\(\*SuperscriptBox[\(B\), 
TagBox[
RowBox[{"(", 
RowBox[{"ddr", ",", "2"}], ")"}],
Derivative],
MultilineFunction->None]\)[r,t]->D[\!\(\*SuperscriptBox[\(B\), 
TagBox[
RowBox[{"(", 
RowBox[{"0", ",", "2"}], ")"}],
Derivative],
MultilineFunction->None]\)[r,t]-1/2 A[r,t] \!\(\*SuperscriptBox[\(B\), 
TagBox[
RowBox[{"(", 
RowBox[{"1", ",", "1"}], ")"}],
Derivative],
MultilineFunction->None]\)[r,t],{r,ddr}]
},{ddr,0,2}]];
rule2t={
\!\(\*SuperscriptBox[\(S\), 
TagBox[
RowBox[{"(", 
RowBox[{"0", ",", "2"}], ")"}],
Derivative],
MultilineFunction->None]\)[r,t]->D[\!\(\*SuperscriptBox[\(S\), 
TagBox[
RowBox[{"(", 
RowBox[{"0", ",", "1"}], ")"}],
Derivative],
MultilineFunction->None]\)[r,t]-1/2 A[r,t] \!\(\*SuperscriptBox[\(S\), 
TagBox[
RowBox[{"(", 
RowBox[{"1", ",", "0"}], ")"}],
Derivative],
MultilineFunction->None]\)[r,t],t],
\!\(\*SuperscriptBox[\(A\), 
TagBox[
RowBox[{"(", 
RowBox[{"0", ",", "2"}], ")"}],
Derivative],
MultilineFunction->None]\)[r,t]->D[\!\(\*SuperscriptBox[\(A\), 
TagBox[
RowBox[{"(", 
RowBox[{"0", ",", "1"}], ")"}],
Derivative],
MultilineFunction->None]\)[r,t]-1/2 A[r,t] \!\(\*SuperscriptBox[\(A\), 
TagBox[
RowBox[{"(", 
RowBox[{"1", ",", "0"}], ")"}],
Derivative],
MultilineFunction->None]\)[r,t],t],
\!\(\*SuperscriptBox[\(B\), 
TagBox[
RowBox[{"(", 
RowBox[{"0", ",", "2"}], ")"}],
Derivative],
MultilineFunction->None]\)[r,t]->D[\!\(\*SuperscriptBox[\(B\), 
TagBox[
RowBox[{"(", 
RowBox[{"0", ",", "1"}], ")"}],
Derivative],
MultilineFunction->None]\)[r,t]-1/2 A[r,t] \!\(\*SuperscriptBox[\(B\), 
TagBox[
RowBox[{"(", 
RowBox[{"1", ",", "0"}], ")"}],
Derivative],
MultilineFunction->None]\)[r,t],t]
};


(* ::Input::Initialization:: *)
(* transforming the Einstein equations to characteristic form *)
EOM1=EE[[1]]/.rule2t/.rule1t;
replEOM1=Solve[EOM1==0,\!\(\*SuperscriptBox[\(S\), 
TagBox[
RowBox[{"(", 
RowBox[{"2", ",", "0"}], ")"}],
Derivative],
MultilineFunction->None]\)[r,t]]//Flatten


(* ::Input::Initialization:: *)
EOM2=EE[[3]]/.rule2t/.rule1t//FullSimplify;
replEOM2=Solve[EOM2==0,\!\(\*SuperscriptBox[\(S\), 
TagBox[
RowBox[{"(", 
RowBox[{"1", ",", "1"}], ")"}],
Derivative],
MultilineFunction->None]\)[r,t]]/.replEOM1//FullSimplify//Flatten


(* ::Input::Initialization:: *)
EOMaux=EE[[4]]/.rule2t/.rule1t//Simplify;
replEOMaux=Solve[EOMaux==0,\!\(\*SuperscriptBox[\(A\), 
TagBox[
RowBox[{"(", 
RowBox[{"2", ",", "0"}], ")"}],
Derivative],
MultilineFunction->None]\)[r,t]]/.replEOM2/.replEOM1//FullSimplify//Flatten


(* ::Input::Initialization:: *)
EOM3=EE[[2]]/.rule2t/.rule1t//Simplify;
replEOM3=Solve[EOM3==0/.replEOM2/.replEOM1/.replEOMaux,\!\(\*SuperscriptBox[\(B\), 
TagBox[
RowBox[{"(", 
RowBox[{"1", ",", "1"}], ")"}],
Derivative],
MultilineFunction->None]\)[r,t]]//FullSimplify//Flatten


(* ::Input::Initialization:: *)
EOM4=EOMaux;
replEOM4=replEOMaux/.replEOM2/.replEOM1/.replEOM3//FullSimplify//Flatten


(* ::Input::Initialization:: *)
EOM5=EE[[5]]/.rule2t/.rule1t//Simplify;
replEOM5=Solve[EOM5==0/.replEOM1/.replEOM2,\!\(\*SuperscriptBox[\(S\), 
TagBox[
RowBox[{"(", 
RowBox[{"0", ",", "2"}], ")"}],
Derivative],
MultilineFunction->None]\)[r,t]]//FullSimplify//Flatten


(* ::Input::Initialization:: *)
(* it is important to note that in EOMList time-derivative means dot-derivative *)
solEE=Flatten[{replEOM1,replEOM2,replEOM3,replEOM4,replEOM5}/.Rule->Equal];
replTex1={A[r,t]->A,S[r,t]->S,B[r,t]->B,\!\(\*SuperscriptBox[\(f_\), 
TagBox[
RowBox[{"(", 
RowBox[{"1", ",", "0"}], ")"}],
Derivative],
MultilineFunction->None]\)[_,_]->f',\!\(\*SuperscriptBox[\(f_\), 
TagBox[
RowBox[{"(", 
RowBox[{"2", ",", "0"}], ")"}],
Derivative],
MultilineFunction->None]\)[_,_]->f'',\!\(\*SuperscriptBox[\(f_\), 
TagBox[
RowBox[{"(", 
RowBox[{"1", ",", "1"}], ")"}],
Derivative],
MultilineFunction->None]\)[_,_]->\!\(\*OverscriptBox[\(f\), \(.\)]\)',\!\(\*SuperscriptBox[\(f_\), 
TagBox[
RowBox[{"(", 
RowBox[{"0", ",", "1"}], ")"}],
Derivative],
MultilineFunction->None]\)[_,_]->\!\(\*OverscriptBox[\(f\), \(.\)]\),\!\(\*SuperscriptBox[\(f_\), 
TagBox[
RowBox[{"(", 
RowBox[{"0", ",", "2"}], ")"}],
Derivative],
MultilineFunction->None]\)[_,_]->\!\(\*OverscriptBox[\(f\), \(\[DoubleDot]\)]\)};
Print["Characteristic form of the Einstein equations:"]
Print[solEE//.replTex1//Expand//TableForm]


(* ::Input::Initialization:: *)
(* LaTeX input *)
solEE//.replTex1//Expand//TableForm//TeXForm


(* ::Section::Initialization:: *)
(*Removing singularities*)


(* ::Input::Initialization:: *)
(* loading the near boundary solution *)
SetDirectory[NotebookDirectory[]];
order=5;resFile="NBsolAdS5Ord5.res";
nbSol=Get[resFile];
Clear[Ab,Sb,Bb];
Ab[z_,t_]=Asb[z,t,order]/.\[Epsilon]->1/.nbSol;
Sb[z_,t_]=Ssb[z,t,order]/.\[Epsilon]->1/.nbSol;
Bb[z_,t_]=Bsb[z,t,order]/.\[Epsilon]->1/.nbSol;
replBdry={A->Ab,S->Sb,B->Bb};


(* ::Input::Initialization:: *)
(* leading parts that are known explicitely from the near boundary expansion *)
subtractA[z_,t_]=Normal[Series[A[z,t]/.replBdry,{z,0,1}]];
subtractS[z_,t_]=Normal[Series[S[z,t]/.replBdry,{z,0,3}]];
(*subtractB[z_,t_]=Normal[Series[B[z,t]/.replBdry,{z,0,3}]];*)
subtractB[z_,t_]=0;
(* here we write the dot derivative explicitely in the z-coordinate: Sdot=dS/dt-z^2/2A*dS/dz,... *)
subtractSdot[z_,t_]=Normal[Series[D[S[z,t],t]-1/2z^2A[z,t]D[S[z,t],z]/.replBdry,{z,0,1}]];
(*subtractBdot[z_,t_]=Normal[Series[D[B[z,t],t]-1/2z^2A[z,t]D[B[z,t],z]/.replBdry,{z,0,2}]];*)
subtractBdot[z_,t_]=0;
(* leading Log contributions *)
subtractLogA[z_,t_]=D[Normal[Series[A[z,t]/.replBdry,{z,0,3}]],Log[z]]Log[z];
subtractLogS[z_,t_]=D[Normal[Series[S[z,t]/.replBdry,{z,0,5}]],Log[z]]Log[z];
subtractLogB[z_,t_]=D[Normal[Series[B[z,t]/.replBdry,{z,0,5}]],Log[z]]Log[z];
subtractLogSdot[z_,t_]=D[Normal[Series[D[S[z,t],t]-1/2z^2A[z,t]D[S[z,t],z]/.replBdry,{z,0,3}]],Log[z]]Log[z];
subtractLogBdot[z_,t_]=D[Normal[Series[D[B[z,t],t]-1/2z^2A[z,t]D[B[z,t],z]/.replBdry,{z,0,4}]],Log[z]]Log[z];


(* ::Input::Initialization:: *)
(* replacements for switching to finite variables Af,Sf,Bf,... *)
replfunc={
A->(Evaluate[subtractA[#1,#2]+subtractLogA[#1,#2]+Af[#1,#2]#1^2]&),
S->(Evaluate[subtractS[#1,#2]+subtractLogS[#1,#2]+Sf[#1,#2]#1^4]&),
B->(Evaluate[subtractB[#1,#2]+subtractLogB[#1,#2]+Bf[#1,#2]#1^4]&),
Sdot->(Evaluate[subtractSdot[#1,#2]+subtractLogSdot[#1,#2]+Sdotf[#1,#2]#1^2]&),
Bdot->(Evaluate[subtractBdot[#1,#2]+subtractLogBdot[#1,#2]+Bdotf[#1,#2]#1^3]&)
};


(* ::Input::Initialization:: *)
(* define list of functions which we want to evolve numerically and assign them numbers B=1,S=2,... *)
funcs={B,S,Sdot,Bdot,A,Sdotdt};
funcsf={Bf,Sf,Sdotf,Bdotf,Af,Sdotdtf};
{nB,nS,nSdot,nBdot,nA,nSdotdt}={1,2,3,4,5,6};


(* ::Input::Initialization:: *)
(* introduce dotted variables and multiplying the Einstein equations with powers of S to make them simpler *)
repldot={\!\(\*SuperscriptBox[\(B\), 
TagBox[
RowBox[{"(", 
RowBox[{"ddr_", ",", "1"}], ")"}],
Derivative],
MultilineFunction->None]\)[r_,t]->\!\(\*SuperscriptBox[\(Bdot\), 
TagBox[
RowBox[{"(", 
RowBox[{"ddr", ",", "0"}], ")"}],
Derivative],
MultilineFunction->None]\)[r,t],\!\(\*SuperscriptBox[\(S\), 
TagBox[
RowBox[{"(", 
RowBox[{"ddr_", ",", "1"}], ")"}],
Derivative],
MultilineFunction->None]\)[r_,t]->\!\(\*SuperscriptBox[\(Sdot\), 
TagBox[
RowBox[{"(", 
RowBox[{"ddr", ",", "0"}], ")"}],
Derivative],
MultilineFunction->None]\)[r,t],\!\(\*SuperscriptBox[\(S\), 
TagBox[
RowBox[{"(", 
RowBox[{"0", ",", "2"}], ")"}],
Derivative],
MultilineFunction->None]\)[r,t]->\!\(\*SuperscriptBox[\(Sdot\), 
TagBox[
RowBox[{"(", 
RowBox[{"0", ",", "1"}], ")"}],
Derivative],
MultilineFunction->None]\)[r,t]+1/2 A[r,t] \!\(\*SuperscriptBox[\(Sdot\), 
TagBox[
RowBox[{"(", 
RowBox[{"1", ",", "0"}], ")"}],
Derivative],
MultilineFunction->None]\)[r,t]};
repldotback={Sdot->(\!\(\*SuperscriptBox[\(S\), 
TagBox[
RowBox[{"(", 
RowBox[{"0", ",", "1"}], ")"}],
Derivative],
MultilineFunction->None]\)[#1,#2]+1/2 A[#1,#2]\!\(\*SuperscriptBox[\(S\), 
TagBox[
RowBox[{"(", 
RowBox[{"1", ",", "0"}], ")"}],
Derivative],
MultilineFunction->None]\)[#1,#2]&),Bdot->(\!\(\*SuperscriptBox[\(B\), 
TagBox[
RowBox[{"(", 
RowBox[{"0", ",", "1"}], ")"}],
Derivative],
MultilineFunction->None]\)[#1,#2]+1/2 A[#1,#2]\!\(\*SuperscriptBox[\(B\), 
TagBox[
RowBox[{"(", 
RowBox[{"1", ",", "0"}], ")"}],
Derivative],
MultilineFunction->None]\)[#1,#2]&)};
multp={1,S[r,t],2S[r,t],S[r,t]^2,1};
newEE=multp(solEE/.Equal->Subtract)/.repldot//Expand//FullSimplify;
newEEz=newEE/.replrtoz//FullSimplify ;


(* ::Input::Initialization:: *)
(* Einstein equations in terms of finite variables *)
newEEfin=newEEz/.replfunc//Simplify;


(* ::Input::Initialization:: *)
(* determining the leading order in z of the Einstein equations *)
Series[newEEfin[[1]],{z,0,6}]
Series[newEEfin[[2]],{z,0,3}]
Series[newEEfin[[3]],{z,0,3}]
Series[newEEfin[[4]],{z,0,2}]
Series[newEEfin[[5]],{z,0,1}]


(* ::Input::Initialization:: *)
(* multiply the Einstein equations with the appropriate order of z such that the equations become finite at z=0  *)
powerz={z^-6,z^-3,z^-3,z^-2,z^-1};
newEEf=powerz*newEEfin//Simplify;


(* ::Section::Initialization:: *)
(*Making equations numerical*)


(* ::Input::Initialization:: *)
(* defining array "nums" that contains the numerical values of the metric functions at the grid points *)
Off[General::"partd"];
replnum ={
Bf[z,t]->nums[[1,1]],\!\(\*SuperscriptBox[\(Bf\), 
TagBox[
RowBox[{"(", 
RowBox[{"1", ",", "0"}], ")"}],
Derivative],
MultilineFunction->None]\)[z,t]->nums[[1,2]],\!\(\*SuperscriptBox[\(Bf\), 
TagBox[
RowBox[{"(", 
RowBox[{"2", ",", "0"}], ")"}],
Derivative],
MultilineFunction->None]\)[z,t]->nums[[1,3]],
Sf[z,t]->nums[[2,1]],\!\(\*SuperscriptBox[\(Sf\), 
TagBox[
RowBox[{"(", 
RowBox[{"1", ",", "0"}], ")"}],
Derivative],
MultilineFunction->None]\)[z,t]->nums[[2,2]],\!\(\*SuperscriptBox[\(Sf\), 
TagBox[
RowBox[{"(", 
RowBox[{"2", ",", "0"}], ")"}],
Derivative],
MultilineFunction->None]\)[z,t]->nums[[2,3]],
Sdotf[z,t]->nums[[3,1]],\!\(\*SuperscriptBox[\(Sdotf\), 
TagBox[
RowBox[{"(", 
RowBox[{"1", ",", "0"}], ")"}],
Derivative],
MultilineFunction->None]\)[z,t]->nums[[3,2]],\!\(\*SuperscriptBox[\(Sdotf\), 
TagBox[
RowBox[{"(", 
RowBox[{"2", ",", "0"}], ")"}],
Derivative],
MultilineFunction->None]\)[z,t]->nums[[3,3]],
Bdotf[z,t]->nums[[4,1]],\!\(\*SuperscriptBox[\(Bdotf\), 
TagBox[
RowBox[{"(", 
RowBox[{"1", ",", "0"}], ")"}],
Derivative],
MultilineFunction->None]\)[z,t]->nums[[4,2]],\!\(\*SuperscriptBox[\(Bdotf\), 
TagBox[
RowBox[{"(", 
RowBox[{"2", ",", "0"}], ")"}],
Derivative],
MultilineFunction->None]\)[z,t]->nums[[4,3]],
Af[z,t]->nums[[5,1]],\!\(\*SuperscriptBox[\(Af\), 
TagBox[
RowBox[{"(", 
RowBox[{"1", ",", "0"}], ")"}],
Derivative],
MultilineFunction->None]\)[z,t]->nums[[5,2]],\!\(\*SuperscriptBox[\(Af\), 
TagBox[
RowBox[{"(", 
RowBox[{"2", ",", "0"}], ")"}],
Derivative],
MultilineFunction->None]\)[z,t]->nums[[5,3]],
Sdotdtf[z,t]->nums[[6,1]],\!\(\*SuperscriptBox[\(Sdotdtf\), 
TagBox[
RowBox[{"(", 
RowBox[{"1", ",", "0"}], ")"}],
Derivative],
MultilineFunction->None]\)[z,t]->nums[[6,2]],\!\(\*SuperscriptBox[\(Sdotdtf\), 
TagBox[
RowBox[{"(", 
RowBox[{"2", ",", "0"}], ")"}],
Derivative],
MultilineFunction->None]\)[z,t]->nums[[6,3]],
Subscript[a, 4][t]->a4,Subscript[b, 4][t]->b4,Derivative[1][Subscript[b, 4]][t]->b4dt
};
(* replacements for the boundary values of the fields *)
replnumb = replnum /. {z -> 0, nums[[a_, b_]] -> nums[[a, b, -1]]};//Quiet
(* replacements for the boundary source *)
replb0 = Prepend[Table[\!\(\*
TagBox[
StyleBox[
RowBox[{
RowBox[{
RowBox[{"Derivative", "[", "i", "]"}], "[", "b0", "]"}], "[", "t", "]"}],
ShowSpecialCharacters->False,
ShowStringCharacters->True,
NumberMarks->True],
FullForm]\) ->  b0num[[i + 1]], {i, 7}], b0[t] -> b0num[[1]]];


(* ::Input::Initialization:: *)
(* updating the derivatives of the numerical functions using the spectal matrices  *)
update[n_]:=(
nums[[n,2]]=DSpec.nums[[n,1]];
nums[[n,3]]=D2Spec.nums[[n,1]];
b4=nums[[nB,1,-1]];
)
SetAttributes[update,Listable];


(* ::Input::Initialization:: *)
(* collecting the variables in the order in which they appear in the nested scheme *)
solvefor={Sf,Sdotf,Bdotf,Af};


(* ::Input::Initialization:: *)
(* building numerical arrays for the coefficient functions in the Einstein equations *)
Monitor[Do[
Subscript[As, i]=D[newEEf[[i]],\!\(\*SuperscriptBox[\(solvefor\[LeftDoubleBracket]i\[RightDoubleBracket]\), 
TagBox[
RowBox[{"(", 
RowBox[{"2", ",", "0"}], ")"}],
Derivative],
MultilineFunction->None]\)[z,t] ]//Expand//Simplify;
Subscript[Asb, i]=SeriesCoefficient[Subscript[As, i],{z,0,0}]//Simplify;
Subscript[Bs, i]=D[newEEf[[i]],\!\(\*SuperscriptBox[\(solvefor\[LeftDoubleBracket]i\[RightDoubleBracket]\), 
TagBox[
RowBox[{"(", 
RowBox[{"1", ",", "0"}], ")"}],
Derivative],
MultilineFunction->None]\)[z,t]]//Expand//Simplify;
Subscript[Bsb, i]=SeriesCoefficient[Subscript[Bs, i],{z,0,0}]//Simplify;
Subscript[Cs, i]=D[newEEf[[i]],solvefor[[i]][z,t]]//Expand//Simplify;
Subscript[Csb, i]=SeriesCoefficient[Subscript[Cs, i],{z,0,0}]//Simplify;
Subscript[Ss, i]=newEEf[[i]]/.solvefor[[i]]->(0&)//Expand//Simplify;
Subscript[Ssb, i]=SeriesCoefficient[Subscript[Ss, i],{z,0,0}]//Simplify;
,{i,Length[solvefor]}],i]


(* ::Input::Initialization:: *)
(* Replacing the analytic expressions with numerical arrays. *)
(* For example Subscript[As, 1] contains the analytic expression in front of the second derivative term in the first Einstein equation, *)
(* Subscript[As, 1+30] contains the same expression but with the analytic expression replaced by the numerical arrays. *)
makeformulas:=(
Do[Subscript[#, ii+30]=Subscript[#, ii]/.replnum/.replnumb/.replb0;,{ii,Length[solvefor]}]&/@{As,Asb,Bs,Bsb,Cs,Csb,Ss,Ssb};
Subscript[Bsb, 2+30]=0;
Subscript[Csb, 2+30]=1;
Subscript[Ssb, 2+30]=-SeriesCoefficient[D[S[z,t],t]-1/2z^2A[z,t]D[S[z,t],z]/.replBdry,{z,0,2}]/.Log[z]->0/.replnum/.replb0;
);
makeformulas;
(* The last Einstein equation is a constraint which we use to monitor the accuracy of the time evolution. *)
constraint=newEEf[[-1]]/.{\!\(\*SuperscriptBox[\(Sdotf\), 
TagBox[
RowBox[{"(", 
RowBox[{"ddz_", ",", "1"}], ")"}],
Derivative],
MultilineFunction->None]\)[z,t]-> \!\(\*SuperscriptBox[\(Sdotdtf\), 
TagBox[
RowBox[{"(", 
RowBox[{"ddz", ",", "0"}], ")"}],
Derivative],
MultilineFunction->None]\)[z,t]}//.replnum//.replb0//Simplify;


(* ::Input::Initialization:: *)
(* saving the numerical formulas to a file such that they can be loaded fast once they were generated before *)
restable=(Table[Subscript[#, ii+30],{ii,Length[solvefor]}]&/@{As,Asb,Bs,Bsb,Cs,Csb,Ss,Ssb})~Join~{constraint}//Simplify;
Put[restable,"formulaAdS5.dat"];


(* ::Input::Initialization:: *)
(* loading the formulas from file *)
restable=Get["formulaAdS5.dat"];
Do[Subscript[{As,Asb,Bs,Bsb,Cs,Csb,Ss,Ssb}[[kk]], ii+30]=restable[[kk,ii]],{ii,Length[solvefor]},{kk,8}];
constraint=restable[[-1]];
(* determining the order of the differential equations *)
solveorder=If[Subscript[As, #+30]===0,If[Subscript[Bs, #+30]===0,0,1],2]&/@Range[Length[solvefor]];


(* ::Input::Initialization:: *)
(* expressing the t-derivatives of B and b4 in terms of Bdot *)
dBfdt=(\!\(\*SuperscriptBox[\(Bf\), 
TagBox[
RowBox[{"(", 
RowBox[{"0", ",", "1"}], ")"}],
Derivative],
MultilineFunction->None]\)[z,t]/.Flatten[Solve[D[B[z,t],t]==Bdot[z,t]+z^2/2 A[z,t]D[B[z,t],z]/.replfunc,\!\(\*SuperscriptBox[\(Bf\), 
TagBox[
RowBox[{"(", 
RowBox[{"0", ",", "1"}], ")"}],
Derivative],
MultilineFunction->None]\)[z,t]]]);
dBfdtn=dBfdt/.replnum/.replb0;


(* ::Input::Initialization:: *)
dBfdtb=SeriesCoefficient[dBfdt,{z,0,0}];
dBfdtbn=dBfdtb/.replnum/.replnumb/.replb0;
b4dt=dBfdtbn;


(* ::Input::Initialization:: *)
(* The expression for the t-derivative of a4 follows from conservation of the stress tensor. *)
(* We have obtained this expression from the near boundary expansion. *)
a4dtn=Derivative[1][Subscript[a, 4]][t]/.nbSol/.replnum/.replb0;


(* ::Input::Initialization:: *)
(* relating the YM-equation (f''[t](1+Phi[t])+2f[t]^3(1+Phi[t])+f'[t]Phidt[t]) to the boundary source b0[t] *)
Phi[t_]=1/16*Sqrt[3](64* Subscript[b, 4][t]-7 D[b0[t],t]^2D[b0[t],{t,2}]);
Phinum=Phi[t]/.replnum/.replb0;
Phidt[t_]=D[Phi[t],t];
Phidtnum=Phidt[t]/.replnum/.replb0;


(* ::Section::Initialization:: *)
(*Initializing the Chebyshev grid and time stepping*)


(* ::Input::Initialization:: *)
(* the function "initGrid" builds the Chebyshev grid and the spectral differentialization matrices DSpec and D2Spec *)
initGrid[points_]:=(
nz=points;
z=N[Table[Cos[(r \[Pi])/(nz-1)],{r,0,nz-1}]];
DSpec=Table[If[i==nz&&j==nz,-((2(nz-1)^2+1)/6),If[i==1&&j==1,(2(nz-1)^2+1)/6,If[i==j,-z[[i]]/(2(1-z[[i]]^2)),(-1)^(i+j)/(z[[i]]-z[[j]]) If[i==1||i==nz,2,1]/If[j==1||j==nz,2,1]]]],{i,nz},{j,nz}];
z=z/2+1/2;
DSpec*=2;
D2Spec=DSpec.DSpec;z);


(* ::Input::Initialization:: *)
(* initializing the time derivatives of b0[t] *)
InitTrepl[b0ini_,t_]:=Module[{},bini=b0ini;
a1=Prepend[Table[D[bini,{ti,i}],{i,6}],bini];
b0num=N[a1/.ti->t]//Chop;]


(* ::Input::Initialization:: *)
(* setting the initial conditions for the time evolution *)
init[profileB_,a4ini_,b0ini_,tini_,npoints_]:=(
initGrid[npoints];
tinitial=t=tini;
nums=Table[0,{Length[funcs]},{3},{nz}];
a4=a4ini;
Bfini=Function[z,(profileB[z]/z^4)/.replnum];
nums[[nB,1]]=Bfini/@z;
nums[[nB,1,-1]]=SeriesCoefficient[Series[Bfini[ztmp,tini],{ztmp,0,0}],0];
Blist=Bdtlist=Sdotlist=a4dtlist=timelist=a4list=b4list=b4dtlist=constrlist=Philist=Phidtlist={};
b0numlist={b0num};
update[nB];
)


(* ::Input::Initialization:: *)
(* loop for the time evolution *)
order=4;
(* the initial time steps are smaller *)
dts=Table[10^(i/2),{i,-8,-1}]~Join~{1-Sum[10^(i/2),{i,-8,-1}]}~Join~{1};
step[change_,append_:-1,b0ini_]:=(
InitTrepl[b0ini,t];
Do[
If[solveorder[[kk]]==2,Anum=Subscript[As, kk+30];Anum[[-1]]=Subscript[Asb, kk+30]];
If[solveorder[[kk]]>=1,Bnum=Subscript[Bs, kk+30];Bnum[[-1]]=Subscript[Bsb, kk+30]];
Cnum=Subscript[Cs, kk+30];Cnum[[-1]]=Subscript[Csb, kk+30];
Snum=Subscript[Ss, kk+30];Snum[[-1]]=Subscript[Ssb, kk+30];

eqm=Switch[solveorder[[kk]],
2,DiagonalMatrix[SparseArray[Anum]].D2Spec+DiagonalMatrix[SparseArray[Bnum]].DSpec+DiagonalMatrix[SparseArray[Cnum]],
1,DiagonalMatrix[SparseArray[Bnum]].DSpec+DiagonalMatrix[SparseArray[Cnum]],
0,DiagonalMatrix[SparseArray[Cnum]]];

nums[[kk+1,1]]=LinearSolve[eqm,-Snum];
update[kk+1];
,{kk,1,Length[solvefor]}];

nums[[nSdotdt,1]]=If[Length[timelist]>5&&((timelist[[-6]]-timelist[[-5]])-(timelist[[-3]]-timelist[[-2]])<10^-12),
1/(12 dt) (3 Sdotlist[[-4]]-16 Sdotlist[[-3]]+36 Sdotlist[[-2]]-48 Sdotlist[[-1]]+25 nums[[nSdot,1]]),
Derivative[1][Interpolation[Thread[{timelist~Join~{t},Sdotlist~Join~{nums[[nSdot,1]]}}],InterpolationOrder->Min[order-1,Length[timelist]]]][t]];
update[nSdotdt];
connum=constraint[[;;-2]]~Join~{0};
mcon=Max[Abs[connum]];

If[append!= -1,
Bdtlist=Append[Drop[Bdtlist,append],dBfdtn[[;;-2]]~Join~{dBfdtbn}];
a4dtlist=Append[Drop[a4dtlist,append],a4dtn];
Sdotlist=Append[Drop[Sdotlist,append],nums[[nSdot,1]]];
(*Blist=Append[Blist,nums\[LeftDoubleBracket]nB,1\[RightDoubleBracket]];*)
timelist=Append[timelist,t];
b4dtlist=Append[b4dtlist,dBfdtbn];
a4list=Append[a4list,a4];
b4list=Append[b4list,b4];
b0numlist=Append[b0numlist,b0num];
Philist=Append[Philist,Phinum];
Phidtlist=Append[Phidtlist,Phidtnum];
constrlist=Append[constrlist,mcon];
];

If[change,
t=t+dt dts[[Min[Length[dts],Length[timelist]]]];
If[Length[timelist]>5&&((timelist[[-6]]-timelist[[-5]])-(timelist[[-3]]-timelist[[-2]])<10^-12),
nums[[nB,1]]+=dt/24 (55 Bdtlist[[-1]]-59 Bdtlist[[-2]]+37 Bdtlist[[-3]]-9 Bdtlist[[-4]]);
a4+=dt/24 (55 a4dtlist[[-1]]-59 a4dtlist[[-2]]+37 a4dtlist[[-3]]-9 a4dtlist[[-4]]);,
nums[[nB,1]]+=\!\(
\*SubsuperscriptBox[\(\[Integral]\), \(timelist[[\(-1\)]]\), \(t\)]\(\(Interpolation[Table[{timelist[[\(-j\)]], Bdtlist[[\(-j\)]]}, {j, Min[Length[Bdtlist], 6]}], InterpolationOrder -> Min[order - 1, Length[Bdtlist] - 1]]\)[T]\ \[DifferentialD]T\)\);
a4+=\!\(
\*SubsuperscriptBox[\(\[Integral]\), \(timelist[[\(-1\)]]\), \(t\)]\(\(Interpolation[Table[{timelist[[\(-j\)]], a4dtlist[[\(-j\)]]}, {j, Min[Length[a4dtlist], 6]}], InterpolationOrder -> Min[order - 1, Length[a4dtlist] - 1]]\)[T]\ \[DifferentialD]T\)\);
];
update[nB];
];
)//Timing


(* ::Section::Initialization:: *)
(*Running*)


(* ::Input::Initialization:: *)
(* saving and loading results *)
saverun[file_]:=Put[{dt,Bfini,b0numlist,nz,t,tinitial,Bdtlist,Sdotlist,a4dtlist,timelist,a4list,a4dtlist,b4list,b4dtlist,Philist,Phidtlist,constrlist},file]
loadrun[file_]:=({dt,Bfini,b0numlist,nz,t,tinitial,Bdtlist,Sdotlist,a4dtlist,timelist,a4list,a4dtlist,b4list,b4dtlist,Philist,Phidtlist,constrlist}=Get[file];
initGrid[nz]);


(* ::Input::Initialization:: *)
(* function for plotting intermediate steps during the time evolution *)
plotfunc[n_]:=Show[plj[numsstart[[n,1]]],pl[nums[[n,1]]],PlotLabel->funcsf[[n]],ImageSize->350,BaseStyle->20,AxesLabel->"z",PlotRange->{Min[numsstart[[n,1]]~Join~nums[[n,1]]],Max[numsstart[[n,1]]~Join~nums[[n,1]]]}]


(* ::Input::Initialization:: *)
(* routine that runs the time evolution *)
run[profile_,a4ini_,b0ini_,tini_,tend_,gridpoints_,name_]:=(session=name;
init[profile,a4ini,b0ini,tini,gridpoints];
(*dt=1/(25 nz^2)//N;*)
dt=1/(2 nz^2)//N;
ttotal=0;
Monitor[Do[ttotal+=First[Timing[step[True,If[Length[timelist]<20,0,1],b0ini]]];
If[stepnumber==4,numsstart=nums];
If[Mod[stepnumber,25]==0,timeleft=ToString[(Round[(tend-tini)/dt]-stepnumber) ttotal/(stepnumber)//Round]<>" seconds left.";
(*plots={plotfunc/@Range[Length[funcs]],Show[pl[connum],PlotLabel\[Rule]"constraint",ImageSize\[Rule]350,BaseStyle\[Rule]20,AxesLabel\[Rule]"z"]};*)];
If[Mod[stepnumber,500]==0,Print[{stepnumber,t,ttotal,mcon,a4,b4}];
saverun[name<>".dat"];];
If[Max[Abs[nums[[nB,1]]]]>10^4||(mcon[[2]]>5),Throw[Null]];,{stepnumber,Round[(tend-tini)/dt]}];,{t,mcon,timeleft(*,plots*)}])


(* ::Section::Closed:: *)
(*Run simulation*)


(* ::Input:: *)
(*(* parameters of the simulations *)*)
(*tstart=0.0; tend=10;Ngrid=30;*)
(*a4now=-1;*)
(*b0now[t_]:=0;*)


(* ::Input:: *)
(*(* running the simulation *)*)
(*run[#^4&,a4now,b0now[ti],tstart,tend,Ngrid,"homogeneous3"]//Quiet;*)


(* ::Section::Closed:: *)
(*Post-Processing *)


(* ::Input:: *)
(*(* lading results for IC1 *)*)
(*SetDirectory[NotebookDirectory[]];*)
(*loadrun["homogeneous2.dat"];*)
(*a4int=Interpolation[Thread[{timelist,a4list}]];*)
(*b4int=Interpolation[Thread[{timelist,b4list}]];*)


(* ::Input:: *)
(*(* plotting the constraint  *)*)
(*constrPlot=ListLogPlot[Thread[{timelist/Pi,constrlist}][[1;;-1;;5]],LabelStyle->{Black,20},PlotStyle->Thick,AxesLabel->{"Tt","max[constraint]"},PlotRange->{{0,2},{10^-13,10^-7}},ImageSize->500]*)


(* ::Input:: *)
(*Export["constraint.pdf",constrPlot]*)


(* ::Input:: *)
(*(* holographic EMT *)*)
(*Table[EMTdd[[i,i]],{i,1,Length[EMTdd]}]*)
(*epsilon[t_]:=-(3/4) a4int[t];*)
(*pl[t_]:=-(1/4) a4int[t]-2 b4int[t];*)
(*pt[t_]:=-(1/4) a4int[t]+b4int[t];*)


(* ::Input:: *)
(*(* plotting holographic EMT components *)*)
(*plotHolEMT=Plot[{epsilon[Pi t],pl[Pi t],pt[Pi t]},{t,0.,2},LabelStyle->{Black,20},AxesLabel->{"Tt"},PlotRange->All,PlotStyle->{{Thick,Red},{Thick,Green},{Thick,Blue}},PlotLegends->Placed[{"\!\(\*FormBox[\(\[ScriptCapitalE]\),*)
(*TraditionalForm]\)","\!\(\*SubscriptBox[\(TraditionalForm\`\[ScriptCapitalP]\), \(\[RightDoubleBracketingBar]\)]\)","\!\(\*SubscriptBox[\(TraditionalForm\`\[ScriptCapitalP]\), \(\[Perpendicular]\)]\)"},Above],ImageSize->600]*)


(* ::Input:: *)
(*Show[Graphics[{{{}, {}, Annotation[{RGBColor[1, 0, 0], Thickness[Large], Opacity[1.], Line[CompressedData["*)
(*1:eJxTTMoPSmViYGAwAWIQ7ZX64uKBl6l2DGDwwn6BqH9bl+QUexi/kWOzeqfk*)
(*Ejg/InLh+g8ft8D53P8bY9zl9sP5qoHvXhzuPgLnl5+ztJ/x5SScL75F8B4j*)
(*43k4f9sj5zPnbS7B+Yey5kpHvrkC539Vs+Vfm3sdzmf6uVpAWvUWnG/2P2Fb*)
(*8Lo7cP6ZFpNzx7Tuw/nyE17/XPnuAZz/+vv9e1cNHsH5S17tPsXX9xjO55Nd*)
(*/v7Jkydwfk/m1rQFKs/g/G8Oq59k1D+H80uMjmwRO/wCzr+/7kj3RcVXcP7v*)
(*rCuzNatew/najuz/0na/gfPdFymmbxR/hwgv2XilhzHv4XydN6Hc130/wPnC*)
(*/B84jvxC8Fn2z16vPOcjnJ9tsvsmp/cnOH+1Z9KCfd8Q/K+vpQSeTfsM528N*)
(*dmee4vIFzq/3aJ2/5RmCr3jX68bi1q9wvpV5yxRJo29wfozc+Z+MlxH8r6d6*)
(*HxU0fofzZ7DElQSp/EC4/97WRUxHEHzW2sYQy8KfcD7Dyrj1j4R+wflFs9b3*)
(*Ch1C8Pdqd3AI5v2G8wW9inSa+P7A+Xd0L1yu3IPgb5hnr/Ik7i+cH7veTFD2*)
(*H4Lf0Lqw8c6qf3D+2heys/l8/sP5Si0iHuzLEPzp8txf//9H8AFVHxMn*)
(*"]]}, "Charting`Private`Tag$393242#1"], Annotation[{RGBColor[0, 1, 0], Thickness[Large], Opacity[1.], Line[CompressedData["*)
(*1:eJwd13c8lW8UAHBERkIIV4oiUUIRJTk3KcqIUlGRjJaRrChCIrMhycqIZCUz*)
(*mzcZlXDt7L23a1zjXr/n/f3l8/28132f55zznPPcvaYPL1kw0NHRyTDS0eF/*)
(*L1iMNWDjFirqnJb76Oj4McYzgra9BDV4Spfd28vMj8nFhEVlE/RhaOpMdi4n*)
(*P6ZZ25b1kWAOR5tntp4R5seuma+UBBMcgHWqeXToFD/WLfPyTBDhBfwVuv1b*)
(*w4Ufi9150cefEAKt7/wIvmP82NPaDxVXC0NgYXDxjvgsP3bVu49B9MZ72O0S*)
(*+KlwiR9jX3rkXhQVCrVuZkMVDAKYc3OI85RwOJQFcXubCAlgeu86HuiIR4P9*)
(*0y+zWdoCmCdL9gE/QgJ0tPT4ZCUJYAfV3/pm2SaAVKZ7lVu6ANbo/XC8szoB*)
(*evaW9CvnCGBiW6RSpZ0+Q+HborjQMgGsipog3dyYCKEpYiz5LQIYOzn0mHBg*)
(*Mig8ol7N2BTAco84hmoMJkN+kpBsPiMBM7a9vPJIKQW4eJYLs1kJ2LdpzoKK*)
(*sRQgc2u2efMQsEtjvsoPzqbBkxtXi96LE7Cw7idncmnpcPKFj7zJBQJ2Wsgw*)
(*oefKN3jc6LU3ToeATVxXZGL5+g1m5SocWi4RsD1TLmZfD2dA2fnrigI3CNjD*)
(*+wr76boyIEjtRgCTJQGr8P3cEsKcCcwn7Fj7bQiYQBKvj6RcJuRl25V9syNg*)
(*2MjCyCX/TJCkfi8Sf0LAuM2/JX0+ngUjozZ6T3wJ2B2vPYZK5llAq1+amAkg*)
(*YEWfgljrX2eBzdOSoKuvCZh5v+UDykgWsKkx/aO9J2C5xhJSmu+zQezZVE/W*)
(*JwJmYBj3bW4+B7LZNbPjCwjYbKc+2yfBXDjr2zhCKCZgPsYsFpfP5ML99HjX*)
(*F6UELNvsoeD3kFxoobkXKfwkYBw2p7yfKn4HCa2qVY6/BOzz3FyvlMl3YM3z*)
(*zT5QR8CU7ROUeny/w/eA9kR5EgF74LxtjtjxHWSY/cUONROwyuft17e65UHI*)
(*mE+fXycBu8kUlJv3OQ8yXT/WnOkmYAsviVz36/Lgl99y+XwPARMJ+lJZI5wP*)
(*2axHSyUHCFge5w0RN418cK0pyc0eJGA6wRxPpR/lw66ontgjwwTM9YOj7Nvy*)
(*fFCxPXWcdYyA8QhKBqhO5oNcgH2b6TgBS4nqGibzFIC8kYxuxgQBa4s7E3nV*)
(*ogC2PFDOFp8mYDZiK0vMrwqAh9L5+uIMAWP8kqJb8B09/xl90GqWgB1N27FV*)
(*iKUQcj/qPvOYJ2C/pStNamULwWzrgoTzAgEzyXQuemZYCAEeQo5mZAK2LCfF*)
(*J/u8EHgHXG+oLhKwoO+9tv0phWB541sz7xIBEzvxria4qRAE+IntHciFRefE*)
(*1TYKoaQj9XbIMgEbw752JmoXwX3j0t7+FQLmfua2goFTEeRd5a9ypBAwvire*)
(*t6wxRWBCn7mDipym8WuysLoIbrpWFD5eJWBnap6es5orAtp+94Jh5A5tmbjd*)
(*hGKwktyyVX2NgNmSBtbrThfDW8Z7YZHIzJdDr3o8QM+3Ft0bRI5uOZ955F0x*)
(*nKjctNmzTsCOGVC3DRYVQ1KAbLwW8t+OjDshQ8WQUqxCtUY2NTL/cXZ7CXAf*)
(*5nF8jrzayy+0cqwEOu1eMfgjvzGtcUoyLoGHkn4fXyCLDz9rMHxZAqUHm0/Z*)
(*IZfcPSq1LaME6oqPd15C1p8Y9in+VwIZ2c7W4siTVuH91vSl4PtSZ2oKre/5*)
(*rJay8MFS+GlsfTERmWBH94F0qRRusHu+uoScsZg97/m0FB4xSMfNo/2rP76r*)
(*JZdQCuwndrh7IfesCn4Z+oues2LCLMgOrnX0oUulsItW/cQdxXPbpudN9T1l*)
(*8OtD+stxFP+w+9eMpc6VgfFVUvB5ZLFmKZMdNmUwxDf6MwrlK1OFwXT5fRlc*)
(*tu3mHkT5VEluM+ssKQNGg/ePhZBreL5aYMNlYISNjmugepCCH9mNWzGYk0m4*)
(*fQ/VS0FyaG6eNAaL7neJeH3R3fJUSFHFgPdcR6jzHAE7x2uVF3UVg9j+oAYz*)
(*VI9Nz04XeD7DILeTq5Eb1e/05alizToMJIL27WgcJWDyrG2nVAYwcISXkSYj*)
(*BOxJ6Y9S2WUMGqbcKvuHUL4lP2A79/yAy1PJzUX9BGwv7XRFr/UPaA7qG6zt*)
(*IGD3sqXONnr+gH3HJB7O/SNg6ff4qyre/4DXA8f1mNoI2Mmmqerkkh8w+Led*)
(*zNREwK4kfaix314OBjYXP9/+Q8CijJ5r3dlbDr3OCe07fxGwAW7rWoNj5TA6*)
(*XsFXXIn6o5tq/SmjchjZnvK8ByNgfpemG5m/lkNDxA2ay3cCRmL+d3kNKwfM*)
(*fdu1kmxU7yXlzVPN5bBnn4v5QgYBiz8Q1tqwUQ4/f0U8OpKK6mdDtSNS6yfE*)
(*x/nsYohB5zfr8I1XJj8hsdyyvyGSgF24K9Dl4fATRmoc6d+FofPfMN1tEfUT*)
(*PvZkioy/JWDziWH9MlM/QbCFI2DQi4Dt15sZ+xlQARZpFXJbLQiYoLu60t24*)
(*Ckj66vMxwISAcX2NDWDLqwB33z3dNNTv11guS+sNVEDbBDE8Hc0DEpZn33O8*)
(*EvVDhT1HAMVf1pNGGa4EWRW+gixetH/jjotR65XQbFoqv50T9e9AuTjYUQUu*)
(*xCALfTSfLo4Nn/FRroKRnSoRSVQ0/2Iv+PG8qwKvXblH7wwLYHWcvLzSxGrw*)
(*lbJrEssUwH6esrZouFIN5Pt1D0+nCGD5llXfHSyrYZAo8Px8vAAWX+1sWBRa*)
(*DVcp93N2vUfz2b07WmO6Gi71RW1UPhbA9s0mSppF/AL14F3cwicEsMd1JyBs*)
(*8TccFyDUnUnlxzikc4d/s/wBzzB57vOf+LGEINnAdaE/kEjnPHcinB9r0DrQ*)
(*bnz2D6j9/SLR48OPHarhtd8f+gc094moSN/mx3qrZxOzFGvAJdbtZscOfszp*)
(*gKX2kGYNzP2kGbxG9xX2lyPknSY14Mk0ESZN5cNOnOsmuvjWwB43ef9DY3zY*)
(*u4o/HdBeA1/TeH4IlfBh6j8+c9Q++QvtKrw/fpnyYT0ie3Npr/5C09QRS/5r*)
(*fJiDR9QN2fi/MJreM6OjyYfFwbukdzV/oZJf7ICzHB+2XuKhel2oFi6vZv7L*)
(*YODDvhXecBotqYWAfKGjX6J2Yvw5O3ro6euhw1wniLWUF7sZuK4/xFcP/UkY*)
(*jTWdF/tkPlxTJVUP3rL8lMmPvJjUzoKCAIN6aE1lE9R348VUnUze78yoh/Af*)
(*TjpZSryYQtS1xBAZEtyx3mB6+40Hu22caJTUQwLpwYE8FU9u7MWFXeHh4yTQ*)
(*kyN9O2PNjSUqvGn2XyRB5eLN8AOG3NgUxxNNa7YG6H9SyhEoy405lWmfOKrQ*)
(*ALu5tdWqundggSJLvMVBDVCsJCRBObID+z5w5m+9ciOk1GvGPHnOibVqeM4H*)
(*qTcC+xuh+e2XObHl9FI+rUuNwE2l3LgtyokpPj1x+9fdRkg8ZDGn8ZMDy+eR*)
(*WcaCG+FoZvz3PXQcWIGaoEjWWCNk8z49x2rDjhUnztmHhDRBq3/XlIcQK1b+*)
(*4OMuw6lmeJR6ZPxJBz22JH6K86t1GzwhPOy9MdNepjKgWu3v3AZvDp8Kb1lr*)
(*K/P5qOF+zws91yR4PWBpLePj1Z8VDW8DCd2m3zwHGssU6C3rIyraIOPBpQ3N*)
(*0Yoy586w1767/oFOqC35Q0YWbLxe5DT//Q8ix4lSw2V9oKa5Vn26+R9ImM09*)
(*HY/qh8CtdB7Cvf/ARo5mkuY8ALvdts11LP4DgY8qrIZSQ6DyYB9JT6QdomSk*)
(*tpZ6jYK7mu4beNwOp5JMuE1mpoFhNZVr1/4O2Dj+7G/c70XYVatKBpkO0Bqv*)
(*bKojLIF8XHuL+YkOiGVdnDp6fwkszjNHpmt3QFMjmRDJtAy/wk33qzp1wJZM*)
(*Zp9f8ivwWomgdL8afZ/oB+Uoy1VI4sgQetXQAfz7939IT18FbODcZlZnBxTt*)
(*vDIwMrcKC/4OFRuzHVDfeUWuym4NrnbW67wR6IRHGu4CUlbrIOTqY5Z3vxPu*)
(*/7Cp2HmMCsd0d5/rsu8Er28HDE49oIKOWI4E/bNOiPCOT3weTQX3v/3TF4I7*)
(*wfDlxRQLRhoMCJ1y7inshIE7SiS/ShqszzVd31LZCTfZo10jyTTgrXxwSqK+*)
(*E8Qn1oh/RDbhnHUYg91gJ4obycbdeROSS8iBTOxd8EnPwT6Rg45Y/tb/4UG+*)
(*LpAfMavdv4+O2GGx99JFkS541VHsLX+MjsjOcZE/XL4LLh2/W093nY5oa5wS*)
(*J2XUBdXv88evR9MR/Y6efqF3twuaDZhP3kinI37a+u+O06MuEJ7kap0qoSM2*)
(*pTNJYd5dkGLNIFPZSUdU2DT5fjkdWeiaoj43PVFLG3t+Nr8L9sWVyafsoSea*)
(*RQpfVCzvgs2fgrFhB+mJbxR7RgVbuyBH/+994ml6YqL3qRz2vi5YqLr0eU2T*)
(*nljcFOVBG+8CiQs7WS9epSeOP7xBGKB2wfb1eIlf9+mJdKWFw03M3ZCuebZr*)
(*2o6eyMcumFW5oxv25eV8+faUnqia9O9C8v5u4GZ6FcYaSE80WFbkj5TphvBX*)
(*M3Wf3tETH6p9GAw80Q0FDXMqHRH0xKi+K6622t3A7ObGIZ5ET8ySztUwvdYN*)
(*B9W8Q0+m0xN/ufLu1L/dDaRQjuT5bHpizx/7/rOW3aCsoN6vUUBPXBJo+qro*)
(*2A11J9i6T5fSE9nvHn0i6d4NcuKNyt3l9MR9uW/P7fLrBtXRlPhd1fTE41vm*)
(*ube/64bzGaJ51D/0RB093V5aVDdEVhCFX9TRE81jvqXOJXbDq/3WPt8a6IlP*)
(*pzmcBzK6ge2wVbhHMz0x+KSNWnNhN2zj02SmtNITk/xquaoqukHX9pInXzs9*)
(*sbRNqjuvrhsUvFcKuzroic37A5OT/6H4fKu9r9lFT5ywn3SMHOgGloTHZ826*)
(*6Yn05RdUg6a6oTc7jHl/Dz1RgCuFw325G4Q3Ky3fIksbs3ba0vWA8ZfHMsnI*)
(*amn3vpiy9YDsyuoOa+Tra9X2+rw9MPrSuLkbfZ+txgHiuT09MG9hfWIDve9l*)
(*qA/7cYkeGDoayvS7k54YPTT8T/JoD9Cd76epofXlHD37eZdyDzAM9E3Z/qMn*)
(*/vFIeLT9XA+oqn4K00b766vborJ5sQdsx6L/djbRE5eFzNjmDXvApPKrvhCK*)
(*z3bL8tYBsx7Y+myRyoniJ1awN77Zugce5/4LzUPxVWL2fFj1uAf2mM/P8KD4*)
(*34kHlpSAHqjaUyQyg/LlOh/dHPm+B55FbI+xLqQnvgNabFBMDxS8KNz/OZee*)
(*iHUWn3iU3QOxV2pjIJWe2CoptNWspAe6/YUg4TM9cerx00b96h50vl1VKmLo*)
(*iQReJcvjneh9w4olJ1C9yZqGKx4cRutXppt5g+rxXAZli9BsD1hjH/oTvemJ*)
(*dlp5UZsMvWDM9dmF+zE98e8L+bqqg70gVMVgRHcNrZ+zhuHtsV5gMO+P/KKN*)
(*8h1+W/EGEVl8vxybGj3RK/1V7OyVXviekFjHJIs+/2/MTsCzFy5HzNkuMaLP*)
(*H/rI/6ANfd8rHvnUeDoi33c5LfmBXkg6aysfEEpHfIEiTpvqBa3y1BVFPzpi*)
(*Rsq6w3n6PsiIArYNazpiQJlETzdfH+QKl25oov6hOv48Y6tqH9Tm5o+PPt4E*)
(*F+m6P5bqfVDnu3zopMEmZNgThklafUDeaTqlcXwT9tC+ESKv9cH9mMDiyBUa*)
(*rHJ3P5ex7gPe9JV4fXsaZCgrXjEI64NCMVOnPBMqCL+ZXE2e6YMpeo2dvsLr*)
(*UHrbPTmL3Ae2bFmRotQ1MJbjMSyi9EHXnyuXsjrWIKpVKf8vQz+Ypmyb+/p+*)
(*DQR2+zvN8vVDSr+Vx6lta8CdIkGWh37ge3M5On2JAkwVFlNlr/vB1mPeIqZu*)
(*GRLer0b+CukH/6z8yvC0ZVC7G6TZEN4PJ+8o23z2X4bnbLkpA/H9YMIn+ZB8*)
(*bhlouoz3mfL7IdNdUdMNWwJK96fhC339cN7I9XZr3iJMrvT2tMgOgHNL7La7*)
(*RQvw/e7r8gsKA1DLd2OUELEAHm0qiWUnB4Au3TVh0XkB+PI+WiefG4BXq802*)
(*oooLoOp0Y8P15gDsmmiQfZQ7DxGLrQQx3wEgmL9KIxTOgcZ8rb5d7wA0xHx9*)
(*X/BnGnhuux0fHRqAr6fCzp98Nw09JCmhmxMDEN5c9WrXzWlwyAgYPLs0AApx*)
(*vqqUmSmIeXjejrBtENoIrR/ECFOwPF3xGlMYhKNTu3fquE9AwkTRH45Xg/BK*)
(*lXeLfvAo2Hy4KPrg3SCQXFwSLO6OgqLa4NPKsEHwvTdmWa88Cn8+sh12jR8E*)
(*8f7fRiljIzCva/h6PH8Qhmbn1hzVRkAlb/lyxeAgqNG/GBZmHwYWc/804fFB*)
(*WE/Y6/lmZAgaufYwPZ0ZBCuLAlX9H0Ng8eDs9yOrg2C4/rSy6fEQBO4OEYjh*)
(*GAKz6C/3ksYGod3rSLfLiSFwfGxaodc+AJ9kK4+1qAzBJ+W7B6vzB8Cq2+CV*)
(*rNoQrGe/d/YLQ3FW9IBRnSHo9ru3l2wwgOZoXZy++RAQn8nV63T3g4Ou1R2Z*)
(*10PgbpetWo3qjGP3l9mhoSFwprTE6iT3APGqV/fg+BD4T2OVi649YPf6Vs3A*)
(*zBCcDDA5y6LXA630Al/6KEPA8pLdymGtG6JHfI270D454nczaF/qBumMB7VN*)
(*8sPAXCFsYLSzC0zGzxU1nhiG0HTG889nOyF4n2hyg8ow+F7Q2sb6pxOWQzpf*)
(*1GsMwxPJdakTHp1Q6qKlXHNzGHRcHghxzXeA9hnp1PIXwxBhd4lpY7Adnrmy*)
(*hf/wG4Yjl5Pbxn+2Q0buiA/2ahjEzigbSie0A69EjGlp2DCs8cXm5lu0Qxc7*)
(*l2Bh2jA8S15mwab+gVXrvG9m8zBoVHmxMbH9g2jOOqeM9mFIImjycsy2AUkj*)
(*xfxbD9qP7x4ts+Y2kC8yJX4dGwajtOiy2Zg22IhpWknaGIaQVnsFd6U2CLyf*)
(*eydWbAQE5v13zLq2wgTxq++2gyPw4Mjl+kDzVlAX+JziJDMCn5WWkp21WoGh*)
(*OmRGS2kE5AJtfijubgUXMQen1YsjEB4evB3KW+BO71GfS09HoFD80ddGQgtU*)
(*fD+YVOIxAslGajs2mFpg36t9fyR8RmDb7wLNewvN0K3MzbH5ZgQkTEU+bv3b*)
(*DPoR8+9TE0cgZiffbY3nzXD6SsZnxsYRUOQsr9dba4JoqaTqh20jsD/s1OMb*)
(*E02wsSV2vKNrBNISn/d/7GiC/KzXhzNHRyB13PjTn6ImkOZ6mGtEHYGN+p43*)
(*2R5NEDh6p+0XwyicFedisLdrgolS41U5llH4p/4g8o55E3y21jnFxjMKB9Ne*)
(*coxpNIFQzeHK7xKj4DmulKrG2wRPPu0f3SuNzpVYCVsocxO0uexmDZIbhUhK*)
(*kAr/eiMES2zXNlMZBeY1/+1fBhqB1WeqmVN/FPX7sKnp7EZYJqYO3XNH55C1*)
(*8mCycSMUJ0xGPn4xCvkf9LzMrjTCcxapyz5+o+Cwrz+TqNUIHKS0H/HvRkGY*)
(*QZTHVKkRxG+lR/d8GQWO5yYfFQQaYap85sp02iisinEaUDkbIUtcZvtG5ig8*)
(*FI7bHGRuBJWZb08JxaMgqn0pnZPSAFfdMg30SaMw4TLU09reAEL985xmLaPw*)
(*dKRY911jAwyoHa1+1DEKx+1EyVY1DWDDnn3s9dAopIdccXYqaQCfyBzuP5RR*)
(*GKOPs23/1AC5BXl/Ye8YTHAfiAuza4AnuyleOuJjwMg7M09v3QBEz+MnjQ6N*)
(*QdTLARevuw3wV6Mg+cmxMYjtVWWuu9kAQ22FPrnnx+CzCWWvi0YDpJxcP1Wh*)
(*MwZBZf1Hos40gG3MyaXGy2MwWPr3eItKA2zcKTabMxoDBy8Bft9jDcC7XEI8*)
(*ZDcGzRWUxHHRBugwpK2ceDwGKloHNnKEGyC2ROWbhusYHOmLkw7Z1QBS3mW7*)
(*7/iMwf5eTDKIpwHUeH+sxUaMwfOukkBTxgZwOFqRw/dzDIZoe59boN+Vd+RF*)
(*Yt/+GgPvr4m+XCMkMFBwDWCvG4O0lys2dQMkOKkkb8rQPgaXuRowzy4SMKgm*)
(*cM7MjAFXhMBkKYkEi2fo1u8tjoHXnWc3V2tJMHz25sjg6hgcaDoWqFZDgl/n*)
(*eUvaGcehb9ZSlL2KBK/1XjyoFByHA1jkacNiEnhc7rtCFBkHe0wlk7uQBHZX*)
(*lE8X7R8H5tzyiY48Elw1XOTPlEXPA7bXhWaTQOi2eWXUuXHQj/u5kZtKAg4z*)
(*LINfaxz2rJGMepNJQGchFBWsNw4SUQOlO5NIMHiv2c735jg8ZDuelJRAghRb*)
(*NREHu3Gw3MGpdTuaBFF2sdtmH4/Dg3cnuqeiSBDksLF8320cClOe/fWKJIGt*)
(*c07tLd9xCNqTtNkYRgLTJ1wFHUHjgBmLDvh/IIG+q1XClXfjMHN46JZ2KAkU*)
(*PcSeakaPw6Tue9v5dySQeO5xpyp+HGIPhCY2BpNA8EWX3unkcej6uVZS8pYE*)
(*tJchEoo541D3YUt3xmsSVLxmbN33Zxz413dzbA0kwfe3Jj8+1o9DVcATtyMB*)
(*JEh6V5wm0DIOU1cm7C38SRDwwcGLo28crms8vjTiSwLXcJKN3/A4HK5jmjiK*)
(*bBMpdZ1xchxKPry8+/IlCS7FDMmuLY2DqLW3s7oPCdTiiEKO6+MQPHayLMub*)
(*BMfio5jn6CbAevLapf3IAl/0e4a3TQDR8cRv4RckYEvO+G2yYwKkq/+ZffEi*)
(*wUYKe24n3wS6r7/KP4bcm14R0LB3AixtSxnuPCcBaZpx6uiBCWi+OniDFRmT*)
(*OqsVIjUBP2lKqVmeJMiw9E5bOjIBt5qqSLeRY1Mq2a8pToDnn/E0PuQ340zW*)
(*+coTcMXwrkiDBwk8Jc7VElQnoM7KlP0tst1dn8NP1Seg0eLmpavIpolVQV1a*)
(*aL0E+WYR5EvDW2dOXULvaxJzmHMnwRkxdZ2YaxMQd7uStxJZ3uxlOp3RBNgL*)
(*ModHI4t9quYwNZ2Agjn5IVdk3n7mhz/vToCX90L7LWQmEY16MesJYCeeNFZH*)
(*Xjb2lfGxm4CL929fkUMe+fjr9ejjCYjV3J4gitzaxTKn4TYBZ64qywggV+86*)
(*r5vyfAI21IhdXMj51/0ytvlOgL9m6ZvtyEnhv7msgyZA4Gz6UQ7ksH+sj+qC*)
(*JyBq1DCJG9mP/0KDTNgEbFu/8U8Q+clV/yNvP6J8hcQkiyM/eP/n7cKnCcjS*)
(*v7VFAfl6M9vC5SQUv1SxWg1kTR7NS7lfJyBcqHrGGFn5UkAWX/YE2AVY6joj*)
(*S72t4XbOn4CtL706QpCFSNvs20vQPSzyr2MOMjunVpPSzwmQ/XJxWxvydODf*)
(*d9TaCXCpiPwjhuLdU8O+aNw0ASd1x6svItexaetj/9D+fJzvuCGn+9byeg1O*)
(*QP708yN9yNHV2x2Hxibg8/RPuZ0o36+36rScnUHr3d7wXAvZ9kXde5bVCdjv*)
(*2pv7A9nkJ8fyfRrKh1HLo01kPYaLV2u2TEKBjftBQPV11KOe7xXHJBx6WKlW*)
(*gbz4lPSBR2wSQuKfcw6i+hwq5KI4SE6CWFZinDSq55ZVXYNW6Ukw+dIb9RT5*)
(*++MGgfATk+B07rUfAdW/s31j+J6Lk/B0MPKOGTov6w+aIyWfTMKMiafOXj90*)
(*/hWM3kq7T0KqZXCFI3Ia/YiP3ItJ8IS3H/8gG4etPDr1ahJ+i97WckTn9Uel*)
(*oIbep0l4Ul5jV4vO80uR24vOfyZBWPgzn8crEvRNjo+71U/Cs8KKiRbk43l2*)
(*vc+bJ6E4PebFIdQfJrRf/AnsmQRLp/jRFmTtp19iYxcmQXGoxVsc9ROe1mmt*)
(*X4JTsLtgq10G6j9WcU6qtcJTgJ1mWaIhV1htKjaKTUHYqegqrRASOG3ZIdol*)
(*PQXZiY7KI8jtsvKrs6pT8DDK4g8n6mdHN4pnFtWnwE4mV8AA2b/67NCq1hR0*)
(*mmX7xCArG1+r33JtCn5mhaYdQv0wJuDJZ37LKRDc6xh8HPVLytUtkUK2UyCn*)
(*03vRBVl3X+CbvY7o/UlJTAXIDAUfnx5yn4LbbHUax8JJYDGC6cG7Kbg7/DJf*)
(*JIIEpZnn1dXCpuDQBc2fBsh8bo3K5z9OQV8eV+Jr5F88Qwcuf0Gfv2cyvIp8*)
(*6DQz7U7RFNS1ZE6Vof6dkZJ64B82BdNarkcXkOV5dfXOV03B9RI76j7U75VH*)
(*wxIONUzBntwwXg9kTPdUXVTrFGSlzBmmIZ8t7F/Z3jUFhM8VfG3IOkEHNedG*)
(*puDOefeJAx9J0Lhc53B7agpiBd3XtZGvmdhHN85PgZ6LlLMdsolc8XzOxhQ8*)
(*qKYlf0ceijLZJc4wDcWKe6+3It/fynT2A/M0NLtOWi4i27Vrh7lwT4PqfHTq*)
(*ITSfllUXfkzwT4POZeMWNeSnaaGTN3ZPQwnxkcFNZDq+kztr903D7K8tcnbI*)
(*L9x7VVQkpkHs1g0zH2SWca973w5PwwHf1tEw5KBLEsEictPg5DdVmozMXfy3*)
(*6O3xaWghHB7NRw7d/2iYQQW9j4d4qwpZ8PVOTocz0/BSfHh/I3IMpeD4sMY0*)
(*uLwmn+hCFjM1Nr2qMw2+NYYRQ8hJNQyB1ZenYVh397lJZKljX3KPG06Dovp1*)
(*5TnkjGjN3mTjafA5fcaVjCzPMseyy3wa7FZFmJaQCx6FHA28Pw1e2y61LiKr*)
(*dB6/SbWZhpycC7MLyOVq3d42Duh35tuQi7PI59I9v/W6TEPAjbrlceQ//OLt*)
(*uu7T8Ck9dWgAWcfzD0P5i2k4X7Uo0IHcNGEjJeePnqv6vKtHNtDnuZrwehry*)
(*J3X1fyJ3leS573w/DRxOhjdykU0O3Ez2iZiGbRr5SZ+Rh9/QNa3ETMNJYrhy*)
(*CPKDtYSNe5+n4azrSX5P5Bmz8+IdKdMQJTJ/3ArZrnb6ombGNDSKCH+6gryi*)
(*EOxSnDsN4SynLp/C88fW+TcaQ/F6+Pw1M54/e/dlziq0X/7FXROoHli7RUU8*)
(*a6bhWND4wh9k7gwre7PWaWj9+t3rJXIoYcfH5s5pmB81BzPkXV65VWf7p4H2*)
(*+cz5U8hiV2kEiSmUP/1VTfx+k1T26Uz4/DRcLpFS/4F8WFLdmm1lGq7Ln38f*)
(*gqyw8Rqbop+B+8uJMgrI6p/23s3knwGlePCxwM9PGId4ldAMHLi8ne0Issir*)
(*9aGOvTPQmPxs3zo6f2suLaaMUjPAt+sYfwDyVz0/42vEGRjjjTsZgc53hLrT*)
(*bquzM/A0ceb1dWSfU2ZdHhdmIHH7D28Csonkqesp+jOg1idq9A71Bx76+SvU*)
(*ezNQz6Yy64T6C91KNy+3zQwwBf72kkKemvrTJG4/A6/MJX/3oX5U9S9BT9dt*)
(*Buazc23OIrtkGGjHv52BKg6KySbqb723MLXzhTNwOW5LpyK6j/298nWLcdkM*)
(*3DFT92lH/TNfM6LcrmIGAvPZmJ8gBys6EKPqZkAnV+xd3hvUDzgllGcHZsDA*)
(*0Gr3ftR/U0tfy73fNgvjjsMaf1E/f7z71r7+m7NQ/yWXaIzmiamu7NWu27OQ*)
(*tjl4cQDNHx0vev+2O7MQWrAZYoEsPh4/X2s7C6NfKDX30bxqzhnDCl/MQu/+*)
(*nhP30XyU1bK7FZI2C7xnjd3OuZJg7Il3lPrGLNCuzv8Kskfna+rKtjbtOTCP*)
(*rh8bNEHx4pxjqVibg4zbldY/FFF/KvF/eZ46B0Ya93tkkIOt9m+t35wDZZp3*)
(*drQCqqc/17d0MM7DmOORt27HkL0rqbMc86BfFnuGKEcCrvWIhV2i8xDih/3o*)
(*kEbzf/Rsl73mPNjVtB20Fkfxe993fVV7HgRuzldN7CdBoNrT9me688Bemj57*)
(*D3nbp4xWvyvzILSyMnFHDN0fjXY1xNyaB6dgpvj7+1B/apqrrLGfh4Nqo13+*)
(*e0jAWBb5TTRqHvwTC/JO85OgQ1lNozd6HnI21g//5iPBt8Kpvoi4eXBw9jXX*)
(*Q77+/RQ395d50G6o5TXfSYLMr70OdFnz8LWc5W0wD5q/UaJK3dXzoPfr3S1e*)
(*LlS/gn8bw/7Mg6gr+84kTnS+whws9WvnYR9b6Vtl5Jx3lVE1jfNgptTTf5cD*)
(*rT/g7mZB9zw87g0ZrGZH9eGSWhFKnoc9vPrBX1lJ8GrlstGl5XlYOqDaqI1s*)
(*5rixtH11HtZnMjtnWNDvk0faB7xp8+AYzasph2xxb9bPnnUByt0oEZVbUbyv*)
(*yV3UFV4AwVLKBCcjCSzli9pZNReg7zGlxp5WDxipJS1fewGa0nrSxqj1sNN6*)
(*zv2u7gJYHTZkvIVc9nm/eOWVBaD4xfHobtQDD/8bOw+TBfg1ox9xZq0eilbN*)
(*t604LgCXl4LVw+V64Ap17/nsvAD3mUXU1pfqweJoRKb+0wXUdy2e+yJzWtZf*)
(*y/JYgJKA44e/LNaDWdeJBOuABfgofoWwtFAP28o4Tg3FLoB6zW7xxtl6MLkh*)
(*yfUufgEy86p+2yPnrpwZPJ24ACtuNqV8yLdkXfxiUhfA6GBViMlMPWTHDbZc*)
(*/74AKf8Ya7ZM1wOLCi2JtWABXhw68vfbVD0YdQi45hctwKGP7NxGyMw8Ovv4*)
(*fizAP/fbFkWT9XDjRb5NQ80CCMl/3/CZqIcMkabTHnULcPR40pIyMlPJNK9M*)
(*wwI4cPOIL47Xw7elfUWBrQtwld9V/A7ylrtBzBr9C8AsJ+hoMFYPqedNY0uX*)
(*F2DJmLHEYaQe5KX2qsAqin9H/bbTyKUcfZ1l6wsQfWpAlwOZ1GTE/4OODGwL*)
(*bXmpw/WwaGT46icbGUL8k+vmh+rBjSggpbadDN5KBp8rkbeKtv2u4CRDcaHE*)
(*nQhkwpg+UxUvGTJ/s8aeQ1ax03X9tYcMD45E/k4erIdqfU7B83vJIDFzuOoF*)
(*sq5iXd5vUTIsSXjHmiCbbmiS/0iQYd5WfUwQ+aWP+oPao2T4EmkXFj6A8nt/*)
(*K4vOMTIkK7y49hQ5QrPyc50iGW586Bk1Qk7jOtNfr0yGzZVDTvuRSREqBo3n*)
(*yPBs783dpf31YOhGXbp0ngzObaNKicgDt4rfNWmSwef9id2vkRfFlOqbdcnw*)
(*WmNO0gzZjXnV6splMqQ6KejpIG+dyGNrvUKGc0eYFJWQCd+OnWu7TgZNS4bj*)
(*vMif3i4OXjMiw6kNXl0GZCmHbM9/t8hgoO0sMt+H4nHiSEm7ORkG7WiDJOTq*)
(*XXPXr98lw5wSa/MPZF1aOqXjPhnsC+SdspFNf0rJd9mQ4deRij9hyJOfJxtu*)
(*PkL7WeB9Eojs4JvysNueDOxMJi0eyLQH97cbO5FB0Gd3pyPyS22J1B5nMljq*)
(*VPlZInPJjmrceoo+v9rUexs5gjtxpNeNDFfHsU4D5LQ20X39z8nQJC9epoF8*)
(*rHCg7LY3GcQdPsSfRi6NijMaeEmGi3/bD5xEVnc3WTf1J0Nbj+WFY8ik28Lh*)
(*g4FkMFKoZZVFNlTrUTB/TUa//56ZH0IeEP/YPPSWDD1+dIYHkK1Yb9pZhKD8*)
(*6bcOiyIvTQpyjYSSgZfNi3UvcsjH0eqocJSfA4Jhe5DlLua4X44ig2GsQ7MQ*)
(*ciOdpyJbDBke7cws3oX8KEt7FosjQ0HUrA5uLnPBL48TyMBUreeH+9vOUWPp*)
(*L2i9n9fu4/+vU53NN5xMhlkSx+xu5Glnj7rINDKUW+bwiSAHHtT2ufSNDNhB*)
(*nt59yIe6CCqsWWTwMNA/J478J2hkqSwH5afj04WDyPch+6tTHhkcFQ9NSCOz*)
(*zLtbHC5E9RG+S1Qe+csnrd1Dxah+m+PJJ5DP6RNaIsrIsNtqyZCIPMw0EqhX*)
(*ToapkDP66sgv8rLUWCrJ4H7yZ58Osuh9943SajIUMqdSryLf/itgJVVLhkr7*)
(*0P57yHTPhkUH68lAJ8f12Q45RiarM7yRDDTjnnlX5O5gTU3mf2SY5HYWeYfs*)
(*qiawpbSDDE+Vf9GikXctDxU6dKPP73a6lYrnz/DZwYEBMoQ3ssRVIrcIZ7KW*)
(*TKN6UTu7tInXX4PbD/s5MhDuZ37nRPXO43XB5SCZDEpdY1PCyHojg2MfKGR4*)
(*/3tn5mnk2jS+X3ZbFiFu0+iBP7KV8aC75NZFqF1hT/iIvI0rQ7GPZRGmwh3O*)
(*ZiKftz//RZtjEcTF2AvakStPuPpIEBZByF9CUQadb/NJDZXeXYsgeuX687PI*)
(*Wz7uXH6/ZxF4EgokbiKfpvtmsUVsETDRYQ9/5JKqfrUe6UVg4/0qP4mce1l9*)
(*S4jaIuTu8rlVhPqNr2a+t5D6IhSMyP1rRr55RpLl8/lFWC1jC5xBZpDbtv27*)
(*ziJs0nJj96H+psNdz/fPYBHir1ONg5DHSFcPClkvAlfUo1NWqH8K6tzRS3i/*)
(*CPqVe69rjNaDu4Z3TM7IIpy/kmqjjvp9o/G9m7Lji6Bysd3uLrKYoyYhbXIR*)
(*kucLXF4i/4nb8S5+bhGM1eSDqpF51z/6BK8tQoNcU6E6mh9JX79b23IsAWOV*)
(*YY46mjck7rGTUgpLQFXd8VYVzad9kjWUL8eXYHlWeeg2siOk54qeXAJT9cMH*)
(*PZEFrRxkdhGX0P3mR2AZsnklnSjbhSVg/vIr7uRcPVAeE9jGjJZgT/+xfNn5*)
(*etjbfeFfvPcSiKpRgZ1cD/7lJ79d8V0CF9+S7EPIC1+kfJgDlgCOxW7TRC63*)
(*45C3fLMEAX2+d/yQzVkaXx+NXIKfOSYYI5q/SXKG6j8yluB1cV3FErK0/93v*)
(*PZ1LsBDkaVKG5nnoQ4OgNz1LUPd9R1sH8qb+eXPV/iWQ5DA6uIxMEj6EOtgS*)
(*SIS/vSO1gs5/7oy19fwSPP6dduQDck6f4/6NrcvgnXSfYkGpByXFFyGEo8tA*)
(*yz6QObpaD62mBIWP8stwjSd6KxXZ/lV6m4jiMsSPnjzJje4XacP/BCWUl+Fo*)
(*wtBpZeQ9IYfjFM4tg4xZXcUrZIb51q+Xry+Dr7XCp8Pr9RAtZHWx7eYy2N82*)
(*mSAin9Sgn79+axnKLFt26CM7xBySNzVfBuKWUpoL8oi2R6GtzTLcO2WvV4Fc*)
(*kyJZHfR8GfgkjQcvofvOvdbSezu8l2GLQKaYOTIjgz7b+5fLkL2VetYRWcXw*)
(*mXZU4DLMby2S+ICcwdzUlBK6DJ2JXrbtyCHmrn3VKctgaiMiYIDuUzf31K/S*)
(*Ny2DRWP2RRV0/+Ig2OUWtSzDDkeFRi3kHzw7Hzn+W4ZWHsajN5DFWW+MjXWj*)
(*/fFS/R4jz5JHWuvHlqFLoy86Hfn5b1r2R9oy8O/Y78K3ie4XFXEPr9GvwGxa*)
(*YO8+5NFStUM7GFfgpuql/TLIWjn+n16wrsBiyYuL6sg7Y/jeWvKugHWd7T4n*)
(*5CQHaZsTB1dAY8R4oBb5+sMGSbLUCmAany/+Q97+wGE4TWYF1vxDoweQ7W4V*)
(*3hQ5tgKxKjwDy8gnL5zTZCauQKbfP3pOOhLUCRtLtFxZgVttE/flkD0E6Yde*)
(*GayAQci2PEVkuZ0JMRo3VuCB+Knlk8hhbBN8xSYrwFAjqqmGbLrkyBRvuQKs*)
(*Hcf19JGX/gQN2HquwNcZ+75HyF3bWu5teK3A4W+k3Y7I5VpCsy99VuBIuJiu*)
(*M/KbupT16IAVOJW99O4Z8sHGat7a9yvAqygQ4Y/MxcMZaRC2AgE3WT2DkFcu*)
(*X907FLECjC9djN4gV7QMHV6PWYHhVcGV98jG7fTqkikovlUq32KR1QTP1+ak*)
(*rcDCE0a1eORDN95cJn5bAZfzDXWfkSlde0yu5aD9HKioSkEO7jvp4l26Ap9a*)
(*xk/nIFeNOqb2N67AOJVmX4mcJlFyxLplBYJjxR5VI7+7z5hPaVsBFslbFr+R*)
(*TSaDKzi7V+CH1xW5WuS1mW/dp0ZXwDud/24zcp/Mitnv8RWYYsjd04pcbasy*)
(*oT+1AmEvJ2vbkEMW/i5bzq9A4WsLnk5kmeVxzoh1FN/TV5T7kfkUj4Tup61A*)
(*3pYXWQPIG4+dhTLpKFCRVSsyhPxndavkLyYKeLbMD48gm1NFVZc5KWB9V4l3*)
(*CvmCiuUvT24KqO+iaU0jH3HP0tm+kwLfrW48m0Gm0Z2+ISZIgYYze2vm8Hwz*)
(*GjtcEqPAklwAxxL+3Iz/WIM4BVRKulmW8e//SVq6KEmBuXi/Ddyyz8881pam*)
(*gGS0OImCHNq/ofhXlgL5BPmMVXy9xO+UC3IUCClv8ltD/rUp8UTjOAVKm+gP*)
(*biBL3xpQqlaiQKN82hzukNLI9bOnKBCbI5JJxePrxuF2RhXt906M8CZeD2tk*)
(*dxUtCvQ22+fQ05Pg5vWvxBIdChjVzV5gQC4vuEOvrEcBS5H4TtyvnNufn7hK*)
(*geLrw5NbkBfbgs/kG1Dg3PArS0bk64pajIo3KDDC/nkYt/hyqbe8CQUet2v+*)
(*ZkIOvOJ8LtuUAhZfl49uRV7IOcJ81IICD1xVP+A24J2szrhLAQYl5WXcpfYJ*)
(*vjIPKKBZM3eRGVmsyeh8uhUF6JitE3D7H+VnO/wQ5Su3hIz76rxfwEEHCmSq*)
(*cXqxIBfrntFKdqKAi43kT9z7MjbYJVxQ/j5eoOH25fxem/iUAvN67vKsyDM2*)
(*D1/tf0YB9s0mC9z6dRIXEzwoQGa5Eoy78PAAp6gXBeoqOQpxiwRFkuK8KfDO*)
(*bXs3bu8p/bcivhTgfmC0jntCk+NSjD8FVr6z7WRD1k2t5t4TRIF6z70HcX9n*)
(*82yKeo0+3/FVCbfQA6WQXcEU2P+j6Bzu57/J+hEhFLh18JoO7jGJrzsJHyiw*)
(*Yz1AD7eO753WD+EUCGS7o4s7Z1T4A18UBYT3zV/ALajefu19NAUOUI+fxu2R*)
(*GCzAG0cBlvPn5XCPMGm1B8dT4MTnw3txa1kwRexIpMDZ+Hk23FkVpdffJKH9*)
(*VsXO4fsREHPexZlKgWDvc424n3kd6Qr6SgEr9ZlvuIcGJqLYMyhgeCfeD/cF*)
(*1QSjgCwKOEk/uoU7I85oD1suBbb2mB7BzUfP3+ubRwHbz16beD5cTUgxzIUU*)
(*OPOl5zfugTI/E59iCshRnr3BrSF8Zi9TGQV2xTy8jJunJ/cTQwUFaBXWdXj+*)
(*X7G7ueVVUcDUSM8DN9tJNQOr3xTA6HikcTOENW5vrUPxl7/rhtfXs6rwMf8G*)
(*CoSv8wrjXls0+QnNFHh5oq0Yr9cFvVnn5HYKbGTVTeH1bOPx/bJxF1qf/x53*)
(*3BPpbtI8vRQQ48rfjrt/G/uQ6xAFBv02hPHzQaqUuKg7i9YTd0wQP19ai7OS*)
(*TAsU2NJ/6g0d8q99eYyFixRI/rDMgLvM/Wyh6BrqJwapPTR0PtNPmO5fYVyF*)
(*UoOntvh5Dvwauf5RcBWecJ1iWUFm6TJtvbR7Fe4bUHXw/vGC7WAms8gqiBRJ*)
(*vcP7zdO7+Xce7V+F8Uwm3kXkB3tbGtRkV0F7aysF708a7zmSJs+uwlUBQ4MJ*)
(*ZCY3T/0Tj1bhVu1gLt4/c+anVsvsV4G0TqzvwPvVHYOYc06rUP3y8Wg7Pj8u*)
(*ykxcfroKV7BA3n/4+/d1u9t4rwL/SwFTvF8z/jqeGh++CsREP9JfZAaeOXqO*)
(*8lVYO2wmXoJMl2z8bYB7Dd5n5Q9F4fNiqs0mZucaHHQf/xuJ71dWT/qmwBp4*)
(*O+vkRiDz55/52rJ7DeSHqn3C8H5XLZH6S2INcvauSYUg3xxeSExXWYN5fSX/*)
(*AOSCvS+jnzxYA4WDVxKeID+zoDNWtF4Dwv5jkS74fEt22b34cA1ebq0Oxucp*)
(*SdYqytpxDb63sr9wQh4DvQgTjzXQc3R4aIfPD+NdoedC0XqsD1o9QLaL+BbE*)
(*Xb4GQzLuHDeQ8zYiXpZXrAHbEbb91/F+bezjaVe9Btt0/ZUNkV+KGjk2/l0D*)
(*4yO2NteQo9LYjILb1uCz5Ub3ZeTKUgsp7uk10L2qPq2FLDC4q2aHwDooUwyb*)
(*AdnoLHPlD8F18Hcb24b705eF0ke71yHFK0hNBVnK6ndWw751sJXcXaCMfHrJ*)
(*KeLt4XW4tXHr2wk8/8yN93ecWYfHKtUk/L5ScsiXZYfNOowQXMckkBWa+WuD*)
(*bddhrcRBB3eG65e3vPbrsBPMcw/g76utEhRwXgezcwd8xJF9HzIdFn6+Dn+z*)
(*j8uJIV/Kfq53OHQdklkMi4SRa25w830LW4e8mi+yuNUYP3XIRq6DWMF44h5k*)
(*xSs/zORj12G27GzobmShlU3HkynrsBKj77ELeVTJLeJ86TpISDt58uPza5D9*)
(*Vg22Dqvmf5hwdwREiWr/XAc7rW0BfMh/u4rSdH+tQ/dxvbCdyFnP1kqvNa5D*)
(*5LXeEh5k1/LHgxYj63CiI0CeC3nHBTup5xwbYPN7roYVOfYH5zalHRvQeOmn*)
(*OW7ZE1/H53k2YGo+hMaCrCMxlnibsAEtqUeO4fbfarz3tNgG6B5USN+K1/vP*)
(*83x0ShvwXiKobgt+v1EaXcxT3oDb1p1OuPdlvWh6CBsQUCwggpsYV/amV20D*)
(*2ORvOjIgu7kf21Z2cQOyT48dpEdePrmX7pnFBkiYnHxORfdP7+zSnmP3NuDX*)
(*/azZDWTeQzdLph9sQMzmoDFuOcEPT4xsN6D8vbXqOrIthX1J+ekGtN9r372K*)
(*PJ5DGV9/swHNElvFlpCdpUJ/Zb/bALkP29MWkVkS5L5Yhm5A6WzyMdzi72zM*)
(*OyM3wK/ugzYZ2fTRUE9h4gYYHWJ6M4/cdZjU5FKMvt+K3WgGOWZyh9uWsg1w*)
(*C323Bbdp8mXxoB8bIHPleuo0/n6xNufYqg2o/RFNm0Je3tWzu5q0AZGFf7Mn*)
(*kQvbhat1mzagYZpgjtvtw23bjpYNKBL+sBM3I89w+XTHBro/EVwnkHewTd3l*)
(*Hd4A+MZ6bRy5ufrwjujRDfiTEMuFO8z7YeGBiQ14LGZQM4a8h57MfnJ2A5R3*)
(*7lXDLUVZzby9ugFnpYdVR5HPjTBvpG+nwvqnQPthZNaE85+Pc1Hh8Hd1Fdx/*)
(*bwfolHNToZSRiQ335W6OuBZ+Kgz3P0sYQjZp3nluYy8V5PNPDg8iiwVfm/UW*)
(*o8LRouY83KMXw8M4D1ChT/Z2AG6bGqHJfVJUeCGmfQz303LRN+cVqFCvrBKC*)
(*//5Q8bA40XScCn7dYIObQeXLwM2TVPC3ET+P27fg4DFbIhV+H4ukxx2acaQj*)
(*9AIVRO7Pu/YjG9rYe4loU2Hh3ZgRbiGpXKmUi1RQdPoOuOO/HHcv0aeC8I16*)
(*JtwZ0SA2ZEwFDqwisg+Z/ITMfPs2FawkCr1wK1z7MtltRoWBYFsr3EWcnNlt*)
(*96iQsUgg4qZNlofqW1JBIYv1EG7iL6cnDdYoHrRsPtxVHj2na+yogHUcmevF*)
(*680oeL+GIxValRh6cGueOMda+ZgKTr98/uImzX8llbpSwU0gMRU3T93tnJPu*)
(*VPT7+lQU7qspO8PyPalAlPEKwt1p6nYry4cK317ssMO9B46ckfGjwp/T1yxw*)
(*m+waFk8LoIJPgpoh7viVMDbJV+j79/7Txj3SpDXz+Q0V3khzncEtkUHXuO8d*)
(*FcZVJ4/jtgzMyY15j+qj9a4M7jk1IbfwCCrE3YE9uOX2kkz4PlIhzfADH24n*)
(*qpfauxgqmBX5c+LOb1eU4PpEhQ42AVbc67mT24ISqOD+69QW3CrBMbOsX1A+*)
(*nNaoPcgeNpebfJLR+0O1VnH/vMCctyWNCjbfji7h3nqgKMIjnQpJZ1Pnce8I*)
(*85d0yKfChqzkOG4h1uv5d8upMHfh+v/+9OOH++1qKkRqhU/hPvBEUv3GXypI*)
(*1w38//8PrnYf/FdHBaN19f+/P+3oW44rDVSQDTn0//tnOc4uNDRR4dCxwP/X*)
(*d3SS0qLTiurD9R4Dvn7H6rSCmn9U+Mpaxvz/fuNNPmp0UkHr2avtuNfceT0r*)
(*u6mg7tbCg/vUzV/mqn1UOJcaLIjb47irRtkAFcLSK/biLueVlVIepoKSzANJ*)
(*3Bq1H8jHJqjQqMumhNs/WbMtawo9v8Dwf75qvTcLZWap0P3onhbuSyp3n0su*)
(*ovVwPjHBbfRNYcduGhX0VF/9Xy/fr1M4SHQ0EGsYiMDNsbWQ3WsLDR43pyXh*)
(*xm6eYhlnoQFjU0Ql7n1saps5PDT46pjFiNfzk1wm6l0+Gmwz8tmJu8mkek2Q*)
(*QIO3Mj/Fcb/Iu7DssYcGdM8fXcA9Yq43rSVJg9E+p1DcKlw8k5uHaLDEcy8F*)
(*94ei5rEsaRq4aH8vxa3BbTAkIE8Ddh7zMdwpZcadQyo0yJbv/v/82gha/XbV*)
(*p0FKJRcNd1Xl4WqZazSYPnyIH+8Hwo9mKwYMaRBWHSiLm1Rth2ncooG5g4g5*)
(*7qOOLnm8D2iw4XOnDvdKvc/nNHca2N+5mY73Iw/vOM+uFBqodQy/w/vd9sfM*)
(*F25/pUGstGsm7oh71twj32jwm/1IPe5szRPxszk0ICqQ2Ubw/e5oqGAoo8FW*)
(*9Z9euDWj6Vgkm2hwcHzUBe+3fHm3Xjmu00C0nWiP9+uvY7sjObQ2offvohM+*)
(*L3w/epv06GzC5ak3EbjNL03vT9fbBD3mfaW4hYqLM7SvbYJ6pQgTPm/8X9+o*)
(*CjTdhJVx92Dc9xTC59lcNuF6nXrKLJ6/F7wazImbcFzv2R98ftGOu25vS9qE*)
(*h5Azibt9erAxMXUTnkXXsy8gv7mWbXQucxPYlHN1cG8evGTvXbwJBNmjDbi7*)
(*Gl5Hb2nahN3GKSR8Pub7rJg1tWxC8S3OOdwhJ29Jxv/bhJ2/9Dnxear5WTpH*)
(*tWcThMrdtXEXOtf+9hzfhA+mF3/hDj187LXe1CYYhWaM4H40EKW/d3YTZE/P*)
(*MeLzWkLLqhdb3AReYbbTuBnpmxPerGwC/+igMe6+3JMPTNY2YfNipCvu4gfx*)
(*MrLUTQg/IBuB+4PwtqXNzU3YUP6Uh/s/U5lIZg==*)
(*"]]}, "Charting`Private`Tag$393242#2"], Annotation[{RGBColor[0, 0, 1], Thickness[Large], Opacity[1.], Line[CompressedData["*)
(*1:eJwd13c8ld8fAHBku9nJbKroK4SGis9NQlYlSRkhUhnJihApQqhUMlORlZCd*)
(*dWVnXXvvde11ca/9O8/vr/t6v57nPs85n/M5n895Dpo91rGgo6GhOUBPQ4P9*)
(*qltMNJVMWiiqclgdoqFhxdNfErQbEFAGN5rMgQEmVrxsTFhUpoAujM5cyszm*)
(*YMVr1HdkRAuYg0zrHOOl/az4W+aUohABR2CZaSWNKrDi+6ReXwoSeAV1wqb/*)
(*1FxZ8V/3XPUNEPgI7R/8BfwmWPFu9Z/L9fI/wtLI8v2j86x4PZ9BusMGn0DE*)
(*NfB7/gorHrfyxLMgKhTqPe6NltOx4V1aP7rM7A8HQhC3j4kwG/76h+5H2ke/*)
(*gINbwnyGFhv+BXPmMX+BOOhu6/fNSGTDH1d975dhFwcSvz0rPVLZ8M0+jyd7*)
(*quKg/2DR0IUsNrzoLomfks4/IP99wbdQAhu+citOsrU5HkKTRZnz2tjwOHLo*)
(*qf2BSXD6yZZe+g4bPvukU6jaSBLkJQpL59Hj8MZ2NyhPziUDJ89qfiYLDp82*)
(*y/GnfCIZyNwaHT48OLzOhN+FR5dT4JmBXsGnozh8WN+zS9nbqXD+la+ciToO*)
(*f1H4dlz/zTR42vzy4DdtHH7qzhkG5l9pMC9b7timg8Pvm3G99+tEOhCu3DnD*)
(*b4DDP354+ghNbzoEKRu8YbDC4cv9frR9ZPoNTPL2LEO2ODx/Iq+vuOxvyM20*)
(*J6TZ4/Al40vjOgG/QXwrp+DoMxye2zwt8cfZDBgn2V5/5ofD33+57/Y58wzY*)
(*Jq5Mzb3B4Qu+B7EQ32aArVtRkN5bHN58yOoRdTwDWJUZOrc/4fDZxmISGp8y*)
(*QfT5TH/Gdxxe//a3tIXFLMjEaWTG/sHh53t0Wb8LZsNlv+ZxgUIc3teY2eLG*)
(*pWx4mBrr/qoYh8+891gw52M2tG17Fpwuw+HZbRV83M7kgJhm5Rp7HQ7/Y2Fh*)
(*QMIkB1hy/TKPNeDwFxzizvX75UDOm654uUYc/pEL2wK+OwekmAJE/2vF4Su8*)
(*u+4weuTCxwnfQf8eHN6QISg790cu/HaPrr3Uh8MvvcZzPmzIhWr/1dLFfhz+*)
(*QFBCRe3+PMhkkSkWH8bhczkMDnio5YF7bVF25ggOrx3C7ib5JA+Eovq/nhzD*)
(*4d0/O0m/L80DRTuFsywTODyPoPgbpek8kH3j0GE2icMnR/WOkXn+gJyR1LX0*)
(*KRy+49ulSD2LP7Dr0YXMo7M4vK0oZYUp+A/wUHveXp3D4ekTkq/9yUHXy74c*)
(*t57H4WVSuBiFmfMhO/rac69FHP6fZIVJvXQ+3GNcEnNZwuFNfrsUPL+dD2+8*)
(*hJ3ukXH4VVkJPmnvfOAddjdQWsbhg3IG7IaS88HKIK2VdwWHF5X/UBvSkg/8*)
(*e/Fd3cj5BSpHlTfzoaj7p+nHVRx+ouRXT7xWATw0Lh4YouDwnpdMT+s7F0Cu*)
(*3t5KJyoOz1fJ+54lpgBMaH9zbSGnqFVP51cVgKF7ef7TNRz+Uq2bivVCAWwf*)
(*8fwzhtytJfVNRKAQrMV3Maqu4/B2jcMbDRcL4T39g7BIZKYboXpej9B1xoIH*)
(*I8hf2q78PvmhEOQrdmz3beDwp/S32EYKCiHxjXSsJnJdd/r9j6OFkFyouGWD*)
(*bGZk/vfy7iLgPsHj5I28NrBXmHKqCHrsg+kCkN+Z1TonGhfBY3H/6FfIR8ee*)
(*N91+XQTFx1sV7JGLLGUk2NKLoKHwbI8Osu7UmG9hZxGkZ7rYHEWetg4fsqEt*)
(*Br/X2jMzaHze85oX9h8vhjJjm6vxyAL2NJ8bdYrBAPciWAc5fTlz8YVbMTyh*)
(*k/y2iOav+tRSUzauGHDyXJ4vkfvXBBNG69B1lpL9zMiO7g20oSvFILRd9cwT*)
(*xZNt54Wh6j4CVH9OfT2J4h/28JaxhAoBjPUaQ64gi7ZKmHDZEmCUj1QWhdbr*)
(*tyKd2eonAtyw6+MeQeupmNRxr6eIAPT6n54KI9fy/LIoGSOAUQlpUg3lgwT8*)
(*zWxmLIEFqTjTByhf/iSFZudKlsCypyUeyy+auy9OJyuVAK9Kd6jLAg6vwmud*)
(*G6VXAl+HgpruoXxseX7xz4vnJZDdw9nMjfJ39sZMoUZDCYgFHeJqJuHwciwd*)
(*CorDJeAEryNNxnH4Z8V/i6VXS6BpxqNiaBStt/jnkj37/sKNmaTWgiEc/uD2*)
(*xfIBm7/QGjQ4Ut+Nwz/IlLjc/OIvHDol9nihE4dPfbC3svzTX3g7fPY6QwcO*)
(*f75lpiqp6C+M1HWRGVpw+JuJn2sddpeCvu3VH6Y1OHyUkbfm/YOlMOAS17Wn*)
(*Gocf5rap1z9VCqTJcr7CClQfPZSICkalML472bu/BIf315ltZvpVCk0RBtuu*)
(*OTh8I1PnjfWSUijxZLtVlInyvai0daa1FPYdcjVfSsfhY4+FtTdtlkJZdcST*)
(*kz9R/mwqdUdqlkHsN18huhi0fzNOGASblEF8qdVQUyQOr27J3+vlWAbjtU60*)
(*H8LQ/m+a7bOIKoPo/t8HJt/j8IvxYUNSM2Ug2Mb+ZuQlDn/k+txE2ZtysEgp*)
(*l2W0wOEFPVXPWX4rh8RfvtFvTHB4zl9f37DmloOn376+bVTv15lvSF4fLoeO*)
(*KXx4KuoHjSW5Dv1nK1A9PL3vJKD4S7/Ypo5VgLQi358MXjR/4+6rURsV0GpW*)
(*LLebA9XvQNlvwFUJrvggC13Un65OjF3yvVAJ43sUIxK3UP/7qu7P86ESXgpl*)
(*y9wfY8M3cPDySuKrwE/CvkX0Nxu+TMHGoulmFZAfNjy+mMyGz7OqzHG0qoIR*)
(*PL/3lVg2fGyVy+2C0CrQoz7MEvqE+rNn3xe12SrQGYzarHjKhj80Hy9+L6Ia*)
(*VEOEuPfLs+GfNshD2PI/OMsv0HDpJyueXTJ77B9zDbwIk+O+8p0VHxckHbgh*)
(*XAPxNC4L8uGs+CbNY13Gl2tAuS5BrN+XFf9fLa/DkdAa0Dh0QFHSlBU/UDUf*)
(*n3GmFly/ehh2c7HinY9ZaY1q1MJC2bb+W3Rewb0eJ+8xqYUXDFNhklsseHmV*)
(*PryrXy3s85AL+G+CBf+hvKYbumrhVwrPX+EiFrzq3x/s9c/qoEuR92+1GQu+*)
(*/8DB7O3gOmiZOWm19xYL3tErykA6tg5Iqf1z2hos+G/wIfFDbR1U7BU95iLL*)
(*gt8o8lK6I1wPN9Z+d6bTseDT8g2cSUX18CZPWCYhihm/N4urn5aWCN3m2kEs*)
(*xUx4w8AN3VE+IgwllmyzpDLhv5uP1VZKEMFHei91OpoJL7Hnz583+kRo/8kq*)
(*qOvBhFdyNvm0J50I4X+dtTPOMeFPR92K/yjVCPdtNhnepzHiTY3jjRL7G0Fy*)
(*ZDhX8QUD/pW6UHj4ZCNcl21Mu2TDgI8//a41YLkRKpYNw4/dZsDPsD/TsGFt*)
(*gqFnxeyB0gx4Z4KWvMzpJhDh1lKu7KPHBx5Y4S0MaoLCc8Ji1JP0+JzhS3XE*)
(*C83QRvpxVeEvHb5d7cVikGozMKoFU32T6PCrqcV8mjrN4P9881f6ezr8GTd5*)
(*02rLZmj5JOMSZ0qHz+ORWi0JaYbwzusGHDR0+D/KggcyJprhToBCLP0ZWnxh*)
(*/ILDx48t0B7QO+MlvAOlj6KFbs+0wpOfJyefdVNg5agCxy+bDjCu7KGEug2D*)
(*4rBSVYBLB5hrXL2fvT0EvtFqng9edkC/TuVXQe8h4OPVnT8c3gGsrY5L5f6D*)
(*cJrWihhR3gEpEbw7zfJ94NIT9tZPqBNOz76nnU1qg823yxzm/zrh4+1NbSua*)
(*IlDWWK+62NoJxg6KnHcC/0AgI43X/oFOUJFhctcWyQERD7aF7uVOUN7LlnMm*)
(*OhUUHx1qvH6gC2gElh2fc3qCp/K1d/C0C5a/el/x+pNLoFv7ySl0pBuoTzbD*)
(*Ln1oIgjVK5FBqhvCmt8kRMs2E+S+dbWZy3cDnmeiy6K5mWBxhSkyVasb4n+Q*)
(*ZE+ztRKqw82OKDl3A6MZB+79w3bC23MC5x5WoeepSrqe7egmJLKnCwc3dYOk*)
(*urqkkVEPoWRYZSejpxuc/0surx/qISwFOJZvznfDl9yf9wPHegl6PUTtd/w9*)
(*MPxuW1q/s58g7O57L/dhDzz8a1u+59QQ4dQ1EZVehx54mXZMX+HREEFbNEuM*)
(*9nkPRPjExnt/GSJ41g3Nqof0wO3XV5Mt6IcJw8IKLv356Hn3zzX6VwwTNhZa*)
(*7uyq6AFD3Bf3SPIwgbfikYIYsQeOTq3jaw6MEFRswujsR3pQ3BptPV1GCElF*)
(*5EAGXC9cip0Y1xQZJZS+D3h8nK8XLOuqIjSURwndFgd1rh7oBZFsV7aHj0YJ*)
(*OPare8PleiHOruELbdYowc44+ZuEUS8E//ekJfvCGMFf5uKr65a9kPSb3TnH*)
(*cIzwnbHzvvOTXlBsP5jS5TZGaEllkCjx6QXdHnrjgNwxwukdk5wbqcjyqbaJ*)
(*ouMETa0S78t5vaCs5PlQV2GccC9y/9Uzpb2w/foU/uTNccK7M/0kwXY0vqSw*)
(*Fi/vcUK8j0IWbrAXxoJ+qK98HicUtkR5bU/2wtWPx7m/pIwTJh8bCAxv9YLg*)
(*qur1dy3jBJri/LEWpj4w/ByQ1zM2TuDDCWZUcPXBZduHV+9SxglKiZ3qSUf6*)
(*4Air8EkefhJBf/XM3kipPpAR6Yu+fIxEeKz8eSQQ5bF97JBH1ikSIWrwprud*)
(*Vh+wr68La14jETIks9XMbvWBrh/riaeGJEK1O+8eXdM++HpCRGfEkkTor3EY*)
(*umyF9oF1cOl7exJhhb/l1xmnPoh24M1/6U4i4Cxlnol79sF9zTjXfB8S4VD2*)
(*exUh/z7wabipeiaYRDi7a5F794c+CDFRtd34RCJoX782sB3VB6f9XygyR5MI*)
(*5jFpPxfi+0BIoxt3N5ZEcJtldxlO74M9dzplaJNIhJDztsqt+X3AJxHKvvCL*)
(*REj0r+esLEfjG4yjO5FBIhR3SPTlNvSBFeu8059sEqH1SGBSUmcfKBtHtoTm*)
(*kQhTDtNOkcN9wKk2H1CZTyLQlqorBc30QbHlSQ71QhKBnzOZ3XO1DxS333Qc*)
(*KiIRJI1Zeuxo+qFQe95YB1k55UGCGWs/mM6RD3ej+++sVzno8vZDA0dRYkEB*)
(*iWCndgyvsq8fhmu77al/SITXob64s2L9UG0hwRaSSyJ8GR3rFJfpB9qQ0i3/*)
(*LBIhS+byD6EL/UBf/rdrKJ1EqPGKe7JbpR98XimfjEwhEQYbdinuXO2HgUaF*)
(*8NwEEmFV+B7r4u1+KAkwSDr7nUTYbVXaPnyvH9g2Jzb2R5EIon8Oxrba9MPc*)
(*o98nHqP4nmN68bjyaT8oVA/3HUPxvx8LzMlv+iFIwflirweJ4L74pTXyUz9s*)
(*yAkp9jiSCB9g+2tQTD/YszhpqFuRCCU9hfJPMvvh4s8I8NQjEdrFhRnvFfVD*)
(*/h7552oaJMLMU7dm3ap+kGNbc/cDEkGA95zV2Z5+OF/z+JnDURJB2iz8zPGx*)
(*fhhyoek9KEAiqKRTdwnP90O3l1SZFhuJYK+ZG7VDNwAFB9TJonPjhLpXcg2V*)
(*xwfgTCBDPk3qOOEcRy3d+1MDsKu69MzVyHFCYrjpGQM8sqb6fV6/ccLL1OCv*)
(*8zcHwOqKVjSbCbq/c8Ke/8UAxMsNDU2xoPv/i977qGMA5IWPPdBTHSPw5chq*)
(*yg0PwNU3gw/4TowRXuFrvLZnBiDs1a35x9xjhPTkDccrtINgIu/JS+0ZJbwh*)
(*iPX38Q3CI3Bf+4zqh9Kkdzqj0iDUZ+dNkp6OEFwlG2qsVAehwW/1v/P6I4R0*)
(*B4GxRs1BIO8xm1E7O0LYt50mEHlrEB7GBBZGUoYJa9x93lI2g8CbSonVdRgm*)
(*pF84c1M/bBDyRc2cc02GCPvfTa8lzQ1C5653xzmhn1Bs6pmUQR6EQV7zM6rr*)
(*fQRjWZ7bBdRByPv0M848u48Q1X4ur45uCMp19w4a/NdH4BcJcJ7nGwLdsi4a*)
(*N75eAneyGFkOhkDiQMIFo+kuAkO5xQzh7RAM0ozUKka3EeI+rUVWfxwCXosn*)
(*b2RvtxGULYM0msKHwLnOpVd9TxvBmzU7eTh2CEokdPrHg1oJ29foHzLkDYHZ*)
(*jm3oulcLgdr3fUx9cAhC8qmlKY+bCNOUgf426WEYvvlSMmq2hpBj+bZU/fQw*)
(*yBlYR1s+qCF4dSjGE84Pw3bu0vvmkX8EvtxomySVYTg7fqMgqLeaoORssOlu*)
(*OAzWiapl5W2VhIjldgFRv2EwfyefcX+ilKC2WK9rPzAMyp45euPHcwk8ph5n*)
(*SaPDkLVqcnRWM4fQ3yghbDiF7tcROm9rl01wTH8zcnllGDpfHWIdKsgkxDy+*)
(*Yi/ANgKGlvp3Q8zTCauz5W9LTo/AO9P5RLv1WELcVEENe/AIfDWcP7fzyA1s*)
(*P189/OjDCHBWNtpp8LyEM8ojbhVhI7Awt+TNfd8PaqJZT7jHovtnTQ/fcXoP*)
(*i9duv53MG4ES2sVrX/MiQTF39Ub5yAiMrgy00MQkALN5QMr+yREQ61xadatN*)
(*hGbOfQxuc+j/rXJTRitJYPHocs7JtRE4e+RAq+nFFAgU+cgfwz4Kjr6mRsZG*)
(*6dD18mSfq/woSBxynsuLyILv0hWn2hRHwXn3Pb0R4Wyw7tMPllYeBV7pJJHu*)
(*6GygOeMFJO1RCOw0eiYblYP6aMM3XfNR2GaUKSsPyAPHa9b3pd6OAv1pmZVh*)
(*6UJgF0mYHx0dBT+m43d/6/0FvN7LvpFJ9HxuOMf28i/Yv71bOzw3ChWOMfmU*)
(*1L/QTsufMEgdBaoL+cMbhlL4Mu5n3IsbgyXD4a0M9B0lmf6ovkVuDNbODyZ8*)
(*nyoDk0mVgmb5MXh0fSPHg7McQg4dTmpSHAOX7Mzl1VPlsPqx5xVRbQx82+lb*)
(*Kj3LodhV80Kt4Rhk+H4amWevAK1Lkj9LX43B/aBf67uEK+G5O2v4X/8xIKbo*)
(*ifGh75T07HHfkuAx6C0si28yqAResRiz4rAxYJy8q64aUQm9OE7B/JQx8NZj*)
(*oQBPFVi3L/r9bh2DPHn3lfWVKvjC0eCc3jUG+uNZ04uc1dColmye1o/m83Qo*)
(*M+q/apArMMP/mhiD2BumeE6TatiMaaEkbo6BtXjgP6+Kagh8mH3/q+g4TLA7*)
(*zXG+/AdT+F9+bMfHIbQhRdoh8h+o8v9IdpYaB4MKVr3Xmf+ArurjnOa5cah3*)
(*eK9YPfwPXEUdndeujoPlfZulEoUauD8g46vjNg4qXcE6J8ZroDzneGKR1zjc*)
(*+l4wt2u9Bg4FH6oR8x2HldMqWZ9310LfBW72nXfj0Bk9YLYmWwu6EYuffsaP*)
(*g+nU5Jdcj1q4eDP9B33zOPxbUJBOZaiDLxKJVY87xqHHstQvbk8dbO76Otnd*)
(*Ow66tz32mR2pg7yMtyd+k8bhJt83o1PKdSDJ+TjbaGsc6KUPPtb0rINA0v2O*)
(*ajoSFHTNbwUG1cFUsfGaLDMJxPI+mYdH1sEPG20FVh4StN9wWdybWwfCtScq*)
(*csRI4MlXoVswXQfPvh8hHZQkwb8epZWH1DrocBVhCZIlgQWT/d8J+noIEdut*)
(*dU+RBGsMTkv6IvXA4jvTyqFLAnUDSx5uzXpYxf8cfeBJgtOr59r0vtVDYdx0*)
(*5NNXJFB7kOoelVwP3swSN3z9SRDYt0+bkFkP7I0pf2M/kGBoq3cmuqIejt5N*)
(*/dKfQIIljxizf6R6mCmduzmbQgKmnoUEuoV6yDgqtXvzNwneDxrTCFPrQXEu*)
(*zU2gkAR9Gb+uLzA1gJ7Hb33dRhLw+QodFD/aAMJDixz32kjgI3ApzfpEAwwr*)
(*y1Q96SZBdVAv7oNcA9jiMk+9HSWBjlXya3+lBvCNzOKuoZKAf9v47VGjBsj+*)
(*g4J0cAL4ZjuNLYMa4JkI9aX20QnYnOZi3w5pAPyLs+eN/psAcxcRX/ewBqhT*)
(*+5P07NQEmBwoop6MbYDRjnzf7CsTYBDD1O+b2wDJ5zcUyrUnwAG/r+FeYQPY*)
(*xZxfab4xASIXZauO/22AzfuF9xaMJiDQnTTxtKYBeFeL8P/ZT4DEeabbfL0N*)
(*0H17myL/dAJKMzt3aQw2wNcixTQ19wkg7jdushptAAkfgsh93wnoOQDt9jMN*)
(*oMz7d/1rxAQ8P6zkEL3RAI4y5Vl8ZRMgTNvvEYG+W+/LHfj6vnoC3HRuP50X*)
(*IIL+afc3uIYJ0HVhfn9ShAjnz8mZ0XVNwK95SfQMItApxXHMzU3AggWJ96IU*)
(*EZYv0Ww8WJ4Aj/AXsYwyRBi7bDg+sjYBXRI1DgVyRKi+wlvURT8JBzg/9pLl*)
(*ifD2+qtHFYKT0AXmxfGXiOB1Y/Am/sAkBMFf7dnLRLC/eeFiwZFJWFNX2HNE*)
(*jQh6t5f3/pZG1x2XTj7UJIKwqXlFlMokpBhf2KWuSwT2eyXpezUnYZhB6vsB*)
(*PSLQWAhHhVyfhM57IhenbhFh5EGrvZ/hJLxfqbp1y4AIyXbKBxztJ+HT3ELm*)
(*F1MiRNl/ZZt/Ogmh1pWHeO4R0Xs2Vx96TILKzRey7uZEsHPJqr/rNwkOQ7do*)
(*TlgSwewZ55/uoEnAf+sVcXpABF1367ibHyaBu1noa8ZDIpzxEnXT+DIJe9Ie*)
(*vWW3JoKYt9f9ythJMOl8ePuEDREEX/Vev5g0CaIXGJSUbInoO+Sj2JmsSZB5*)
(*sHnoqh0Ryt/Stx+qmYRJ+uHFNQci5Lw3+RtNnIRzjj7eDY5ESPxQmMLfNgm8*)
(*yXsCI5yI8Oaz40v2wUmIz/X7JfCUCO7hjbb+Y5PQcnJ9Tz2ybaTEHfrpSbj0*)
(*wCXMxYUIOjGj0usrk9AX8ux1nisRlL/hhZ02JsF2bzle6xkRTsVGMS3QTMEH*)
(*3sRf3cj8Cbr9Y2xT6HxbeXrQjQisSen/TLimoPnssSh9dyJsJuOye/im4KDm*)
(*E9Ua5IHU8jdNB6fg09uLW+EeRGicpZ+ROTYFEknCcavIJRKXNT9KTIECbYWu*)
(*1nMipFv5pKycnIJvEvJSX5C/Jlfgbp2ZAs9TfDcmkd9NMtjkXZiCn/FhA5Ke*)
(*RHghplIvoDQFMh+iybbI9pa+J9xUp0AyIvZXErJZfGVQryYa73jtfwPIOmOM*)
(*cwo66H0SPW84vIhwSVRVO+bWFNz9cm76HLLcvdepNEZTEDRGvW+KLPq9it3M*)
(*bApUOWqFXiLzDjE9LrOcAo9nu49+RWY4oEYUtZmCZUL5tzzkVWM/KV/7Kfgd*)
(*+iW5Dnk8uvot6ekUmGQtGfQit/cyL6h5TEFRUlkjCblK6Mq1ZO8poC8gHJ5H*)
(*zrvjn87mNwXOWRcfLyEnhv/jtAmagon86/WLyGGdLE8aQqbAnD/+1iyy/171*)
(*JqmwKVihjzs2hvxML+Dk+2i0XlYmel3Ijz7VvF/6PgXaP79u/kO+08q6dCMR*)
(*xU+3RyYXWYNHQyf71xRYjpzl+oZ8QedNBl/mFAQ7fkx7jSzxvpbbJW8K1l3c*)
(*j1ghCzeyOXQVTYGhuWyABjKOQ7PlXNkUNOqnL4shzwbWfdiqn4LX581P9aB4*)
(*99filo1bpqAije9sOnIDq5ZuSSean+vrcG/kVL963pcjU6DG7dGwH/lL1W6n*)
(*0YkpMOC+UDeF1vsto3bb5Tk03iVJj0xku1cNn5jXpqDn5QF1RWSTMvbVh9to*)
(*Pb4fD6ZBvk53Va921zSovvdqK0H5JeNF5Atmn4a2d+cKziMvuzV+5hGdBmtD*)
(*jwVhlJ+j+ZxUR/Fp6NW6bdyE8rlt7Zp+u+Q0fNU/cO8Vcs7TJv5w+WkI+GPn*)
(*PI7y38WhOXzf1WnwETYPj0L7ZeNRa6T4s2ngjnn+u98Z7f/TRu8lPafh5keb*)
(*8wHIKbTjvrKvpsGzxNbsFLJxGOWJQvA0nOn9khmA9uvfCkG169+nwVdBLkgG*)
(*7efXB0yXXWqmYWjwzqTnEyIMTk9OehCnwfvy+T3Hkc/m2g94t06D8nUTt1ZU*)
(*H6a0XtUE9k/DJ39D/uPIWm4JX78uTcM/oePPulA94Wmf1awWnIERlbWgq6j+*)
(*WH9zVqrfPwP4YgorLXK59c6ZZtEZeFBqKp9pRQTnXVyHeyVnQOt2QJkAcpe0*)
(*3Nq80gy8vxdxagHVM5nNwrll1RkIblQnJSAHVF0eXdOcgSNRmq4myBeMbxF3*)
(*3ZoBBa2HN1pRPYx58+zHXqsZGO8PsKlC9ZKqtytS2G4G6n8fSPdFvnYo8N1B*)
(*J/T+W7fWVZDp/kS7/ec5AzErJ3Nr7hPBYrzkOnyYgXBBF9UBCyIU/76iqhw2*)
(*A23ZWRcSkPk8mi9ciZ6BA2rzt+2Qq3lGj91IQPd/jhFkRP7vItP2/YIZkDmu*)
(*zYNH9Ts9+eexzpIZ4Ml8Wb8bWY732vUrlTMQrxRE14fq/QVSWNx/TTMwrG45*)
(*7Ylcck2hIap9BrRvcsTfQL6cP0TZ3TsDpDvnJ8WQtYOOayyMz0BEjteeTjMi*)
(*NK82OJrOzIDJmBd9BvItE4cvzYszkObb8joI2US2cDFrcwZCz9LqXUEejTIR*)
(*Oko3C8rV/T/EkR8yMlz+zDQLEi95P7Ih23dphblyz0Ixu6luK+pPq0pLf6f2*)
(*zkJGyrfjBchuKaHTBiKzcIkQnBCLTMN3fk/9oVngOrNZF4T8ynNAUVFsFnq/*)
(*xkW5IjNPvnyQdmIWup6K81siB+mIhRyQnYUAZ56LesjchXUF78/Own/jzfyq*)
(*yKFHnozRKaL3zRC+yiMLvt3D4XhpFly7BLtPIMdQ/5wdU5uF13a4ysPIombG*)
(*Znras+AiF28hhJxYSxdYdWMWhNKG//AiS5xKyD57exb+5f0o40BO/6IxkGQ8*)
(*C8+KC1/ikOWYF5iFzGchmHFgnRX5z5OPMoEPZ8Fj+Zc4G7Jiz1nDLdtZ0NTI*)
(*5tyNXKrc52PrOAuztlbpnMgqqS/SBlxnwSnuJAsfcs3eo13XPGfB+LqukAiy*)
(*9osautJXs5Arz0Y6gtwyZSshG4CuF7laSyPr6/Loxb2dBTXetJ8XkHuLcj33*)
(*fJqFJf/4OHVkk2OGSb4Rs7CSq3rrDvLYO5oWSswsVBDul1khP1qP23zwYxYK*)
(*XpZPPEeeu3flaHfyLJgPsFd9QLavn72qkT4LkgODRsnIlNMhroXZs2BJKU0p*)
(*xdaPtafuSwmK1zsPOyq2fg6eqxyVaL4TbKN7UD6w9B0+8KJ2Fmrt+XafQuZO*)
(*t3a41z4Lx3WuuLsghwpwRbf2zAIHf2RJFLLQy+zKy0OzQHenMKcUWVRvW0Bs*)
(*Bq3fT8Ys7HyTSPh+KXxxFn4pteQpIp8QV7VhpcxCfG3OIyvk05tvS2Zo5+Az*)
(*y+3Gf8iq3w9a/t47B5WGJa4R2P4JYz9aKTwHXSlLKw3IB4I3RrsPzoGk3os+*)
(*erT/1l3bzOgl5mBqtGbCEfnXdX/jW/g54J82LrdA+ztC1VnE+vIc+NzmsvuB*)
(*7Ktwr9dLfQ7uLCk+G0c2EVe4k6w7B4X7e79bo/rAQ7t4c+vBHJxc+cvpj+oL*)
(*DaWPl9t2DjYcTru3IM/M1LQcdZgD+8j20/tRParsjLt+zWMOODTV3+cju6br*)
(*a8W+n4Nzi0wxNKi+DdwtUb6SPwe/jDdFq9F5rO7mr13GhDmIiMpzPYrqZ55G*)
(*RKl9+Rw4qq5QfZBDzjjioxrmIEO9x1rtMaoHHGIX5ofnIDH+w3A3qr8/i9/K*)
(*fmKbh70BgrmyqJ4/Fbl7aMhwHk7qqxO+oX5idk1ar9d0HnRphNNFUP/Rfkkb*)
(*0HF/Hh6p0FhFIB+djF2st5sHAX0muVDUr1qzJkryX83Dwe6DlaGoP0pr2t/9*)
(*mDIPM/nfvP+8JMLEM58o1c15oEtiP2MfiPbXzE22Dq0FiDKV3iscg+LFscBc*)
(*vr4A176cC1GsRvWpKOD1la0FiM0NPdiIHGJ9hJG4swDltM80Tf+hfKq5s6ub*)
(*fhH4AxpsvWuQfSq25tkXIQV/t5BQRwTOjYglocOLYO0MikeaUP8nXe510FiE*)
(*YDmxtpAuFL9Pg3fWtBZhIpZdfk83EQKV3bqeX1uE5YvXOT8js31Pb/e/uQij*)
(*zMx7wnvQ+dFIqCnm7iIE2Kwbhvah+tSyUFHrsAjtBfyHnYaIQE+ITDsctQjO*)
(*t1XUiieI0H1BWW3gyyJo7qJvPj1JhLT8mcGIb4sQ+PppZCrynRwFbu6ERciU*)
(*lJmOnCLC718DjjQZi3BDgWJrM4P6b9Thc31Vi5B2xvrr9DzKX8G65rCaReh7*)
(*SZ66tYD2V5ijlW79IvSvXLQtQ876UBFV27wI0RUH94UtovG/sdz507cI/ges*)
(*hM+SUX64/iwPJS/C8PRPG51VIgRTbhjprC4CW2fRiQzke06bK7vXFoGBS1uU*)
(*i4K+T55oHfPZXoQ3ptNZdcgWD+b9HViWQNGbyeLcGor3Ldmr1/YvwfhFpj0L*)
(*G0SwkivoYtFYAuZh2t33aBqhpLEtJU9rCWp1q1t9kPfYLHhaXluCH1mTionI*)
(*hB9HjlbcXALv6L0b08g8e9/Ze5ksgXb5I4HHtI1QsGbORnFaggt3l0TN6RqB*)
(*M9Sz/4fLEhzdpki+QLaQifit67YEIbH3B6KROayItzK8loCr77dcO/K9Xvk4*)
(*mzdLcDsn+Y/SrkZgI7ArjH5dAskgF2c2+kYwMRDn/BC7BLw6wbsPI2dTLo1c*)
(*jEfjMTG/dg75rrSrf8zPJXjo1NBtiZz5baTtTs4StP9T3E1AZlbcTmT5swT7*)
(*s8m4VmSjbn73vIIlyO50L5hAZuLRPsT3dwlYvyYc5WJoBINXebZNtUuwZt+M*)
(*N0ZOP9By0athCZYXP5XaITMUzfJKNS3BvfNuq97IaSuHCgLbl+C/dRPneORd*)
(*lkFMakMo/g5/hieRf14x+1q8ugSXJoobzRgbQU7ioCKsLYEaC67oMXIx+2AP*)
(*YWMJwu2qrNyRG1uM9v6lIYPHLlbuT8jLRreDy1jJYBB9M7AM2QPPL6G8mwx3*)
(*F5NCG5AZD3f8K+cgw/Ump/tdyAITugyVvGT4GLR1aw5Z0f6ae/U+MsTLhgft*)
(*YWqEKl0OwSsHyeD8yrFhH/K1Mw25/w6ToSy7v+MYstmmBrlGjAzlh+cuyyO/*)
(*9lV9VC9DBr7bI+G3kTkfMjJrnyKD3tK+46bIERoVPxrOkEH8DrPvA+QUzktD*)
(*xAtkoKFz+u6M3BihqN+sQoavlESXd8i3PbZWdK6Q4RCrypNQ5OG7hR9aNMjw*)
(*JeL32SjkZdFzxNZrZOidfXkqAdmDac365g0y3DTLsEpBZpzKZW2/SYa0TEX7*)
(*38gCaadUOu6Q4YTFzmI+8vf3yyO3jMiwWLJuRUCWcMx80XmXDMkzQ7llyIry*)
(*J4u6zMmgeo+mqhaLh9DCnTuWZFD+vRVAxOKxnUrtfkiG2Cb7Qy1YPMok5Hpt*)
(*yRBM1q7vQp7+Md1k+IQMpmXPhnqRHf2SH/c5kKFoO6FmAHn70cPdxs5kcP1y*)
(*1n8Yi5+W2M9+FzJYWAeLjGHxkyap3XVD99Oy+ZGw+HHHjw94kOFnBce/SSx+*)
(*HYcPDXmTUT8+3TiLfCp/mGDqQwbKIcHP88jFUd+Mhl+TYSaQ5fQisqqnyYZZ*)
(*ABnUmS1+LWHxNt0fPhJIhocZgVtkLN7K/afN35JBxGbw6AoW76PRraPvyeDf*)
(*RxBfRbZmMbS3+EiGyUcBjBTklWlBzvFQMrgpxBRh/hhNqooKJ8N+57XbVGTZ*)
(*q1meN6LIMNUxXIO5mebFGdYYMsTlt3CuIT/J0Jov+UaGZn5lCcyc5oIJT+PI*)
(*AME/RTCn7SEZSyaQIUDvxRj2f+2qTL6xJJRvGhdeYZ518WqITCED55jOOvb+*)
(*wONavjppZLh26rwa5v96BRRZMsgwOPnIFht/TdD4CiGLDNpXbtlg83sImb+c*)
(*c8lwWMbp8jIy86KnxYl8MvwSkKdg8Un4rikyWkiGdta9L7H4qegKtEUQyLAW*)
(*5z2JxXeMYTzweikZKgTnj88hv8rNUGauIMOBk0TVGeTDDz03i6vIwKX4CT+F*)
(*bFrHby1RTwaOexFV2PrSPB87PEIkA55coDeCHCOV0RPeTIaLD/zKBpH7QjQ0*)
(*mDrJcHlj6EwPsrsy/67ibjIMLWjhO5GFVkfzHfvI8MLlkFgbtn63nx8fHibD*)
(*ndqizw3Ibft/sxTNkmE6NZWuBMu/Jo+/DgtkyP9xg7sQmeeluutxMhmWqivo*)
(*85Cvj49MfKaSITIo/0Macn0KX7X9rmXg93pgEY2tv/GIpzjjMuwu3QwNQ2bj*)
(*TD8zyLwMPl3fUz8gX3G4kqDFvgyU1cL3/sgV8u6+YgLL4Np3SsYR2XxaTXFA*)
(*aBlyHz2MsUXeFb1n9dO+ZdgIbVrB6sFFmjSLXaLLwJlbedcIuahySLlfchm2*)
(*8ltlLiNn31Dd9VF5GXhc94lxIftp5PkIqy6DDiP5Hyuy4SVx5h9XloHu2pYh*)
(*PTKdLNvuHO1lwHs136Kg+qbNTeTr1F+GW0n4xF7kiUa948I2y7AZFrsQhyyo*)
(*ff963Kdl+Bn8dPIEsqeaT0zW+DKkHr8pKonqc7PxA0PpyWWQTrly+CiyqJOG*)
(*QMr0MoS+fH9wH3LNN64PsQvLMDGXIcyOzLsR7Ruyvgw4WQ2uOdQ/En/l2Nix*)
(*r4Bn8CR3KnIj98R5idMrcDGtwEAK+ZB4LTXh7ArQlS/4HkN2gtTsw+dX4Fiq*)
(*Y/Z+ZEFrRykh/AqsJxKFOZHNK2gOs6qvgGJ70N5F1A+pTwVYJ4xWIO/1UlMm*)
(*8sE+9c5YnxWgpNIsXEAOKD2fdtNvBYajOXlOIS8lSPgyvVmB9KW9CieQS+3Z*)
(*5azerUBflVG8CLI5c/NbmcgVIDaTiNuo/ybK3lb9m74CvdfqgkuRJQMsc/p7*)
(*VuBV1MCxa8ihj/WD3vWvgFbIefUryDu6V8yVhlbAufCBkxJy4/7/UAVbgZxw*)
(*g3E55CfZczY2iyvwvf6jrCBy1qDTkU3GVRj4NPZiHJ0Xzp159VFAZhVoeXJ4*)
(*fJHbzQROR8utwnE3/WIvZIfg1I4DZ1ZhL+PJJ27IKWOdgmIXVoHc7jtvh7zv*)
(*44lvp1VWQc6yTtsQmW6x/deNO6vwxdz+sxzyF2Hrqx2GqxBLsqyQQj6vRrt4*)
(*5+4qcP4I2DiO7Bjzn5yZ+SpIe17zOog8ruWVb2e7CvHSsdPsyLXJ4lVB3qtQ*)
(*4CSmip13HrQXP+DyWYWdPy2r48j0dLqsn16vwo1ir9RhZMXbz7WiAldBWbHp*)
(*VDdyOlNLS3LoKlzhG4j9h/zR3H2wKnkVHoxS1pKQDfcR12hbVsGSjcfaDpld*)
(*wD67oG0VCs3sa62R//LseeLUuQqsG89lHyIfZTGYmOhbhf58LxEz5HnyeDtx*)
(*YhX+zarq6yJ7/9vOjN5ehfyCHNNzyHLl3x7foqVAeauw2hlkUrHyf1z0FHj4*)
(*+OdpOWTNrIDvr1go8LJxn4Qkdt6L4XtvxUsBi7p7Tw8hJzpK2sofp4CtyoPX*)
(*OOQ7j5vEyRIUaPzFN8OCvPuR41iKFAWUjO7oMyHb3803PHCKAvrSbqp0yOfV*)
(*VTSY8BS46nxse2GHCA37jcXablLgm9ie0DpkL0Ha0WB9CiRaLatVI8vuiYtR*)
(*M6BAaFcpSzlyGOsUX6EJBbblerMKkM1WnBhirShAOVKV+hN5pSZo2O4FBW5w*)
(*Be4PRu5la3uw+ZICLdekhgOQSzWF51/7UoB4vyftNfK7huSNL28oUKbJav0C*)
(*+XhzFW/9JwrMVJMsnJA5eTgi9cMo4BS7+twemXJD7+BoBAU2XXy/P0Yubxs9*)
(*sRFDASHGMeZHyMZdtKriyRTgkv977S6ysuCV+qwUCrD7bBQYIv9n8O4GPo0C*)
(*r3MkT95BpvbuM7mVhebTeV7+JnLI4HlXn2IKGB/nK9ZAriQ5/RxqpsBeOtrA*)
(*c8gpYkUnbdpQ/O/2BJ9F/vCQPo/aQQFq+9eI08gm0yHlHH0UAPfkOhnk9bm0*)
(*PgUSBdyuT4T9hzwoRbn3b5ICvFvqQ+LIVXaKU7ozFHjgwisjhvxxqW7VapEC*)
(*KnYRM6LIUquTHBEbKL7FyWX7kPnOnAw9sk2BK5tuWiLIm09dhH/TUOGClsyA*)
(*EHLNGqN4NQMVPI+zCwogm28dVlrloMKHsIppHmR1RavqF9xU+DNKm8mNfNIz*)
(*Q3v3Hiqof4h7wYW8TXPRQFSQClKF/XIc2HrTGzvqiFKBrc5xkRW7fm/vqaaj*)
(*VChVOkRhwZ5f1rhyVZwKnIbOuzBLe196qiVJhQ7TLikm5NChzTN10lRQG6+9*)
(*yoiNF59DVZelgrWChDMDcvWO2DO1s1RQkthu24UseXf4XNU5KkjW3uDA/LE4*)
(*cuOyAhVMNAa06bD4erB7XFJC8w03GaTB8mGd7KmoSYWD/wVqbG8TwfDOL3yR*)
(*NhViT3JmbyGX/rlPe+E6FT4NGIpiDnbp8pbXo4LyD0HeTeTljpBLefpUyBd8*)
(*8nED+c4ZTfozBlQQJN8RxHx0tdhHzoQK/kezTq8jB950Uck0o0KkDkv9GvJS*)
(*1kkmGQsq6kdFDzDr805XpVtSYbuijAVzsUOcn9QjKmTLcaRTkUVbjK6kWlOB*)
(*Zi3EAHOAzF7WE4/Reqkr4TDrLfq/Oe5IhasFC+4U5MJrlzSTnKnw+n37BcyH*)
(*0jdxYq5o/cyyaTH7ceTUx7tRgSPVq3YVec72cfCR51RYppGIwKzbIHY1zosK*)
(*uynJNpjzTwxzHH5JBZlzi5cxHwiKbPzmQwUb76VDmH1mdN8f8KPC3Kfv9Jin*)
(*NNh1YgKowHJlZWoF+drPKu59QVQ4+by/DXMO64uWqLfo/iM6FZiFH537KBRC*)
(*hR5F5T+YJ8R+7RH4TIV5esdUzFmk/Z/5oqgw1MeejVlQtevWpy9U6KKrKsbs*)
(*FR/Cz/uNCtScnDrM4wyaXSGxVKi609yPWdOCIYIrngoFhuwrmDPKi++8S0Tz*)
(*kb/LgY2XX9RFiOMnFWyf/TmB+fnLk71Bv6jwMY/rGubR4akoXDoVEsINnTGr*)
(*K8UZvcmgQkBT8FfM6d+M9rFmU2H9YHQDZj7avQN+uVR4d8edBou3u0ljDFM+*)
(*FYr0D57GPEzwN/EtpEI904vHmNX2XzrIQKDCmMm7FMw8/dnf6cqpQHc+5CS2*)
(*vsE4D4/cSip8+Z7qiZn1vLK+9T8q4HdmmjDThTXvbm9A8a0N88by53ll+ERA*)
(*ExUs6acHMa8vm5RBKxVcK8UuYfm4dH3eJamLCvRaJ3mwfLX1yrlh3IvG5zTk*)
(*hXkq1UOSZ4AKvfOqS5iH2HCj7qNUEHHeNYjlf2OF2NVr82g8xjVj2P7RXJ4X*)
(*Z1iiwta+0sc7yNWHcunzl6lw6wHLFmaC5+X8w+uoXiToHqRF+y9V3uwIhX4N*)
(*lBJevcX2a+CvyI1owTXwnS+lMCMz95q164iswecEut9YfXjFevw304E1GFRu*)
(*scbqiZtl3v0nR9Zgr/b6NBvyo4NtTcrSa5C5Js6E1R+1T+yJ05fXIJkUn7AH*)
(*mcHjha78kzX4JiOsjtXHrMWZNYLDGkjTE6SPYPXovn6MivMayLv48R/F+sNV*)
(*qakbbmvwExymj2HvP9TnaeuzBpMupGisHtNXn/0ZG74GJbedpWSR6XgWaNlL*)
(*14CxOapLCZkmyThtmHsdrLRUhe5h/WCmwzZmzzq0e/HJmmPzlb4uaci/Dm6v*)
(*f6tbIO/Nu/SrTWQd6oTOulpi9axK7Ge12Dpo9jO0WCEbji3FpyquA8fPCidH*)
(*5D8HX3959mgdatqSDXyQn1vQGJ+xWQdSd425L9a/klxFlh+vg+vaWRusXzZK*)
(*W0fZOK2DujjZzR95Aq5HmHitQ1rAm3dBWH8wFgpVCUXjCWn78AnZPiItiLt0*)
(*HYQbvRbjkHM3I16Xlq/DasNK9w+sHhv7vrCvWoeVNKeyeOTXh42cmuvW4XvD*)
(*2/eJyFEprEYhHetg8HHXoRTkimILCe7ZdUhPyuPOROYfEarl4t+Acqb4/0qQ*)
(*jS4zVfwV3ABn773LmL8nLBU/EdkAPXf7gr/IEtb/MpoObcC79mGVMuSLK84R*)
(*709swLddX69VYuvP1PyQ69IG+P89K4WdR4r+82Pmst0AwfGXezuQT7furQ+x*)
(*2wBGpTe/Mae7J7znddiA6ZJI9U7sffWVgvwuGxD9p9O1C9nvMcOJ/d4bIKdZ*)
(*VdeDrJPpff1E6AbcosQrDyLXGnDzpYVtwBU5fSJmZfrv3dKRG9Crwnd7CPnM*)
(*zb/35L5uABc+/+EwsjBlx+l88gawmPz0HEUmnfOIuFK8AZ1N/s8nsP40grtb*)
(*W7IBTJGn1jF3v4k6rFW2AcGZy46TyHW9BSnXqjfgcFWq5RRyxvP14lvNG2CR*)
(*eEBpBtm99OmIxfgGVB1xrJ1H5lK3l/Bm34SQ0xxyq8hf/3KwnePaBMlfFyIx*)
(*S8v/mlzk2QReditaCrK22ES8qcAm/KfbUIM5gNH44EXRTUhv+3d9Dcv3sit8*)
(*NOc2warD/uQmdn45R1rOvbAJMSGi/pgPZbxqeQyb4HSJNIAZ/43wbkB5E1Zr*)
(*YwO2kD08T7ERrm6CVvHetm3k1fMHaZ5boOvHnsTQoPOlT2Zx/6kHmxBswVOO*)
(*mfc/w6LZR5vAT6M6gVlW8PMzI7tNeBthLkWLbEfFrVxw24SaH/55mCezqJMb*)
(*7zah1mk7BzuvukiEVmd+2ATZiM1WzMxxsglWoZvA9erTEuajH2zNeyI3wRgX*)
(*IbEL2ezJaH9+/CaIZ29HYu490djiWoieP1r4gB45ZprLYxdhE0wFhTwxmyXd*)
(*OBr0dxOyHol/wjwp2uHytXITgq7ql2BeFeoXqWrchNtNOB4G5Pyu/VXXWjYh*)
(*8NXaUcwen03tutvQeM5GnMNMzzNWOtuN4v1uzQQzF+uMJe/YJkh/VEjG3Fp1*)
(*gusLaRMa2jsKMIf5PM4/NrUJw6vi9Zj30ZJx5+c34Xc+ZQ6zBHXtt+naJjzO*)
(*UpFiRFYZZ9pM3b0FFzuEYzGzxF35cZZzC5y4bdMx15m+0S7l3oLG4udFmG/0*)
(*sX9r27sFr1UH2zGbtO5R2Ty4BWSu34zY94FoyK15H9EtkG1i5cZMuhoexnFs*)
(*C6ozj4lgtq0Vnj4ksQUDOSkymN1KD7+7cnoL6uwW7mBW9LKQbzm7BV+Y7cwx*)
(*0ykmDBue3wIj8xwbzH5/jp+yw29BsEy4J+bQ9JPdoepbkPej7Cvm27YOLw9o*)
(*bcGuHp9EzMIS2RLJV7fA3ow7HXNswlnPIt0tOCseWIw5/QuIjhpvwSaRvQsz*)
(*+RmZydR0C8acuAYwn76VMN13bwvUekZGMRdwcGR2PNiCVu+1eczb06Whulbo*)
(*+R8UVjDjq52fNdlsgQPNjXXMlV79F2vtt+CaP3kXM5ZvRiFH1Jy2QP2JJzNm*)
(*DXkVloqnWzCs1Y/D3Lj4q7HYfQuGXPn2YOZpMM0677kFtB8W+DHrJe8Jy3ux*)
(*BdJyMcKYe8w87mb4ovF8cTuMeR+cvCTlvwVBUslHMZsIjR1NebMF/e2p4phj*)
(*KWGs4sFb8I3iL4F5vEVz7se7LTBcPi+FWSydpvnQhy1Qkao4idkqMCs75hPK*)
(*D9ajcpgXlIU9wiO2QP+o3VnMzlsvlT/EbMGx6yIXMG9kT7MFxW3BYJA0YPay*)
(*vdHim4Se151yETPjsYIIr9QtCJX8dAmzMMudPMvSLVC2SbqMWWaa2qbdjuI/*)
(*aPH//+fFmkSr9WyBjUnU/5+vYFhtrjS4BVI3DeQxl/JKS1wY2wIZxx+nMKvV*)
(*fyafmkL5ZOXx//EHJGl0ZMyg9bGB/8+v3mcnX2p+C/xHxv4/fx1FS2/x5S3Y*)
(*4z50DLNR2mkuke0tsH4c9f/45tyhsjfSbEOuhqoAZnbGfNzLXdswXNP6//Up*)
(*MVRgnmTeBtAIZ8d8iFV5J4tnGz6ZtWxj6/8sm2HLkm8bih58+X8+tJhUrQsK*)
(*bIOw3NVVzK9y1Ve99m2DV8y9Wczj5tdnNcW3oVLVrPv/+4GTZ3rnv23YtWHR*)
(*hvlzQetEhuQ2fLfWacSsxq0/yi+3DQob4xWYkwnGPaOK29CSwfz/fLcVtP7n*)
(*rovG0+Dm9f98rDhRJXVrG3zkTj3DvP/JfPnw7W0QbOh1wNxYZV+idncbSMOU*)
(*+5hlnFxzeR9tw0Xj+5qYKUTfHyme23Bv/AEvZi+fby96k7fhsT9jJFYPdj9l*)
(*Ujf9tQ36y18/YI54YMM9nrYNDYWygZgzNeRj57O2wU7mkgfmca6mcjrCNihK*)
(*ctzFrPGFhlm8ZRtyGHwPYubLvRvstIHi/08qFqtfvyZEItk1d6C6Tuk7Vk/9*)
(*on1M+rV3YLrcIByzuc7skdTrOyDx/Ok7zMKFhelat3YgtSH3OeaAtwaVgWY7*)
(*4M1gYoD5wenwRVbXHXiIs92D+dArXjWm+B3IsDIJwur99ln33R2JOyC6wO2D*)
(*uWt2pDn+5w6Y8Ne5Y353K9NI5fcObKXrWGPeOa7j4FO4A+sOmRr/7xdNb7/s*)
(*atkB1wdtbJjzfCn3Wtp2gEtscxfmj+fvisd27sBG/fFNrP9o/JDMUupH93N8*)
(*ncGc71L/78XkDtw5NlOPOfTEqbfXZ9B4u29UYn4yHKV7cH4HsubLijGLaVoP*)
(*lCzvwHmXojTM9LStce8oO3COUTUR82D2+Ucm6ztAI9H1FXPho1gp6a0dEFx1*)
(*CMf8eT/bys7ODvy12xuC+X9aH0Ia*)
(*"]]}, "Charting`Private`Tag$393242#3"]}, {}}, AspectRatio -> GoldenRatio^(-1), Axes -> {True, True}, AxesLabel -> {"Tt", None}, AxesOrigin -> {Automatic, 0}, DisplayFunction -> Identity, Frame -> {{False, False}, {False, False}}, FrameLabel -> {{None, None}, {None, None}}, FrameTicks -> {{Automatic, Charting`ScaledFrameTicks[{Identity, Identity}]}, {Automatic, Charting`ScaledFrameTicks[{Identity, Identity}]}}, GridLines -> {None, None}, GridLinesStyle -> Directive[GrayLevel[0.5, 0.4]], ImagePadding -> All, ImageSize -> 600, LabelStyle -> {GrayLevel[0], 20}, Method -> {"DefaultBoundaryStyle" -> Automatic, "DefaultGraphicsInteraction" -> {"Version" -> 1.2, "TrackMousePosition" -> {True, False}, "Effects" -> {"Highlight" -> {"ratio" -> 2}, "HighlightPoint" -> {"ratio" -> 2}, "Droplines" -> {"freeformCursorMode" -> True, "placement" -> {"x" -> "All", "y" -> "None"}}}}, "DefaultMeshStyle" -> AbsolutePointSize[6], "ScalingFunctions" -> None, "CoordinatesToolOptions" -> {"DisplayFunction" -> ({(Identity[#]& )[Part[#, 1]], (Identity[#]& )[Part[#, 2]]}& ), "CopiedValueFunction" -> ({(Identity[#]& )[Part[#, 1]], (Identity[#]& )[Part[#, 2]]}& )}}, PlotRange -> {All, All}, PlotRangeClipping -> True, PlotRangePadding -> {{Scaled[0.02], Scaled[0.02]}, {Scaled[0.05], Scaled[0.05]}}, Ticks -> {Automatic, Automatic}],Graphics[{{{}, {}, Annotation[{RGBColor[1, 0, 0], Thickness[Large], Opacity[1.], Line[CompressedData["*)
(*1:eJxTTMoPSmViYGAwAWIQ7ZX64uKBl6l2DGDwwn6BqH9bl+QUexi/kWOzeqfk*)
(*Ejg/InLh+g8ft8D53P8bY9zl9sP5qoHvXhzuPgLnl5+ztJ/x5SScL75F8B4j*)
(*43k4f9sj5zPnbS7B+Yey5kpHvrkC539Vs+Vfm3sdzmf6uVpAWvUWnG/2P2Fb*)
(*8Lo7cP6ZFpNzx7Tuw/nyE17/XPnuAZz/+vv9e1cNHsH5S17tPsXX9xjO55Nd*)
(*/v7Jkydwfk/m1rQFKs/g/G8Oq59k1D+H80uMjmwRO/wCzr+/7kj3RcVXcP7v*)
(*rCuzNatew/najuz/0na/gfPdFymmbxR/hwgv2XilhzHv4XydN6Hc130/wPnC*)
(*/B84jvxC8Fn2z16vPOcjnJ9tsvsmp/cnOH+1Z9KCfd8Q/K+vpQSeTfsM528N*)
(*dmee4vIFzq/3aJ2/5RmCr3jX68bi1q9wvpV5yxRJo29wfozc+Z+MlxH8r6d6*)
(*HxU0fofzZ7DElQSp/EC4/97WRUxHEHzW2sYQy8KfcD7Dyrj1j4R+wflFs9b3*)
(*Ch1C8Pdqd3AI5v2G8wW9inSa+P7A+Xd0L1yu3IPgb5hnr/Ik7i+cH7veTFD2*)
(*H4Lf0Lqw8c6qf3D+2heys/l8/sP5Si0iHuzLEPzp8txf//9H8AFVHxMn*)
(*"]]}, "Charting`Private`Tag$392830#1"], Annotation[{RGBColor[0, 1, 0], Thickness[Large], Opacity[1.], Line[CompressedData["*)
(*1:eJwd2Hc81P8fAHCjhJydUJqSInzLKMVLIjIyKzuEJCJRKhFJVKIIiZKspGzK*)
(*PMKRdXb2yDzCneNwh9/78/vL4/m4z70/7/Ea77Pf3t3YkYWJiWmFlYkJ+6vj*)
(*ONWKn3ZUdd1K39zcxOG3nBP1GBLRAD3JtTjMJz7GxueJmELyVs4BzGO7xQPu*)
(*SdgAk+L6YSYmbrxuU3dugogDLEOcAWbxl2Ipm3tvgKQ/uwlmhs61GDuJW1B1*)
(*VF0VMymoo9ou+jaMOtHEMF9xoJW9EfGCedcPdGz86nOiRb/i78Li4N9+zHIH*)
(*VbKpe+/DvcCL1ZjjWWy/HPr8EJa3SOZiZh8NTLos4Qf6PgYpmIuH25p+mT6G*)
(*1/v63mP2MHTQsYoOANK90EjMEvilWmp3IDAUOZ9jHpB9di5MJAgeJh95hDnq*)
(*ozD+kOVTGD086YJZhyfjTHl8MCQL+xlhZvI//fPy4DMQHnSUxVw416gwvzcU*)
(*Hk3qsWJ2tbHJfWb3HCb6eRs3kA80z8vs+/wCRGIKQjD/UQn4+mPsJRB7LE9i*)
(*fvWNX9JI4hWo3jzZu46sKZacPH09HCxHXrhh9quN4600jQA20/dkBrK539+L*)
(*YUsRwLctwxGzvKJ0mHn0a3jGTa2jI/PMeTUcUnoDu8a+iGImpZRxULrfwOvB*)
(*WfM15BprNu1yn0govNf4dBU5cYdB8HORKJjW1UhYQX7YFFN9uTgKgsqUP9KQ*)
(*Lz8dZjlo+RaGDz1+voz8n8qRs/P0t1Dr32y9hMy1dNu/JD4a7t+t30lFnsws*)
(*LnumEgMehI1CCnKVAyvDZDAGHMcnTpGRfTqifGb3vgOJDd2xf8gmLwcKf+Df*)
(*AeXFFfZZZBkNiaUguziwPq3AQULmYNw6YcTyHgRSmMcnsfjLK7ot9vk9bDX9*)
(*/W4cueImU/b0uXhI6X116C9y3MELcwVj8cDTczZ4GNkostflosQHOP6ZpbwX*)
(*WVr34BdRwgdgEi6L6Mbih9V1cuL6RxhrwMt2IF/9uZFTYZoIW8hR9Y3IejsO*)
(*th5KSYTE6rqKOmTl21oLL5YSIUwXd78aeceRCBmz6E9w+8dJqRJklqf5+uUT*)
(*n2C9T2NvIfLC8B9XcaUk4DyZ2ZiN3BC7/+tCdxLEZDy5kIL8g6r5+7LkZ6io*)
(*7BP4iJxq6DJd6vMZFLNtwmKRA9jzDoeKJIPGB4snz5GPar0OyfVIBoO7QqaB*)
(*yG1P3af7CMkQvZ/64z6yOKv0V5m7KdBTMSZ3HbnxLCeXWVMKcHGWnbFG9n48*)
(*5RogngqjFLVmY+Ta9WSZjrZUWO6Rv3MG2f3Mk/D1I2mgX0Lxk0MWfmi3IBGQ*)
(*Bsv6NCZx5BsrYrn3ZdOBraD8Cgd2/ovRCntffgH+zM2+xg0cvuA/72jtv1+g*)
(*xnCPRCmyjYcJ7bZyBlBrEqozkLP+8fysnsoA++NPZp8im0nPiczBV0iL+Ojq*)
(*icx0s/HBzpivcCglVMYG2Xgq5IyLZiYYs3k7nECmS1xPiIzPhEv+Sc27kZMd*)
(*NTdKFzPhr+ay01bkpVFmPE/SN0hXzsJ1rOPwsQMPzhVsfIdEyt5PTshnd5sn*)
(*D17KguMDWSRtZJKF0lb2b1nwOmVN6Sjyntn7174dywbBZivCNAOH99w+LaBu*)
(*ng2533dtEJBrj5pVdwVlw7nMsf0pyO43FA8x9WfD0B4VESvk6pCUzqhtOSAn*)
(*Tx9WQBZOFww+ciIH6I37/LiR8ROUCePnOfAryGpfKR2H53fISk85mQtKNzyv*)
(*sCM7PdljruyQC9aXssX+rOHwJUlhHC3huWAbofMsFdlh5KbLykQukL95HwXk*)
(*AhtJad23eaB6KGzZZhWH5/SL6R/C58EDJbrYQWSbBLYwr9k8SGplzhtfweHZ*)
(*+sf+JZzLh3CtDg4nZDPzT1kL5Hxwrd5MMqbh8PN9ppxJogUgJ81ixo0cbMPu*)
(*aHKuADpLLHzrlnH4vGvuooVRBbBVOHlNCZn7lsrTh0qF8Mzr317WJRw+ZWFh*)
(*SNq2EBbC9ZNyqDj8mTvJyoMhhbCFY9zOBtnFZ/uCWm8hfP9W4Je7iOpRYI8F*)
(*26MiyNg36K1BweGttoYVFKUUgfwz0yNjZBye8kyN90ZzEUj1fBoLQN4XllbT*)
(*sPcHbPgIXClawOF9Y7zlXlf9gIX6BCeOeRz+eCYf2272YtBJ2j54bgaHr5ep*)
(*sW2SK4b7Jg/y8SQc3jbHp8TPvBh+dg+KnUYOKxzyGMkoho+Kt59LTuPwU/hv*)
(*fan6JRBiwzU3NoHD+5+zUzS7WwKBYn3++shCtYKvOT6WgMuLbY554zj8uYaH*)
(*510XSkD4Vgr7vTEc/kPnhZz/IkvBqUyksX8Eh1cwW9/+t6QUiJmWvLLIjb3Z*)
(*TlFjpXAjesLBbxiHXx3auZumUAaSFWq7BYdweFPSeHDpnzJ4MkRoFu3H4bdv*)
(*Blhp7akAVrP8G2GdKJ5vXLGRPl8BTyonVrI7UL53SNvy3aqAqvG7R1vbcXjV*)
(*L93X+soq4MTpz4bb2nD4n1+iC4pk8PBiJefLhWYc/p/JbKluMx6+Hjt8bU8t*)
(*6h8c3Sqqo3jolSnCb6vB4R+UV5bLLeMhQbv9yL9fOPy2IzH4HXsq4fYEaflr*)
(*JQ6/f+Ns9ZBbJdjxVdmvlOLwl9JjGu7gqoA/PejRr1zUn60D9Zz2V0Hpypt6*)
(*5xwcfpTfrclMoQr6tGoecGSjfHmk3qJiXQV1QrnDZ77h8KHG/9q2fasC0Z7E*)
(*Jy5pOHwZQ733vd4vYArdf6I7Dt1Hco9ZvrL9BcVqoiHH3qF+fF24/7HXL/hm*)
(*8+a5XwwO3936b8Ax/hc4Q9R13igcnpwaOyI7+wsuFRQfpbzE4Q8ZzU39elEN*)
(*L3ha+5Me4fCi/lrK1z9Vg+C3hQOeD3F43m+JLziLqmHovayi8n0cfo3dRMZo*)
(*tBp6Ywf9871weCK+6M7gyRqQOO5fanAT7Y9cwMbKeA14Epts+K6g9dj0GsTT*)
(*a+Bok6WKoynKz5cnPgFfLah3czRkG+HwBlPj54LP1MLGl4PpMnrovBJ1QgUi*)
(*ayHsJ8PwlxoO38wjKCijRoC3Zp/+A0kc/peKm2PrJQKwnDv4d8sh1B9u1hZ6*)
(*3SRA4dsMtsr9OPxngo95STQBPuWsGezehfqp/8AH7X8ECFnyeGeFQ/eN+dQj*)
(*1+LqYOeqCit5ngu/U4zpIVt2HXBlJBQLzHLhuXTNG7/U1AEPc2+Q1BQXnpa6*)
(*3X1hoQ4S/h5bOjXMhW+08ch/pFUPktrreS0tXPh7zacglloPcmmLn+szufDc*)
(*MgXj9ey/oV27SiYgnQufHCb3kr77N/zn+WvsWDIXvlXvcI+N5m/4XKvu7/Se*)
(*Cy/VIHjnUPRvWLaY328cyoUfIsyn5io1wKov1YzpGhf+7uGb+mO6DeB07fl7*)
(*Jhs0v2cTiztsG+DJfciimnHhT50fULsf0gAjApbzRRe58JHVv3uhpwHSWZcM*)
(*009x4bUqU7ibHjTClya/o7txXPjBffsLNl41QtKRnI8Xt3HhvR7HW8p9boQL*)
(*rwLd7jBz4T9BZHpkQyP0FpRlRy5tx9PLHqtb7G6C6H9H9koPbMdnFVvenSxr*)
(*AhdJ12epX7bjd+bzDTIzt8CZ98fXOZW2461e0k3HhFqg8P3wWLbMdnySw3hD*)
(*rXQL6H9TGFeX2I6X3vHz5wuzFvgsf+DT0R3b8ep3bd/uyG6BoaM7hbdSOPGK*)
(*8VdSo2SJYOZ7XLY2jRNvHtI48lSZCB/2hKfv/sCJ9/U6K3ZPkwiquP8oVlGc*)
(*+Eo9qShzSyL0rYfYxj3mxOutbwTseUYEWmL6yy1mnHg7m1Tr9EEi/NnxfN2L*)
(*hRMfpLPr3btpIpzHl9ZepnHgUxUjOp5TiXCjYp1p3ywHfpb7ga4bZyvI5vPJ*)
(*Ondy4O9W6J86rtgK927xqDSkcuBf7lsSLA1rhenPa3eLNTjwhaPnGlvOtIGg*)
(*tavff67s+C7tAHKYVhtoBzYL0q3Z8cvfy4X0jNtA+n2Ua7oBO17p4Sm7uutt*)
(*UBZ1t+vbcXb8DwHZZfybNkhoOPrnKm0b/qeG6L7cqTYoSi+xOPdwG740deFO*)
(*VFQ71Eyw3JNyZcP3c8m8M/7YDj/E+JJlLNjwDM+b5bwZ7bDUbFfCp82GV4EJ*)
(*9rCKdrC53KRx7yAbvry7L+EpqR0+ujTsK/uzFV/BTiDcVesAn9XQ8H0qW/FV*)
(*Lgm7zGc7ILPi5ve6WVZ83vTNmzbLHdDey7m47w8rPtlZueQaUyeMj7Q/tfzF*)
(*in/q1G1xS7AT2uwPBd97x4rXtueLe3KmExhyh9J0NFjxLeZPhb+/7IQOvhV1*)
(*6ygWfL+22w7WY11w+6uQhdkeZvyShArPN7dueK7rCPRieoXqqDrhuU83rL/+*)
(*86LHj14RnKDt7/ykG7wVTvn7qdMrhARN5w++6wZJR+27e+rXKhSZb7bEVXdD*)
(*eyUnb237aoVPX2x4yK4/UK3QrH9qiFbBCKfyONT/AePjetm+7YsVLKtfeXcd*)
(*6gW+rJVXkb8mKnY1qS+CbC+MVLzc0nVuokL+U0+nw6leGEjvTdpZPV7heGHb*)
(*++/6vbA7yu9iYvVYRd07+0Pqd3tB0cw0e7NmtCJcWUT5BqEXBndwTwTGDFTs*)
(*9g2+VnSjD/YFxpbq1TVUKBiKne+/0wfSqTGtzD31FRfF8yWZ/fqgOSq1upJE*)
(*qPBvHPmn86YP1kSOHH/KV10xulvFZ7C4D1zjdx4S6/1R8aVs8eVWrn7I7lNY*)
(*LRqJAsVN20KT7/0QOfv9kZhXC+jp4wM1f/SDxpRMh0EUEa6932ugVNUPubzF*)
(*7q/cWiFCaXBStKsf4kqlLJd2t8O0u6XI6Ho/4P9wXGV174L44Uu+HvoDED8R*)
(*6cMygL4nU6Btf2UAglU6XV1TBqDOV3CHqd0AdMRqFb93HYQl4fZvSt4DkK5m*)
(*cFeFNgQXjQyHNuIHwKvDJSzObwSYq3TUw2YHYGI49MkXhzEQ5s3g9l8egLaj*)
(*z34alY+BjA1HnwfTIAS/sufW3DkOFmuEO6aCgxC9dml+G2Ec8o9rpuw6Mwi5*)
(*Tqfa+8QmwekzsGe8GATDi3Sd4rRp8CV/6Hj/dhCuXRx8E7w0DZGwkRj2cRDA*)
(*I+HoS3US4PtKT93OG4T2a3sKoZcEIoLKN0/2DcJz1vmvckyz0Bgk31x7dAh6*)
(*lHdORB6aA2WeBpbXCkNgdGu78A7XOUh/Z6dkqTYEvekxBz/mzMGT768S5y8N*)
(*QUuki3vDqXlQ/jPlKRwwBHcNZHRP712AdHv/tNEXQ3DyeyBbrOoCCP3b0Z8Z*)
(*PQRMFfyX71kvAJn5nObZzCHw/PfnQNk79LxUwk6XbuQT5e/tucggVHhCT350*)
(*CJhJZzeJh8kQpPb78cbsEHAFTV2pVCdDdgbd6wLzMEz5X146fI8MLyokBweE*)
(*hoFSpHmeu5cMVVu8VtR3D8NRo7vRYWQyrFyo4E/fPwwhObq3w9kp4NRxSctT*)
(*ehj+nJG7uUuBAurTgdls6mg8y5558xAK3Jdp/n1Taxjyl7WplPcUyL4jMk7U*)
(*G4b0vhp13iwK7NnIEnl/ZRhIewiUP+0UuHSOfoLJehio+I+rQeMUeBly/qKj*)
(*/TA4W4q+LVymwCr/QKCs2zD4ZOawZu5cBDkzyYQoz2HQLj3ywePwIjgl3Cla*)
(*vTcME3ZjlqWKi9BxmHP2V+AwiDfh1kkmi5B9RumSWewwLHzj7MkJWYTJgED3*)
(*soRhmL0SeKc0ahH2EJpCD3xGnxea8RgnLsJLI4fy2W9oPsb2gfsKF+FXdNYf*)
(*o7xhaHug5meBX4S1vjVK4Y9h2BVdVL3/9yI4O70+HPBrGPA37T9Z9C/Cx6/9*)
(*ZyfqhqGeD+fTNLYIXQuHrXSbh2ElYL8ocXYRcIp37ma3D0Nm7k65a9RF0HhY*)
(*HrGjZxgCLWMdn9EX4SGe4+uDwWHgnHO7ocJChdytl2qG/g4DztX731N2KuyN*)
(*mFn9MjcMb46E67UKUKHczv9L7uIwXPar3QwUpoLNCQHzkpVhCPY82t29mwrr*)
(*W9LYq9eHoep0xS3iPirEdyn/aGQZgVcisQ5u4lQ4nd58vXPbCNwWaTTLOkyF*)
(*3vv2Owe5RuBEg8u/6KNUuK+7XDvBNwIJUx614seoICz2/O680Aj4UktDzWWp*)
(*UDQnJrGyawQUG1iHFP6jwmV8TifT/hEwka1xLz5OheXXmk85JEbA6ILa5OQJ*)
(*Kry91iPPLzUCfw46U8rlqSCv4DYmKjcCPfv+gJoCFTrYmKMOKoxAu3Odvyvy*)
(*nT9R56SVR+CqOYuDBjJ/huSiPIwAYbbjYx36fs7D0iQVjRHQF+eYWkXjG+ob*)
(*Gp+/MAInb3VR/qD3L+wZYza4OAI3uLvvOaH5hS/cy7liMgKdzxiyn9H8Zaq2*)
(*29majQDvtqDOcLS+5siPvDesR2CtjXTwmBQV3BxP4G/bj8D+8tHuh5JU4FIi*)
(*uD+4juafeD7L9xAVvrJb7n3iiupUN/cNuQNU0Omda35xewRUajVrYvZQYfpr*)
(*oF/UXbR+V0ZQnigVQh4JySQ8HAHOJVWXICEqHDbIGEh5PAKRVhwnt/NToXaf*)
(*atj3pyMQL3ImVxtHha3VjrMV4SPQIhteQULxkPx29X1d1AjUvQoYNFxH8XM9*)
(*TLf13Qjc2rxR6UFbhEDOgozRzyOwLHWfMkxahP392hYz6SNwZr6r9TiKR/y3*)
(*fg7qtxFIYnoSc3ZgETYMt9zY+mMElK04N2KbFyHhQKwwd9kIyH8Tl/9buwhn*)
(*qFJ1QlUjkC62/OBfOYrPGJPDko0jIOGmckPz+yKsDCSN6wyj/YPTu0WeL0JM*)
(*luJbk/ERKH9y4ciDx4ugGPBbw4o0AgdGnn/5dHcRvMUpn92oI8AkNaWoZb8I*)
(*FJez9hHsoyAnNzHForQIM7ShwU65Ueg/fOv0fBcFCq+HV+kojoIvQ5efuZ4C*)
(*j7tVUytOj0Lu6X8XB4spIFSU4Pbl/CgQWdaoGwmoHt21ZPhajUL68udUqh0F*)
(*4qhdIuIho1CrnI57OUIGbXKTqefQKOzI6o7YrFmAZFLJb+5Xf2FbD8VqM3UW*)
(*bsUYHHSJ/AuWT0pLYv1nQUnj78Oa2L/wKFv+239ms/A7gfOY7+e/8NOR8NWJ*)
(*YxbIhubh0z/+QioXg752cwZUi5ZNqv/+Bf/2Z+lZx0nQ8+S/gfunxsBUrk0v*)
(*vnESkuRqFDpVx+DBNeNPzGmT4Dpg9kpOYww0iHLRgQGTwKT0GCYvjgFj3XWp*)
(*R3ESJEjNn0xR39s6RBqaT5oAL0NXJ9nwMYg7+z3oZMA4cIulzY+NjcHLoe8B*)
(*CeZ/Qe3yk4G/02MwIXLqT4HyX/AMv9owOjcGgZptfIu7/kIXs3Da8MoY/JZe*)
(*ER4YHIUPEyE2/VzjMCRR+NDbaRRksl2a2uXHgUrcNz3ii/LqnMzXqqBxOLnl*)
(*tq9u+BD4+XK+qwwdB2P7lcUrlkOQXTARjH81DuVnb0cWHh4CQcmP9uWx4yC8*)
(*K+OtS+Ug9HPxihZnjgOr9wW5rJUBcO0ih+R0jMPxyJpHWXf74eWNAqdE8QkQ*)
(*7j4jr5ndAyS1byHbj07Aedopd+4nPaAlnJJxV3YC8NNeVxUv9wALIWpOT3kC*)
(*DgS3BH9a/wP3xb3urhpMQLPj+XeDhn/Aaeh4sPHDCUimq5wvYe+G6sKj6WWP*)
(*J8B2TaHAeKQLDrw68FsyeALIh8wzNYq7YOAMP/dmxARICytQFN26wDSO/PZr*)
(*6gTcog416XZ3wtlL2Slb2iYg/fo1L1JJB3yQTie4d6Pn7dT0WOI6gMGaON3b*)
(*PwFNPHsOuPt0wI/c8GM5kxOQeSZG/bxiB8jwuhdYr0/ABwPVp0MF7bC74VhN*)
(*oeQk1BxzkXesboMHSYcm98tMAs6S0ZCQ1gbd98U4wk5Mwo4wdob4izZ4I4nT*)
(*v6Y6Cd/3PvYRM2kDjuDZDh7TSRBRLuX9OtEKy2pfx5z9J2EorB3Xs6sVSpNn*)
(*3t8LmoQ5ngJ7262tEMgubRIcOglEnTOqEvNE4CZmVn6OnISmbRx7XH8RQeLq*)
(*9w+DKA67izXDum4R4fKjHDNTIjJhpDrzUwvsHiHzXOucBJM0lxFF3xYY1ThO*)
(*uN07CcXSz1KPX2mBW1x5CuFjk6Cklmbky90Cwe/z+X+vTAKdU15TI7AZCn4W*)
(*NcL+KUg90MrB7tcED8RWnlyUmAJNtrpPV642gVrAydPWUsjxlm9xZ5ugUfvn*)
(*lwcKU6D19HnNl61NMNZdHFxwYQpSZm73hEc2guBymZqU5xTYhj5/61PZAL3m*)
(*G7RT96ZgNlkC15feAIllqlnavlPAneg3GRvRANJPK8ScgqfgYRerkLhdA2gI*)
(*Vq4lxk1B3/ewXEe2BvA6Xp0v9GsKXCyYbrHY/QYn+X2Jr+umoMz83SV5vd9g*)
(*puj7gqt5CiwP1EjVKP2G08ry9iw9U7CnI96Zifc3sKgn88zNofnsVy6ora4H*)
(*6jkmujN1Ci6zh8X35NbDuKbVxN/VKeBkUePX/FQPdRcEy3q2TMNhEfXQA/71*)
(*EG4U5FIjOg0dj0zoI6r18Nhk+JLavmnoPL8z5bNsPXheOnO25NA0qm8ZBUX7*)
(*6uGyOXVnjtw06Dd3snew1sNuO4ea+PPT8O5DSlxdUx1wX8Nn79SbhvLrjTHs*)
(*lXXA5Lg7/o3RNJx9p8XzNr8O/jp3eIZYTUP40ExIRHwdZHho7PPynIYe//8+*)
(*rLvXQbxn4vb5e9PA9i3BJsyxDsK8GMs3Hk2D9NkP09cs68DDJ7/pasg02BIq*)
(*3o1q1YH9A96fvWHTgPvoaxOqWgemvq7JlyKnQe4o2zEvhTpQeiz+UPfDNEw1*)
(*Wf0nIF4HkoGPnWo/T4P1RhB37a46EA3qNzr7ZRokDOU0igXqYONZlKRS/jQM*)
(*3/Lkd9hSB9XhW7oO/J4G113fRjOmCFD42rYyoWUaRPdIHbo8SoD0yNJM4c5p*)
(*GHSVeqnST4AXMV5PuIenoYq+R62cSADfd8RboeNoffl5upcbCHDrvbTFlplp*)
(*OKWJWz1cSwDjj2Nya+jerzpAwXuXEkDjk9pub/o0+DD1flooIoDC5/htC0wk*)
(*+PPxdE5SHgGE00wHx7eTQDn4J+XLVwJwfsmut+UjQfe1Ecp6GgEYGVwFfUIk*)
(*iJWNKwtOJsDQ9+oXrftJcEHO0lz+AwGI/7bMHj9MAvmEvBGH9wTAS2vqRUmT*)
(*wE72Dm9tLAGybz7NXPqPBFuUv1pciSZAYkYN1xUlEljqXlQUjiJAxPRWtx9n*)
(*SPDqp6wo+xsCBEiebxJBv1OspFn9JCII4Hk9+NhDLRLoHbTq9XxFAPvU2rB+*)
(*PRK88fmWPvsSrXecbU7FmARVw9rXIl8Q4Jy41sWPV0hw7brNU6fnBJC/9uw7*)
(*kzUJoqfozx1DCSCeROC2tyeB7zMucngIAQRHtrn/uk6Ctf1mCuPPCLB1n3aL*)
(*uBsJ4isDuxyQl21CZIM9SVDIwWXKjTyRUBc+eY8Es0kXZUaCCdDVz76g/YgE*)
(*wsn+E/3IhF0XDDMCSdCx8m2DCT3/wyI0e3sICbRrKlf1kNPf1fO6hZGgjfDQ*)
(*Eo8c+4fjdvMbEhj98Eq3QPMJ3anTKhtLApmU3U670HwfXH7+3+sEEpy6WCzK*)
(*gtbj8vb3a0oSCdwfuehzo/VadHBSTNJJUG29NxjQfugK6BoXfCNBQctH/bdh*)
(*BDhj/CJXKI8EO0qDj/GEE0D6dQO/zw8SMBzUszPR/u4mbr/TU4b6er6AnRva*)
(*fy4evXblXyTIfoFrNETn8+9lY+R6EwkSBvu2+KHzHGzgotq0k+DMs+m52jgC*)
(*NHPqm+L/kGDhQ0SDXAIBvoc0CT75S4LEVKYPjkkE+EDAeY9NkUC8yLtQNoUA*)
(*4WwXOzXnSKBO72UTSSeAR1DzW/ZVEpzf9jpB6zsBbH9xL9/YQOOJTXO9zCGA*)
(*EYvB5QZWdA/xV2WbyyfA8cctQq+4Z8B4NYLMjuKb+pAYIyA+A1uN9hmFo3wY*)
(*K+Zd8ToyA3hlmZz8FgJ0rhqadcnMwJxOrgO1HeXbvVbhd6dmQGHK9lQlyi+f*)
(*O23v9hjMgGeAPunpPAHoLh3vjzyYgWXrrXbOIij/Fa1fy/jPgJk3Eyl5Tx1k*)
(*Mk8EnwiaAVESv9biwTqwiaXdVnk1A3v6cHbVMnVQWSOqbZQ0A0ze7DflNOvg*)
(*2T47qs/vGRCJstGgetWBQNc/vTrRWQj8elvxxmAduH66q960dxYKfdvO2I6j*)
(*+uC6qdQmju5hlGxpu9k6uMvKd7BfZhZ2me+Z9V+rgx45+dV59Vl4qDa9enRn*)
(*PXx88SBl581ZCLX6qs9hXA9SZ7dtOJXMQi53Z25YYz1oJe2/nrNzDtxv1t/q*)
(*IjeAfCy3RO3uOTD8yDcQudoA+17Rx3r3z4HUY8NVPeZGWLvfab9Feg763oj+*)
(*y+FthG9GoTZX1OZAV4z3q4ZsIwgwky+tO89B17jS10c3G2HoKl7jQvEcrB9d*)
(*ed041AiNl76x2lTMQd7ih6TBiUb4oRtX5Vk9B28vnHQh/WuEN0peavHNc0Ae*)
(*uT4/R28ETR7JM/Ojc6DouPE2Y2cTfC0PP/F2+zx4yN478le/Ce6JXT0wYjUP*)
(*S3bEGtGCJrA3lLvcbzcPs6kvfv9X0gQXnzA/73aah9Yh29KzlU0gMf2Z3OQx*)
(*D9LvK431mpqgI38KXxw0D6TJ9art400gp+d5NSpzHlTpH5ZlBZph6sHTeC3G*)
(*PDBX5N2i3WgG0mUp509MC3DubdWMuHsz2C4IsOxlXYDn5pwrOl7NoHtwXEGU*)
(*fQG28jmUPPZrhgOhzxJ4+RdgW0nh0Js3zUA0bXRdP7QA7PfevzH62QzSs5e2*)
(*d+svAOHeoXEV5hb4YapX8dwQPc85stqztQXOlanfUTVZgJ9XmDzcOFvA4pVs*)
(*X4rZAlx38wl7INACoXIcX72uLQCfhfz8sUMtMOFVqsN/fwGal38UM863gOdA*)
(*7kbNwwUwHuT7wqvbAuuaX3Lv+y3AqOvAqd0GLSC4M1p05MkCtOt9vSyK7jfn*)
(*frqTsl4tgMm1u7g8pxZIXD/wXD8ZvX/VRsUyqAWknURUmdMW4MSOzq6ZZy1Q*)
(*1MxDyf+yAC9rH7h6vWiB5o8M891ZCxDWmZDi9ga972zXkZmfaH0XTLcwJ6L5*)
(*B4fWh7QswNgk6VLOzxYQ4Flgr15D81//F+sy2QJtZc+fXVhfAPMTgtl90y3w*)
(*xvUQW8vmAhzAu0efn20B/t8WrL1byMB7SWyGg4z8tGZ9npsMkYIXbmmttQAv*)
(*PY6y6yAZ8O4jw61biUD8onD70yEySEoY+/SyESHCjDgvIUmGKNuXx4a3EYGn*)
(*cOu//46RwW1Sf3OUA90Xb3tMaimRQe7hrqRaHBG4JjX77+iSwfucG4N9BxEa*)
(*3w5brOqTgaYqvHMc+aXGwx4/QzJ8pZ8OKRciwvak7K7QS2TwvWXie0OYCJzW*)
(*u1o/XiVDL38jb+IuIrC3L9Q03CHDEy4bocj9RKgLeKFhdJcMY8rvdc4fIEKI*)
(*nMSvLh8yFIaY8NCQt72yxI88IkNX0W8zI3EisF2oLaY9I4NGDtloWoIIWyre*)
(*Zx2MJ4PhEx7uDCki9J7R0B76QIbxf7cbzkkTIat4djjuExlMdTXIvcgWhSr8*)
(*/Glk0MP9DGGWIULOtyEvplwyyEvLOMvJESFYKgRXmk+G0T5bUgGy1Re51HtF*)
(*ZLDJn1o9+R+af0pA91wpGY6sGrcrHSeCTfxB5QECGdhwxfPC8kSQF21si/1N*)
(*BqZggtULZI5Yr5umTWTY873wIR05P7ImvqGNDJee2Ix0KKD9enF98+cAGYiy*)
(*dr9dlYgwzM4b6z1MhhFBIc0m5MLgH3L//SVDc275W6mTRLAP5LBPnyJDxkrd*)
(*7xHkH/e/VkcvkuHRhfNMCspEeEUzsTZeJoMf41haEPI1b8YSbpUM+TqBWm3K*)
(*2PnqH366QYY3katPr58mwtjcUoUaMwVUS3uks5CLXT+YMVgpUFGVOEhFdnSe*)
(*D73DQQG71jNeD84QQXki5oAsFwWaWwtti5F5HdRKSNwUKAuxu7GCXHr19ayd*)
(*IAUCWD2Gb6kQgf/KCQPDvRSoSqjr11QlwlRH3+T2AxTQXlU4cA+5zDjoMUGc*)
(*AoeLfqekIjtf7MxRPUqB/t6DPkxABNXGRzpr0hQoop9uOYosoCPxt0CWAhdH*)
(*DnibIJdr3hM8pkABpn99ZR+Ro37t/TalRAHOw2POv5BvnK3TTFamgOvpw3fG*)
(*kQVVRO7tUqPAOFNigbgaEUjFlTzd6hTIXptdOouMP+mS/kaTAtYGn1OskW/K*)
(*l/Rw6FKAl93gWgT2ObEz84c+BR45N/GmIe9wW/C/bkgB5gJf3VJkF47tJkIm*)
(*FHDXEN5JRK5IOSRRc4kCR2Ou3BtFFlRXW71jRoE3hnPui8g3Bi0aD1hS4K3c*)
(*yy2sZ9F6Hnh/bLVG+40bOMmHLLAzwvOxLQXONIXw7kF2zsvQlL1GgSDbE+FH*)
(*kMsMaoQHHSnQVOL74wQy/+zQzEtntJ9tTBFnkK+HrJWfvomeD78iooFcKr7j*)
(*DcmNAlrtgmY6yHyVso7vPCgw51JnYIDsZK1zUvsO2h+uDVYT5JJVh+00bwrc*)
(*N1HxuYTMG+0/mOKDni/iyL6M7Hg8Lsf0IQUKwka/Yy5uzg9i9aMAYV39/8/z*)
(*3Gy5kvuYAntYkniw8Ry2kY7aPaGAvp2nD/a+n5+3bPAEU0Az40AhNh9utb2t*)
(*5SEUkJQTa8Tme63/VLLbCwrosdWVqSD/8DG9t/sVBUwZQa8UkHE73HUaIihw*)
(*Oy7nrDSyfU6o2INICvwQS+vdj1yEirdkNAU+ehVbCSFzkcp/dcdSIPb0mQYO*)
(*ZLvgnujg9xSo/elzhIHOY3sFt8pYIgUmlVWqBpBtLY/wRn6mQP5WbdYm5ALa*)
(*ub9nUymQxidwtgT5qtz90I9fKXBKo6IiCjm/MdLq4ncKPBNg2vYYmePGd9n1*)
(*bAo89sBfcUHO+/S306KQAhp+Fw+cQWZX3Ujn+Inyw8Us6SCyda+w748SCuiY*)
(*WZzkRN4mcPGAUCUFjmSG5Xag+LUM+nGrtYECa5Sv8leRs/e1n33cTIGM9lc/*)
(*AXlr2T9B2VYKHDgS5LoXOWvpQMnLLhSfXDbO/Si/WK+HbdMeoYAhRw+7PvLX*)
(*C/aJ5csUaC10cdqO8lVeer8qrFKg98BWvr8on8u5h/sq6BQY0u+UwvKd2G69*)
(*s5JpERZWsoqdkKnW5q9+cS5CZ1q5z09UL1Q9DX3r9ixCis2L81qo3hDjVM3a*)
(*zi/CZknqCh+qZ+aP1peMLyyCS01VZJ8iEUavlka26y7Cwwb7gWRkqrhyS4fh*)
(*IiRF7nyiiCySpXC+22IRbsyuaF1G9dL+l7R8/61FUGtlORp6gghLM6K8E9GL*)
(*cOwb6SRBFp2PiRZrlAYVnjwPYRnag/qJ7o+nu7WooGlUyZWEbHXuCHvKBSpU*)
(*LwgqOyKznNiOK7xIhe6IGKZZMSJc5G8R+mNGBcX7ly7Sd6N6Rbx8dLcbFTjv*)
(*p7nvR/1L9KKTUfJbKnznueUevpMI/zS726VjqRBV+XTuCjJeRftyQRwV3hvO*)
(*JO5Ddjp21LL2IxVS6vZ8ykX9MY9rzmHqCxVOOCuK9KD+qdfg7SNdToVQse/H*)
(*5QSJ4K/99GP+BBWGU1dYOfiI0GbjbCU3TQXC7xSfHl4iiHvrimTOUKFSZUEi*)
(*A/n3J77IzwtUuHRxXU4PWZCeEPxmjQoBp3k23/AQIf1boZsH9xLYXn/xUYob*)
(*nQf/1GlpxSXQldV7G7qdiOKkYSXt5BI4qdy9Y4fsDd8LDp5eAuYnk2GnkEVd*)
(*vWR3qS1Bv1HGPRInyt8apoOcOksws6N6jwHyyj0RzinrJSjkx/84iO4P+wd0*)
(*/nx+ugQXqZV/VtF943nV6axLIUvgms3n3oVMSZMO3vZiCUbWf+rkIVd5csvf*)
(*jFgCri8xrW7IDuxt4cffL8ExFpLIBLq/pJ8w16rMXoI9J7i8/25B9UdEZ8+d*)
(*vCXg3HfdsBr5/obyknjhEhjex99KQdap3/05pGQJaMwld5yRZ2yGNwxql0Cy*)
(*rewdmZUIMs+vFw72LcGJx2JSvMjR7mZhEYNL4Ng4sHeJhQibphcc1EeWQKD1*)
(*kWkvMnGvFH/qxBKob/X3TUG+XTDn5kZeAuMY00pV5Pxh70MMtmW4xMY648uM*)
(*+qNSUJTI8WVQ2x66XYKJCF32IooJ8ssg+vBkhwDynVffu/cpLcOOxrwmZuTM*)
(*8T+ikmeWIdHDKrl0swX2RB37pHh+GdUBHsUTyCzkrm8mFsugJVFF/W+jBT7s*)
(*djXotlqG/HRrFwHk09rMZIury5DG6qy8tN4CXh+l5O0dlmHYnDr5E3lC/3Gx*)
(*x61lmOHkctVCbsg4QggLXIbX9q+zPRkt4NxV7sz3dBmaU4remiNvYTHlfPts*)
(*Gd5ojjapIaua++nHv1wGxz4HBV7k7G3t7RnRy6DhxGjPpbeA3glnb5l3y1Bk*)
(*flAjHnnaZl0o9/0yfDoZxhqMfKDwsMXPxGUwZNG+ZYEc5eA7TMhYhrlf0hNs*)
(*yHIRfIG635aBbqX6i4Lur00lqQdbsrDvt8wMIrMJtDp15S+Dq1iY2A9kn8pD*)
(*/8bKl0G13XDZDdlqT8sqczua351pgcnVFuAW8Swo6UTrWzx5sAu5UmDHbe8/*)
(*y6Bz+3pgDbIEh+XU1MAyjJuxyqQgzy9OdLVMLaP7dPYRZ+SkudA3oTPLIEh1*)
(*TzNHNp2Wvnhubhk8lc08dJF/DHrWFC4ug95pvRFZ5MD6jbyEjWVoUfSSZay0*)
(*gHz1J/crzDR4e+vI8jzyZLmGFN8WGlxpt2MdQ9bLf54UxEGDbAOe+UbkHR+F*)
(*Xt8UpMGQ12X6J+S6dz/1Du2kgYfi7okY5AdRVuxDIjTAve/gf4U8GJrkb7yX*)
(*Bjdw/bIPkdO9ZG6dOkqDKIFIZzNkC/fWI4vSNEisPmliiIxz8RrPlKVB+y7z*)
(*h9rInleLrfYp0CC8tDPgFLK4hbVwrxINEvaJXD2O3GXK1BGpTIMc+5R7Usin*)
(*dc7rblOjQUHWTbM9yHMa02yV6jTwehCxVxg5EV5WPdCkwf2U9r38yFsU2k7N*)
(*6dDgOjG8ng25UNZ7KU0fze8q7TYz8o2jwjl2hjRg9rAwYtDQ76W9NpKdl2jw*)
(*N2Epl4L8WJR57JUZDXjO7pOfQz6xI/mjtiUNpO5IkaaRYzlJQqW2NHCRbqaM*)
(*IOtsDWvzvkaDZBMrjUHk9U3ZV7JONODS+1HXi2y/5L31800aGJQ0Xu9AFlwQ*)
(*rrS6RYNo3OOAVmQCqcRX6DYNaqeXG5qRH4zbnCTeoUHFzuzzjcjSw8zU0Lto*)
(*fTNHyfXIg73JWefu06Bfc/I3ATmiU+vm+kO0328d22uQ1YkkiSI/Goj8keCs*)
(*Rl76HTbqEUADjrjMW1XI/ds7nRlPaPAi49LOSuQqvd3zz4JpsH+af7IC+UvY*)
(*NW+BUBo0FvkNlGPjN2fQP7ygQfPkibUy5Ls8lICjr2ggU9x0BrO14Sn2wgga*)
(*yA9PJpcia7x+/OpsJA0mHu45jvloG0Gw6S0NHuzjHitB5hXgeW8WSwNim0kx*)
(*ZprJ5f1jcSj+OMLyMQ9GJaS5J9DgQNaVNszVnWPH6B9psG2vsQA2XoaQdH5w*)
(*Eg2Mj+y6h/n1lTvK/Ck0COIzXsfsE1uMT0ijgUp4dAo2P5seZq0jGTSIaw5x*)
(*x9ajIXqhKT+TBq896i2w9UpZRpioZdFA58WqEx6ZP767pyEHxVdGcSS2Pyv9*)
(*e2yv5NPgfcv3YWz/BsWcJkYLaXDX00cf298am2+ut37S4GRAcQ+2/2+GT99/*)
(*Wo7m90/BADsvn/1PNnkracD2oEipAfmq/e+n8b9oMNB6VrUJO98xszd5dTTY*)
(*4cSWjMUD/6FEYWiggXfESxYsXlYdJz/8bkL5Zp/3qAu5dtL760gbDUq3Pa7r*)
(*Q86ULPvPrZMGm6+8Y7D4i7yx5cdKN4pvia8BWHzazryp5hlA+3HgYeok8tpc*)
(*1oDKJA3IRhckl5CHZWnX6qdpUMPaTVjB4tFDlWQ6S4MnTmv+WH5EURqXb5Jp*)
(*IH6w8OQWlD+yy9M8cXQamEVs+yWILKT0X/ShDRpceKa5RwSZcc9ndw7TChA9*)
(*wsLFkH+vsh2p27oC8VYbuRLIDusH1Zd5VuCMbhcVy//YLTZexuIrsMTyeL8D*)
(*8sa1nQqtEiuovjIn3MCe/0VcMjiyAsl5h4+4I8sFnrunL7MCETzGng+werUp*)
(*+UD75Aqw9vTiXyPT1hb9VfVWwPYdfccv5Mvk0BdHvVYg991VsgyqlwKDBUks*)
(*1Suoj/6x2Ivq99ZHAaanbq/CxcIea6zf5JNnVyvurILDy4AlLWQHJ7OP5++u*)
(*wtCNiyEGyNUGsiSTh6tQvbn9ow1y0IEB/1tPV2Fn4s5wX6x/1Z38+vndKoxx*)
(*XXUpQmYRWGDmrlqFCr2FJEnU/3KfWaZFVa/ClqAef1lke0at3i7CKtTc576s*)
(*iFw1nhAj2bgKnkNyc+eQA3/oHjvXtQrKHScbbZCZbdLMfEirYPFPSCkSmemL*)
(*TdYo/xos5215vYxcO9t96+OONdibzMq2jvxSzkjGSngN3EPYPVlRv97549y3*)
(*TrE1KLbmluVFliFIfq2TXIN3eL8zR5Ctximp31XX4J/d5U/myD/3P/vwwGUN*)
(*qqXVAvKQ/RyZbJTc0Pfn3hF/IGt8uS9GdV+DZo58oXJkopxrvJv3GujcPPas*)
(*DnkKjOJsH6+BF6vlaj+ykM2u6PPRaPxkM/Et6H7hGZcVxl+1BltNTDJ0kYsY*)
(*cc+qqtdA3eLkLwNkhk1wgCdhDaajY9tNkJ8dtPZua1wDE/WhXkvk+ExO6zfd*)
(*a2BLFg26iVxT7ijN/28NQn5rnQ9F5txvJFE1vwZUId+el8iGT87s86SsgWN1*)
(*iEMEcq+WgGAbbQ3evta0j0GeI+Lpr1noYH6TYz0ZWfjvrgY+YTqcT+stKEe2*)
(*1txWUylKR/fs1vpK5KQ0SvltMTrwVvF1VSNLu9bnth6gQ9/p8O7fyGeX7sa9*)
(*PkaH4zmX/LuQXba13eA7R4c3mZw3ZpGzbpRdq9SkA0PUdOs8MrUh3fq2Nh3s*)
(*RZzjyMj+Ef5Grfp0SAjhy1pGjhI5duq1GR1sTTLPMqH7W5lUCDvfLTrYeSe7*)
(*cyArduxseuNBB9XBZ3c4kbN9014L3qHD8zdBXtuRk5pqRYV96DBl+8cNhxzi*)
(*vvXY3kA6/EozUOVDNs4LNDoWTQfT5jxzYeQGS36hrFg6bNz7KyeCrLElqVfu*)
(*PR0SL13fKoqsdKnymnwiHdTjHqTuQt5N2/Q+nUGHVIZS/R7kSeVHcRfK6dBz*)
(*+HKUOLLtX66rDXg6XAp6rn4IufdF/EH9X3TwEzWYw9zYX5JpWEeHRqZY1cPI*)
(*uX5r5Vfa6CDE+abqCLLU4edPejroIBuSY30UOblFRNuymw76l/yXMcfsO9Vq*)
(*00+HkWH5/dLIvlX3/jpO0CGoxMNSBnnJZVv65BQdvuTeHMN8SyDG9cYM2o+F*)
(*4zdlke0cCpdcF+gwLqvgKYd8nm1pm9caHUIZHk7Hkfl0PKUDuRnAJ71lRgE5*)
(*sZJnuzIfA0SuHLRURJY79W2aLMAA8QxyHeaLklOpdiIM6Lf/nqCE/JzNZv9Z*)
(*cQZYtx2RPoUs4kffXJFgwAcmjeeY05diB7OPMKBB9/gEZsJY+/t9sgxwCkiI*)
(*UUbe8uuCEJMyA3Krj42cRo5UnqQWnWGAgbazzBnkA7lB7e7AAMKx9/cxq32q*)
(*iBjSYMATNgaHCnKLsLV7jBYaf/y4AWabiDV9Ax00v407bzA/8lfYXmHAgILX*)
(*EgKqyFy0tum7xgyI7k80xBx/y6NO5hIDrrIqvcRcbP316QcLBvxnW7eOefn0*)
(*fiY/RwZEzK78BuSneeWDCs4MmK5uWMUsKGVV9s+FAXjeAQk15BOiMQ+sPRgw*)
(*wNb4EHPVa3nzHXcYYHYk9zNmY442pSZvBqRnTdZj9ljhWjrzkAFWmUZ8Z5E3*)
(*3TPaqY8YQHrw6jjmV5NauZmP0futpYwxZ3YFuu8OZsCx0bcvMZ++uO9iRwgD*)
(*5EtupmH+XVMm/fIFAxyEcvGYp/NXpukRDPj5JnEOs490dF1eJAP66gxZ1ZHZ*)
(*k0+k3YxmAP/XACHMMbtanx58xwDGSSVJzBKRtxz63jOgzc37JOYCTq5zkR/Q*)
(*fqmpaWHWCPyyX/cTA9aT3ptibl89z8SazIDnT17YYra/PTZYnMqA6w0CNzGT*)
(*pwLKPL8wwO3qSS/Mj233xh/NZMC/vesPMfP+KX0w+p0Bn6ccAzF/NLAwj8th*)
(*wI+Iu88wl6u+FeIsYkBQ9YcwzP3HiO33S9F8d0a9+v/zM3yPWCsY4HVX6/+f*)
(*238xkQirZICE2+jz/6/n+tsWoWoGDDE7/3+8afFun8RaBrQr9v3/fZmjwgeO*)
(*1jPgHSc8wuyRaNGQ18CAwz7RdzGfsIn3UmlG+3F37Bbm5V2DYgQiA86vSTlh*)
(*Lu7ZSzBsZ8ALirsV5kcxdh69nQyg6+YZYVa79FnE4Q8Dzs2vaGLeIjBe9a+X*)
(*ATaV505hJhAlXO8NoPd/jpbC/PyV8w7mYQaE3KHuxszHOXtdcJwBWmr9DOw8*)
(*OwjH+D5MMiB1/OYM5tin7sWHSQzorOfrwbyHeZHr9DwDEu9+zcE8Wi5fWE1m*)
(*QEd9cjzmFN+7Vy9SGcB+uyIYs/TKao7dKgOk0q6ZYZ4vOG05Q0fvT6IA5tw7*)
(*j7Z4b6B4+v1NAvOpeaYrIazrUGWYR8bi+fzENsZ33DrESmr6YeZIvpBykncd*)
(*DOL8r2JutHtxsYp/HeTl6YDZZID7U+fOdSjcgd/A8sm2Y8d5xv51+H6McAfz*)
(*w6qDERcU18GKb1EFy0fVx46n2k+uQ5ypiyBmFtW0UavT65D45wIJy/eQn0cV*)
(*PNTWQeKvThTm6Oz/eqN1kP9FjWD1I/sDiI/ZrMPbr6evYfVm8cHiNju7dfjv*)
(*v7PHMSteSZsZuLYO/GVdTJhLeHjyup3X4XLxz3isXtU+Hjzb4LkO+bpvW7B6*)
(*1mf/6Gpu8Do8f6R18CQy2+GSuMff1yH0O2e4PLJ1liKf2MY6OPNXEbH6XWix*)
(*wk1k2oCDp768wMzNVsz1hHUDnH0DzmPGW6mwT7NvgJPS9lIprP5xamzmC2zA*)
(*StHXVKwfTDgY/dM7sgH/ES49lMTqvahrva/pBrQJHJPC+tHjp58C+jM24FLb*)
(*gWGsf+HubdOx+7YBZW/qkzHHObvxT2RtwOfQ+BuY83RPfZ7P34BUsfpFrP9N*)
(*8LVWs1RswKyvDidm3Q9M7EfaN0DbzFQN659CRVdfedM3wL7LoF0Q+XN6zeW1*)
(*dXR0LEYfMMvFSe31Z9oE9vLnNzBfeETLCtm6CVKr6UyYfc+Ft77n2YRL8TH/*)
(*Yf+/GW0pF6w6uAmMy7c/YP3625TYe269TYhNdEjhxs4z4ant4MVNsP4W44nZ*)
(*wfjfoe9Gm/BGXVgN8+7S0mz9K5vg5MfSj/X/5+GWtS/tNyFT2G8nZmfFd2TO*)
(*+5swkXogAbsvaMwwFfU+3IRQ15PumPcmOvtm+G3CZHX8WcxdHCe36QRtwvMT*)
(*RpPYfUNzoGt3aPgmmHc5K2I+ECSovS11E4iv9P6yI2+c9MV1p2/C4rx2Meae*)
(*f3/bUr9uwivtyNeYI67kWZ/P2QTcM76zmDePGt95WroJaf45n7dh9as1/ANr*)
(*+yac8nC8x4b8I5h2rb1zE7bLy5tgjjp99cjnP5vQtHpRFrNuiky++uAm7B2/*)
(*PrUVqzc+TfUB05uw4X3ABnP0MYVwo1n0vpd6ZzDfHo033T+/CdIqlaKYJfVc*)
(*h/DUTbicEfpnC1Z/mDuSI2ibMMU0UIR5uOC0i+3aJnw6/CAGc6nLZ1m59U3w*)
(*n7a9hzlm7/alzc1N+MD7+grm/wGu4x0t*)
(*"]]}, "Charting`Private`Tag$392830#2"], Annotation[{RGBColor[0, 0, 1], Thickness[Large], Opacity[1.], Line[CompressedData["*)
(*1:eJwd2Hc81P8fAHCHUEQhUVqSUkKl+pbyOqEkSlKJbEkZyS7KFpllJFkhWSEr*)
(*+2yy7s7eIZuMM8447vf+/P7q8Xy4+7zfn9f7Nd7XEYMXak8YGRgYVpgYGLB/*)
(*lZ5MkMsmn8jkuW3S6XQcnllun+UfAXkQ4X3ThPlcTHhktoA6tDB6zmIeERR2*)
(*tRfRAYYvWzzoCfhbTZ1ZUQJGsHGo6ThmYb8D3+iHnkEIeIhipikZftIXsQD9*)
(*PUcOYZ7yaKvSD3sJL7M3ODE/NKKWfBSwAYcLxE3s+VVy+35VRtrB67mX/19P*)
(*8ujVzKVDr+CK7PFhzJGMesnH4h3BkZG3GzPbsFvcA5G3IFL8vQVz4WBLU6W6*)
(*CygG/fv//i1VjZQeh7nCclHVb8wiZcs1S51u8Hb/tmrM/RLv5PwFPECmJaQU*)
(*c0gMf9kxLU9YCLbOxazElXKlNNIL1DgISZgZnKULHgy8A465rE+Y82Ybz88d*)
(*8oFS60RXzGY6Olnv9N+DtQWbMWah5jnxw/G+4NegKY+566prav6IH0zMnBbE*)
(*HPCD+8RdkQCIOS/4bwtZ4UBCwuTTQEifr87D/LYmYle5ehAwpTQ5YH709u9t*)
(*/+Ug8GZqP4tZ6oKY/6OwDyDP+npkE5lr1qbh2MWPwElR98c89a1kO6XzIyi+*)
(*WBHDXK3NoljqEAy90mMVNOTYPXe83guEwNKxuNuYHZs+VT0oDIEKnSjiBvID*)
(*z0HGo1qhMM+NV8B85qqo7NxGKPyF8fR1ZI7ll85FkWFwpWiEHfN4WmHJu6uf*)
(*IN/QWWsNucKIiXZv4BNkWS1GriI7tIU4zBz6DDz0Y5QV5Ht+/Xn5ZZ9hrfoU*)
(*M2ZxeZFlD/0ISI+KYF5G3k6zOHeX8Qu8b3GhLGL5l/3r5YH4L8B4crSRgkww*)
(*ZciclIsENfPajwvIEUdvzuaORIL3zGG5eeS7wT3Pb4tEw+e7ON1/yGK3jibv*)
(*q40GBj+dumksf5jMxseexgDlyeDeKWTdgq2fBPVYcHOoNxpDVt5zlHzsWyyo*)
(*GozojiBffnlj3nc5Fmq+eV0ZRt4jGiSuEfYV/nscGdqPzOiZo1I69hW2zOO4*)
(*epHnB7vMhC/GwTbBB8ZdyA3hR1LnO+NAub08oQU5f0mh/sGJeBgc/OdDRE5U*)
(*fT5Z7BAP+zUyrjciu7JlH/cRSAChOz9kq5FP3vjgnWWZACek2U+WI7d4vpjs*)
(*rU2Ahl2v+4qRhZnEUsXtvsGMrlV4NnKj7A4OjaZvwOKpE52ObOsyYeYqnAgv*)
(*Xx1+moxcs5kg3taSCBtmEZeikV9ccQ/cFP0OidqvIByZ31F/XsT1O2wkOjJ8*)
(*RH62eiDrlUQSuGv+SfXAzn8x7Pwhv2TY3kE3N0bOPWMbpvg3GQxOcPHoIOtY*)
(*3qO+vJwC68O3De4jZ/zjKqiaSIEzn2VXriFriM0KzEIqqNfduXAJqy/Txtd7*)
(*P6XCxxb5vRLIahPeV54rpIGo++UsAeQNkadRwZFpkAJ3n3IiJzxR2CpeTAOK*)
(*kGM2I/LyMK6MK+4HqEd1skxt4fDh/a/lcrfSQXUtUDUbWVbwUcLA/Qz4PPvQ*)
(*NgF5SvPiNrYfGVCntrk/FPngzCvDH6czYcfT9L82yFbskzzXHmXCo86ddEPk*)
(*mpMaVR0emSD0gBKkhvzi2YVjDH2ZYBkYzSGOXOX9rT2E9SfwR7yZ34fMn8Tr*)
(*JXruJ7w13gWsyGVjlDG19z/B4Jr4roFNHJ7bKCPp239ZEClVcMoD2dj94KPL*)
(*RlmQfrKL0wy5KM5/OzEwCyQVv8ndQzYaMn2+OpYFqx2XQw8h5+qcELsVmg0x*)
(*H687ZtBw+B1vP/X9KcuGki+bAR+QdaJY/G1msoE84dJthczSN/IvSi4H6o5O*)
(*MZ9D1nj0NWN+IQcuDDvfTd7A4ed61XfE7csFfj6c2DtkLx22J/fkcmFK+7SM*)
(*EXK24Yt9eSG5wMiBAobMaXHV0/FiHsgXrgS6ruPw3+bn/4jp5cFqrQhZA/mK*)
(*dcLlAe88cPOgSEogP3dgn8f35EFnhyZ0r6F+5NatyfLmF9zfNXdZCPnxNv/c*)
(*X99+QYRcCi9lFYenvMPvetb8C8JmVCllyIf9v1c3HMoH52KfVC1kp0+2kh8q*)
(*8mF15Ha2BxWHP5u2m0WQrRCOkbfNfV3G4X+LV+s1SRZCsehVTT1kvZ8ORW8f*)
(*FUKf6VzAAWT/vD+WQymFoPrlP4WQJRx+ouxHb6JKEchLsNhbLeLwznL6FzTs*)
(*iqA84B9BBJmvhvfD9pgiyK12z+qm4PByDY7XzeaLgOOXmoc0cnT7zZ9ngosh*)
(*e4DDeHYehz+vscn+t6gYTDpOe39CbuzJNA4ZKQapW9ZnAXntz15B6vkSCNE9*)
(*zPl+DodXnxr1Ku4qgYo5w6c7Z3F4drrr4xsHCcAo1pNTM4Xy+dlDHbHrBKjQ*)
(*s3bSQBZuE9PbbUGAISvp0IlJHF4mudOwt4QAEVF3k5iQC5LDcn+Jl8F1Wleb*)
(*8DgO/+/eTPGt5jLoCAs+w/UXzY/tnVdlhsvA/JPWoPswDv+6tLxUcqUMbidM*)
(*hqwM4fCsop/K9hwsh/+sbTfaB3H4I1uyVX/MyyGTTf+M0wAOfz/pU4P1zgrY*)
(*rl5Rqt+N5rO2m7LxkQrQcfptlNOFww9zmzdpnK+A2XiDEmbkF2+uEa9qV4AR*)
(*+yPL6A4c3kftXwvrjwrwNyPJ5rbi8CW0az1flCvBpSros2kTuo9kndYK0KsE*)
(*7cM7q8Ia0Tx+yt/nYlMJDyRuKhAacPhO8r/+J5GVkHOo/hxbPQ6/kBg+JDFT*)
(*CSma2nte1eDwx+7OTlT6VkHNOxMLcikOv8/5xuWnX6vg/QOH3QUlOPyuH7G+*)
(*O35VwXzT3v1RxTj8Ots98bvDVWCuPAeahTg8qeyX9cB/1cDzGQa+56L4SLpu*)
(*rY5Ww6UJ4wzvVPQ+Oj13IjeqIXTsR/TZFFSffue+wu4a+Drt8aQrCYe/MzEq*)
(*53WlBuhtu9X3JqLzilXy4QmugZrHb5P0Y3H4Zi5eXnF8LSilkcIPheDwlVfN*)
(*n5Dv14Lr1w8v3T6i+WBak2djWgu9SvfdB4Nw+Phah0dFYbVA0tg8HuCP5qlz*)
(*f7Tiv1qoep2vnO6F7htziaKGEXXgR4t2XbXH4fceYHBkyawDlvbb/dvtUL++*)
(*9agxuboOvBnMK/bY4PDURPYX8/N10LwQti5oifJTxzLnzY3fwJuw1T3+FIe3*)
(*b74E4Uu/Ibx1jWz0ANW7eO7ob7Z6eC6s/wmvjvqvv6TfhmA98BfoU/jUcHiy*)
(*8vFuHYV6UDP8CtkqOPypBl7rY2H14PiDuitZHof/UzuXmHWxAZxkXou5nMHh*)
(*7Y6bqozcaoDsM9UqDBJof+/GFvfoNUBFcczD12I4/KXr/fhX3g0w7/ODqnUc*)
(*hw+uqu+B7gZoc319ok0Qh79R/o2z6XUjtBnDngAWHH7g8JHcrYBGIId03fnO*)
(*hMPbuERqScY3gvCN8rwCBhz+KwQnBTc0wr8ena7f6wz4jRKXa5qCTdCwEsIV*)
(*NsuAzyjUshsvaYLzvHlyrW0M+L05uwdwOCJEq3x+6/mFAf/Yb0N9hI8IvSqW*)
(*VhqfGPBxRqMNNWJESOyIsDoSzIAX21NQ4KtBBLV9H1RD3zPgr9nphe7JJIJl*)
(*KLsf4ysG/IXIh4khEiQQk/m89+89Bvwj78Yhz8skuBN4Q53zDgPeyUb2gL0C*)
(*CWK8+F+JKzHgy5VPhTzSIsHsW3lJFTwDXnlzy/XgOxI4qapfZxRjwOvrJGon*)
(*DZBgZkf128s4BryH0v7PnydJcL2suOYBlQ6JF4La3i+R4Blhk+HwDB1mOF/f*)
(*Mt9BBomc3RIm7XSwI6hcOnuBDPYWXFcbEungd3iZt9ifDJPx63aF8nTIG5Zr*)
(*JF5pAV5ts7dnzLagQ9F1wf9GCyi6NfNuaG/BSnopn7JaC4h9CTFLurMFFx0v*)
(*6dc9bYGSELuOH2e3IJ9HYqXsYwtENZzs0qVuQoH8vsNZEy3wK6lIU85xE4oT*)
(*561DQlqheozR/pQZDfo4xD+rxbRC/oHdCeKaNKBZmZbuSmmF5Wb9ot2KNLgK*)
(*Y2z+hFbQedAkb3+UBqWdvVGeU60Q87zhcEnXBhDYamvt8G3gsOYTePjqBlQ8*)
(*j9r/aKYN0gim6XUza5A9aWqqs9IGrT07Fg93rUGCyeUiQ4Z2GB1q9dSqXANP*)
(*405NC952aDE45mX/eQ0UDXZHuF9pB5rkse9K8mtAfOTJn+7XDm27V69ph6xC*)
(*n6L5HqbTHfAylU9T4yAVlkWucv0w74Tqb1kxb/oXQWb4Wu17h07YUpypNiMs*)
(*gleUorOJeycURkQCfF0EPl71uaOfOyHkrLB0oNEiXMCZEiOqOmFKbxvb30kK*)
(*OPSGB3rv74LhfeOJkXMLQAtc4jL63QVqZ5UznVrngHEtddf+Yz3g/dCp9rf+*)
(*OOxvurYIEj0wr3udcfrIOEh97W43utQDFurmZPbhMXhyk/VLukoPcCrBcVWD*)
(*Maj7bHDsml0P7Bc72UU3GIXAywKXn9X2gIOne8Hgk78g6ORl+OtZL3gfFpHg*)
(*oQzAedUD1/use8F38okN3m4AbgvnnMC97QWpb6IvM9f7wblx6J/Sx154e/29*)
(*sj5zPwwLXnUYKOyFgMW4BLn9vZBcsui3jaMPhO5ewFM0OuECXS/vXnof9MqH*)
(*xK4KkUBZpcxNIb8PiveST2WaEsHwy6E7Fyv64PacQtDLj80QdHFgfF9HHxjL*)
(*tSXs+NsIky+0BIY3+wB/fCWWFvQbIgfvO1mq9IORgNm7TaFKyBLPVTR42A+v*)
(*K04GB2tWQJ0T7x51/X4Qe5qvYBRcDsv8rT8u2vaDBiHTp4INzb27qn+2Ivvh*)
(*bLp6fei2QsBVKF3zn+mHxtXMv9k9acC/K4XTeaUfLJNrzlntTAFxne29lgwD*)
(*0Og77dep8B0012ut1XkHQPmUqmdZ7VfIOavwbf+VASCJtoZ7yvuBcTywpfgO*)
(*AMNoTkaCYwTBaSG67UvoAEiOjZx+phFDCIatWP+YAVh1nVbX540jlPUWX3qZ*)
(*PQDhBxczN8mJBAHey6b/9Q7ACc7ijAqGTEKjh1Rzzck/YDT9OPxOWD7hMlcD*)
(*44fzf0C+4LSeemYBIemz/kUt/B8widcJcKopJLinB8TO3f8DV+6klJweLyZc*)
(*7pqw4nf9Az1XC8T+OZURkgycvw/7/oGB/B379NbLCHz/9vSlhf0BBpqSRr9t*)
(*OWEBJ6cgm/YH6nINfZeMKwhJp6L2Pu9Eblq1eH+misCXd05ZavgPMCexbsrF*)
(*VBE88PUuWzN/QHxXnO429mpCZsqGzU3cIMTRH7XH9FUTfAknBvr5BqHuD5/A*)
(*JZNaQgWzzeo1wUFIlO9+dKGmlrB6k8CddGQQhGrEq68K1RGM2+7fsBIbBJ3F*)
(*sjTztjrCtUm3TJZrgxD74gnh6ZF6wivx5nrTG4PQ0nOU3PCinpBpLTBKUh6E*)
(*aIu97LLF9YSDWxkCXx4OQpKjYIPhnQbCfbmNcwzag0Bich7jCWsg+Hlfv/3E*)
(*AO3P/Jl6T08DYY27303CfBAGf9ZwZuo0EiQ1TkSFWA3CwYVEh4qoRoJxlPWv*)
(*NftBCHv34cVCTyOh7fiOmUq3QQiSuriUeqeJkHnl4n2N8EGo+HXSqFaomTDu*)
(*6vaiJGoQsnS3Ny+oNRMO1jb5CMUPQlmvwVVF12aC312j0pkfg9B8y2e7W1cz*)
(*oTIso+tu9iAoTePWnuGIhPXedUpe/iCYaQxwvztBJJgYfzjuWjkIW+k+ns9f*)
(*EgkxqX2yY3WDaN5JDZ4PJhI65o8/vtWM9s/qZYLPJhJ2XrC2y2xF69frlPmR*)
(*iQR5x9KgPd2DsMNCP+zALJHgWLY99fXAIPD29ba8ZiYRsrbdr/7zdxD2dU31*)
(*sO8mEQ4FTa8lz6Ln3zv4Oe44iVCq75yctTgIaTQ/+rYzJILOOZ5HRauDwDGq*)
(*np5xiUTYZP7OVrU5CB72b/tjZUmEyI7L+Y2MQyB4TvJ3jyKJIJ3U/LSddQiG*)
(*z4b/1LlDIvS8Mtg7wDEExmHtPafvkwivbq3UjO0egsvkIb+bmiQC/4H3dnN8*)
(*Q7A68YonW4dE+DV7QGR1/xCYhbGWmhqQCA/KfrYzHBmCRJ33f2yekAgrHxQ8*)
(*t4sMQdxHl2biUxIh1LBbivvUEGTeII28eUYiSJ03H9knOQRZspnOLs9JhDYW*)
(*XMjR8+j75CB6F7J1V4ic2OUhKM1irvNG5k45sSgFQxDQnQwf0Pd/OhbHXZUf*)
(*gghFbtIier6qiqra9ZtDYNmXNvoTrT9/cAR35/YQtAimzzWi/QXO2/98eG8I*)
(*UrjWdK/rkgjiFez6ehpDcJiTLfWQFonQHByz65n2ECx9a7+h+YBEMH9yruyl*)
(*wRBcc6rOoKqSCBwXa1+8for2L+uvs6ZEIqSyaR1yNxsCxvQDLXryJIJSz2yz*)
(*78sheO3n+17iKokwmer2NsQOvX/X6van50kE7zd84lGOQ8A79aaN7TSJcPxO*)
(*Sv83lyEQKeC25BMmEWoOy/inew7Bf+ccjD7sIxG2VT2ZIQQOQbTuQef2bSRC*)
(*Qujal7qQITgozfJabhnlz1P/W+TPQ5C/+X1bZT+R4LYjN2U4fgi6MkZ+u6cS*)
(*CUf6FDWnk4ZgtljrRkMQkVD2o2/70o8heM/IocNiSyRsqTI/25Y/BGOWoquG*)
(*l4mEKKFwfs4SFK+cjyS//UTClaVTdXwVQxBr1zWdu9FMcPx07/iJxiEIz1lO*)
(*2FPQTFjtjxtVGhwCPG3G0uhEM+FTxoXQe6NDsMIp8m0C10y44Fov/3hqCLw9*)
(*TiV69TQRbIUp8eZLKF5psb2CPk0EynNZgyC2YSgtC49n72skTFP/DLRLDoNl*)
(*TP6/Yq0GQt7TwAqlC8NAWZC4ySzcQHDplEkkSA9D3WyO9KvpegLfryjz5OvD*)
(*cI19onXDvp5wzU6L5vR4GOJ6AiNa3v0mRCx1CAh7DwP/dOwFSa9aguJCk7rV*)
(*n2FQL9aR29pbSUiYKqrnDEBz1qjBevNLFsHi052jz4P/gtlOyqzB1k/CRfm/*)
(*jtXhf2GpvCWnRu8noT5qx2mn+L8wEnYgI+JYJmFB9VHgZP5fkHwy+KdTN5Ug*)
(*82vlXtXfv7D6RUb+TehXQrf7mf5Xl0agrH08rPvnM4iTrD7fLjMCekeepn4C*)
(*OzDr1wiQlB+BwdkGa80YZ2C46ALjt0fABa+nVi7sAyJTzV/VjUYgk0sNx0gO*)
(*BRtVM2OJwBGQWvTvkV6JB84D3+dGRkag2VV+zedzFuAfuPf/nRwB4/D9MSf2*)
(*ZINVoG7D8OwIrE/m5Z8LyoYOHP/3wdUR6FA+FOnrkQPRY946fRyj4GF0fPdJ*)
(*4zwQz3ze1Co1CpLFTloOHIWgIieeWuExCnUbAe45lmXw1mnH53KfUUiPYuVI*)
(*SiiDzNwxr7KAUbhWGmB2s6sMeE/EGJSGj8LEyP3noTLl6D65a19h2ihsvs8j*)
(*qrJWgFnHgvfPtlFoNrvspupTCX7Pco1jhcdg4kRlQ+GdGpjC//BmPzkGhWw1*)
(*QQtONXCD/1uKncQY4Pl8Y38n1wBjbcis8uUxGHgl+UqHsRZeCdvYrd0Zg7MR*)
(*BcZHMmrB+M9ZLzXHMXjMXFEgT62DqryTSSUuYxC7rV7px8HfIBQgVH/Cawy4*)
(*ehLvFSn8hv4r3Jz0oDFoG6/f+fvjb1CPWAhNTRyDj+yHz+acqAfZ+5nfmFvG*)
(*QCM80nePfANEiyXVvuhEn48mZG8+aQAaU+xkT98YnJsf6g961wD5WYGnf46P*)
(*gXqlSUnB7wY0d1/kam+OgUFm+evDSo0g2HC6Ou/EOEi3hDZESDfB67hj40fE*)
(*x2ExgUnKQKMJOl8d2O5/bhymrahMvTZN8PHEThVDmXFQG3R+N5zWBNu9Ztq4*)
(*1MdhvFpuTl2gGVbwqSMmzuNwxEqMIjLSDMUJ01/sPcaBe14pKma9GdzYxO55*)
(*+YyDZG5lefcuInCS0srjg8fh3OrKUPAVIojopkcPfB8HUYVCK9EPRHjw5qeG*)
(*OmkcdgTUNofsJYHg0AKXYfs4nHrw7QblMAmG5c/WvuwZhxYpnQfnTpLAgiP7*)
(*fODIOFAkwqb1rpDA60sOd/3qOMhuZs1a6JIgt+BXIxyZgIe/fDffxJLg9YFV*)
(*99siE/Biy3/vySQS4F3/k9Y+NQFz/MkRxRkkaFQsSH59fgLSdaICQktJMNJZ*)
(*6JV7cwL2esZvF+wlAe9KCf6U1QQkaUf1fOUiQ8+jLeol+wmofpB35fceMsSW*)
(*yGQoOk0A7eHhmu79ZBDzJBww9poA/R3O63nHySDPW74eGzEBN0MeBxvLkMHm*)
(*bFUOX+UEPJl0NiI8JYOx1OHYD3UT8CH58xTFjAwaF5x8OZon4NkvzlxOKzJI*)
(*X5YyYOyegLX6rlG6IxkYryVwzc6i/dj/VGsOIMOSHMOGydIETJf0hd8MJsOo*)
(*wuOxv2sT8Jb+oij9ExnqbvKWdDNPQt6rDO2LsWQIvOvxvHrfJPgePuV6I5MM*)
(*LvcG7+MPT8It8UK+vTlksLp/Rbbo2CQoO4dyt/wiw4NHS3t/Sk6CuV/95g4C*)
(*GQT1jaojr0+CZviDbv9GMnAalmXuVZ6EXWMcPSVEMjA8EYz8eHcSMgXEi3ta*)
(*yPDXpM3K+/EkCLJV9/d2kSHFUv6wjdUk7Ii1DmcYIUOkVSz7nP0klIZoJEWO*)
(*kcHfhrby7M0k5L4I33Z8kgyWDjlNut6T6P4b9Ih1lgwGr3cV9PhPwtUuvSTt*)
(*eTKoO5kl3A+eBBvb0qU4Chkuugg73oqeBK/GpHMbK2Q44eZiXBM/Cc/KdJw4*)
(*18iwz6PvrmzyJKxML1rs3iDD1ruQExdzJkHRyHBjbIsMVYHMHUL1k3CsQP1G*)
(*+7YWyPugVx5FnITX/+XaPWdtgaTg4jT+9kn4LbwcOc/WAr6fbNw5ByeBS3b1*)
(*BYm9BZw+kyx8RiehnpvH/OTOFrD4IqbJPD0JWenFODvOFlCLGZFcX0bxUawM*)
(*HN7VAvJf8YK2G5Mw7OzTycTdAufjI1nnGaZgB//Pe3t5WoD/u/rAKPsUWEXf*)
(*dd+zpwV2JGf+1tuN/v79hjsDXwvQUjhye/mmYJ/NZ9UB5D/pVb7kI1Mwey55*)
(*0oq/BUj/mGfOHp+CbP7gWhGBFigTU1AOEZuCE1J/HZuRM00905bPTIHzmfui*)
(*z/e1QGxKNcfDi1Mgam5GWUMOmtxmnn9lCiJ326y92d8CrieuNwlcm4JJKWe9*)
(*FWSrp16nHW9MwWmqyU1DwRYwSKzx71Oegt6vbe1VyGqjLLNX1aaAVHv6+P4D*)
(*LSAnfON2zMMpOG6cpGOMLGX4Lp1Bewp6tslGJSILx9VyGhhMwZGBYqY+ZN4h*)
(*1heVT6fAlZqcxXKwBbYdViQKm0/B5zu69SLIKzreEl5WU/Bxk/b8CvJYVF3g*)
(*uP0UKDw0y7mB3NHHNq/4ZgrW+fRqFJFr999UTXGbggZX9TJAztf0yWT3noL0*)
(*gMByMeSkz793mftPATvnkUku5PCu7S+bP07BDHn4/iTaj89eJbJE+BTkdDiI*)
(*FCC/fvD+zIcodD5md185Iz8Prf9AiZuCBL1vZleQNdt2UO4lTUHgCYeD8yge*)
(*t3huqeX+mAIejvC0COQrar5ZfNlTsKEaLXUFWexDA7dD/hSUG8+1tqH4CpLY*)
(*rbtLplA+FX03QubgUm69XDkFvFG0xml0Pv/8GoM3m6bgEdvvt4PoPAcaOJZ0*)
(*WqfgTLSC/G3k5h0q6mVdUyDftb8pG+VDuncTr/vfKdDocA433tsC0bU7bUcm*)
(*puDiRwPuHJQ/gSy32xVmp0BC9mbpGso3S4/mULa1KRAv/cJvwdsCepWcK8+2*)
(*psAjv5oWifLzLuOdBw1M0yCpZ/m2CuXvWRciXwDnNKSVf6ncRPm95Ej6xCM8*)
(*DYTpX6cEUT2MFO5atRGdBtWXOa3cHC3Qvqaq0SE+DfLmrWOMqH7y7Mn8ny9N*)
(*w2L1959EVF8O1i2fD96ZhvgYM8+DzC2w8bzti+jraXBPKh3/hupZ/YL2B3Hn*)
(*aUgZdq5+sEqGNNyY1zmPaXilIC3ORCWDTjj15dWAabjkUzx+e4kM5dX7FO/G*)
(*TUPZ8NXRbNRP3h3WX3Kon4Z1zYnZimEy8HT8U67bNwOHQw2zntSRweyr3bWm*)
(*QzPQotcw/70G9Qcz+sUW4RmQr2zNGa8igx3T7qN94jNQ8GyVWa+cDN2SUmtz*)
(*12ZgKMNT9nwhGWJ8X3/bazoDAydC096kkuGULOuWcdEM3CtWCu7zI8ONuCNP*)
(*f+6dBeFvKkbvmcggFc4pUiM4C2LhRWxfcGQ4HLAx0nNkFpb/pJan0Umw/qrd*)
(*gFlsFm5qFXi0bpDgx10fnYf4WfixKt0uvkQCHtzC/U2TWXjPcjb08CgJ/uiW*)
(*yd8snAXZZddejjo07+7/YNIhzAKPnMZDxRoS5N+KqLCqmoUv/14+8awiwceL*)
(*NvjI5lnwYBWRZy4ngQLXiStzw7OQbeyiuaeQBKmlgedC2edAWCru/Ls0Etgf*)
(*0BUaejwH7uO+tws+ksBAVfJBn/4cKPBp+5/7QILb7rj3ncZzwFE3HpQRSAKR*)
(*yfiFJss5OP/5znSaHwnacibKCj3mYJsXnqvEiwSSyla6IWlzICmrwSjlSIKJ*)
(*156RN2hzgN91T9jtCQmmHpwy+cowDxmaXNUXjEigN8/DeIhpHqKfbcr+MyDB*)
(*raOj5/exzQNhQ4Ssq0cCIZ93Ubu450HmLreilhYJSOqNZpvHkIUefUm9SwKx*)
(*mfvsnSrzwBmX55UJKB7qyoT3qvPwtuSdq48MCeRKrlnL3JsH7il6vNFVEmgG*)
(*SPR+05iHxJGvj49Ik8BHcnuqjeE8OE2e9ci5QIIxm2Il7lfzoFJxl4wXJ4FV*)
(*f9ZWteM8mHoXtUufJsGmQnLWq7fzUJfgvfCfGLpf7A3bN+Q+Dw1i959fQvcb*)
(*uYIXUxkB83DqadyVeyIkiN0Ueq+SMA8fZZMkJw+i/RoLyOC+z0N2odKObci/*)
(*mrkoOcnzoH17aEToAAmaY2iPBDPmIep9eJjRfrSebIfodME8BPybKqWj+5Wm*)
(*l89vb+I8sF73TAnejfKFa56tan0eil2qHh1jIUFLyft3Nzfn4eTiOo/PNpQP*)
(*ZsdYiPR5oHIZac4yk4C7XpOph3kBilP+8yxhQvas3pzjXIDeQnNhCxwJdm1E*)
(*UPYfXQB80MFD4utEICWff/n12AJ0df94d2yNCEEapDmREwtgFmPdcmiVCFx5*)
(*2/6dOb0AwfxZDAdW0H3xpeX4jYsLQPIY0b5EIQLHuEKf9a0F8C3+yESdIkJj*)
(*6KDmmsoCbC8fn9iH7Cfv2P1WdQHuM1fZy04SgT0us8Pn/gJ4fEhzDxsnwg7t*)
(*/eQY3QUQ+XduTneECGyt89UN1gvwZvHrpNkAEepcfeXv2i2AYLVRbkE/Ebwl*)
(*RSo7HBZAyT5tng2ZNUCrbOjNApxUPP89vZcILDdrCqnv0Pvf5kzn6yYCM+FL*)
(*xtHIBch0ml+430aEnivyin+iF2A/d4BUcSsRMgpnBiO+LkBaThHnMWTNvKvc*)
(*3N8XIIdy3X6LTISfP/7YMGQtQGMr+RORSASvU947i3MW4KBwzB4l5MfJkon2*)
(*vxYg7tZeltpmtP9vrp2zxQvQyfJDrK6JCDqRRy/31y7AOkVh13gDEaT2NbaE*)
(*1y8Aw+v/4m2Qt4fbmKo3LcDw3ZsezMg5wdWRDS0LkOr09eCpehQv36f0gv4F*)
(*kCRFnw+uI8Ig265w28EFODQ9WXgWOc8rX/LM3wU4qyL7vK2WCAZu2w2SJhbg*)
(*AevF8weR81+lVoUtLoB7XgG9vpoIAdR72morC+DG1KLhiGxoS1veubYAyrlv*)
(*8k9XY+erctxzawEszFheh1cRYWR2mYDHUaBCTqRVFbnQLFqDxkQB2au6R9iR*)
(*n5jM+Vhvp0CMeKWvZyURLo99EpLgoMBZ8ZsxCsi7jPBFU5wUkLOPDmNFLtb9*)
(*MKPPSwFnWuChDxVE4H547o7qIQrIGFw8WlhOhIm23nF2IQrks9T3eyOXqHm4*)
(*1ApToFvxvOYjZJPb7T9lTlJA+FjfO3oZEWQa3yiti1HgJnOVZDsyj5LI31wJ*)
(*CmQd7H+fhlyqYM97+jwFGHiEr+khh1Qe+jFxkQIrXfs/XUF+JlunkHCZAiFV*)
(*XX77kHmvCtjvx1NgP11XqZdAhKnCcq7OaxRQ3cazoxS57L/nSR8VKBCf+Vgz*)
(*DtlUqqh7+y0KzFMzI19gfye1p+WrUMD909k5DeQ95vPOT1UpQFdyz5FDfr6d*)
(*/R7fPQp8KBqfkEAmfDsmUn2fAh0mSd4HkHmv4desNShgkbE7iAP52YBmo5AW*)
(*BUyJ1hu0UvQ+r21jyNoo3hSh2llknr1BVi56FKg6az83hGySnaIgYUgBp5hG*)
(*yw7kkjvV/ANPKHBO3v1GIzL3zJ9pPxMUz9P0F5XIT73XS6VN0ectk8aKkIuF*)
(*93ycMqdAgdj091zk3eUSTz5bUoA79GJmJrKxttJ/itYoPos4Whpy0ZoRO9WW*)
(*Au/SKt6lIO8Kcx745oA+r7hyJxn5ydmIn+qOFLhldeAu5sLmHA+mtxS4xFjy*)
(*/89zmRIfZrlQYHhTex57nhHr1El9dwpkR/u/w9YriGfe4vKiQNH9/pvYfjjx*)
(*h8il3hToIg6fw/Zr2HcpwdyXAjlrF69VIOc7qNsLBlAgjcnxZT3yzj0vlBqC*)
(*KBD45HZpK7LBT58Dr4MpoDiscWwA+Rdq3ifCKKDvqxA/icwxVVrZGU4Bk6pK*)
(*qRVkfa/uMK8vFLh8/V0HEzoPdgLn1ZFYCghUV1wVQtbTEt0VHE8B5fVftLPI*)
(*uVS5v7KJFHg0O1Mqj6wr+conJpUCtUV4vClyTmPw49vpFHg1Q191Rt7+LF1i*)
(*M5MCLoGQFIqc/fVvu2YeBYpdf/ZXIrPJbCVtL0D1Efpduw9Zu4ffKb+IAnnf*)
(*v9UuI7Py3BbiK6dA5z0rlVMof7U88i3IDRRg2aneEIucebhV1qWZAg/EXl4v*)
(*Q95W8o9XgkyBgQ7H4EHkjGWhIr8OlJ+LXz8dRfXF9NSfVXGIApkrItQs5NSb*)
(*BrGlKxSQuBn6eQnVq5TYERlYo4BI//qsIKrnUs7BXsIGBY5knWzD6p3Uqr23*)
(*nGERdrGqKnxGXtJ+FFC5YxFOaci+u476hYyVqlPdwUXQ+mpTkI/6DSlCRqPl*)
(*+iLg5NHPR9TPHr3ZXFa7uQhhl6+aCf8mwrBucXDrrUXwlIoS0kJeEr5MbFNd*)
(*BB2zCaffyAIZ5693ai7CJx7W/GTULw0qxaT6LBahTHyz3a6RCMvT+3aNhS1C*)
(*q9qe2v9I6Hzu3WAKkV+CN7b2m4eH0Dy5le8peGMJitJlFrWRH8uJsn27uQRX*)
(*uKarI5AZz7HvzLu9BKIvTOg8w0S4zU3k69JYgnqvlJ/Mf1G/Ij04KWi+BCte*)
(*GkEDaH7tu218NyF0CdTmPwRZThDhn0Jnq1j4EpjJvN6dhFx2VfFBbsQSPMng*)
(*1f2DbHz6pFZNzBJoXRzSUUHzMZtj1mgieQmaPv0eE0HzU7nB1kGsdAnsh+82*)
(*EaeJ4KzoGZMztgSHH7HSVmaJ0KJj8lhycgkundd8JzJHBGHbWwJp00sAFVzd*)
(*95Hrv+4Ojp9fgtSfjMRsZN6NKK+P60vgXDXPYD5PhKQfeeaWnMsQG26j17aA*)
(*zoN7QlrswjLkkrKf2y0RQUi0YfX7f8sQUeHjF41sC+m5R6WXge7Eb1WDvM/M*)
(*RmI/fhmE0+9771lG9VvNcHSH0jLsmZIeykRetRfYMaG9DEr/4EYfuj8c6Vfq*)
(*ivdchix2meMs6L7xvkI64773MoTcmQ0SRaZ8F/Ni9V2GQ4zXc5WRK6w4pUyD*)
(*lmHpgYn4R2QjtpbAs1+WoXVzz5gAur8knXt0ozxzGYYbF98LbqD+I6B00Dp7*)
(*GVb+hGdII7/aurwsnLcMmV7wQRNZ6bdgvHfRMmzfkvf7hDytM7h1p2YZuk5f*)
(*M+akEUH8/dO8gd5laHIebptDDnuh4R80sAxfzgkN7tgkAl39ptG1oWX4J+6W*)
(*egyZdOgUd+LYMpSuu7hrIr/MnTU3X1iGdJNUmXLknEHbYzSWFUhdo/G6b6H5*)
(*eNEjRODsCpQt2S1104nQYSBwIUpqBcY8ak/NIFsHpHcevrgC0+eUz24hp412*)
(*7TtxZQW6/k507GIgwcGQ018vXF+BZ9udzp5DZlzo+HFPcwUs7Ihu1sjRgmZ3*)
(*Oh+vQGuoyTFHZGlF3IKm7grw0Z/8dEO2iTklZWC0At6TjF4fkMdUXAotLVZA*)
(*YZMmnI7ckCJa6++2Ao+/9977i2zSUWqy23MF/Pg+RkwiMzOq7wh9twKRFu8a*)
(*55BlHr1VifRbgeO/EykbyJmsra0pYSvwYozhFg+6byqfM7EV/7wCH0Wps/zI*)
(*kzqbfFlf0Pv8FHQ+iCyUd1yzIHYFzOjmRqLIIUZOg7UpK8BMtLkugywZtNvt*)
(*1o8VKDeRvCOH3FSUeJSYgb5/hkNBEZmFh2zckbMC3/IjV+8iO5Qf+zdSugKS*)
(*DaZuRsiPDxLXcK1of/HbpL2QOQWscovaV0Cw8oz9e+Rynj0vbbtWIP2ocUwA*)
(*ssh2rYmJ/hWoFXXOCUOeWxzrIE6sQPFU8PI35LhZn48+0yuwwSgcloysPil2*)
(*W252BQwXJo/+QM4fsKrOW1yB0wtprDnIbr+3sqO2VkCZIrRUhixV9fXFQxwV*)
(*tBJsGauQx0vlT+1mpsJJ3+P0GixeOe/jPLZTgVfsSlUj8p4Yvg+mvFSoMzjp*)
(*2olc97lA+dheKhhRHFh6kF+HPGb7I0CFq91Kjn3IAz5xzmqHqJBIY5MaQk6y*)
(*Ebe4dJIK+6V7RyeRNV+QRRfFqNDVfGZ6Gnnnc5vRNAkqsBeIDv1DttItfHz4*)
(*PBX6gt7HLCALa2rz91ykQvevSyaLyB3qDG3Bl6mg/p1PaBlZWun6LVY8FXh4*)
(*R5+tIs/KT7KUX6OC4dD+1TXkWPCreK1AhTg+docNZObzLZdmlagwxh6psoWc*)
(*J2G7/F0F7e+EWzwd+dlJ/p/6qlRgMEqeZmBEv5cO6Zxov0+Fd+HX7jEhu+zD*)
(*jQRoUKF4jmrFjHxuT0KMohYVzh+19dyGHL5jiq9YDz3fRtmdFVlpm3+LrSEV*)
(*9k5PWLAhb9IlAiSMqbA5Q1bZjmywbLst3pQKYmSOsR3IvPP85Y8tqPCFFhPD*)
(*jlw7VeTE95IKgQqMKhzIr0d1/iNZU2HX+r1ZzGKDuCUfOyoIKeS670Qe6EnI*)
(*kHtFBR/x6zs5kYPab5huOlIhPELEF/M10pTIr7dUWL9pt4l5ud5/2NKVCm+7*)
(*1Y24kPvY201o7lTQDj1ZiLlCWXDunRcV8hWkWXYhJ/sb2vL4UGGnWqwi5qDm*)
(*lI1oX2QWa1fMdlwU15MBVLD90JiJWVv1ElteEBUWWWs6MMt/cAmQDaYCq77D*)
(*IuaTLbW8TaFUGLKnsexG3sXD9UUjnArK7Ka7MVPvPTgyEkEF75I+bswDIVHf*)
(*X0RR4SLvc3bMVe0jpzdiqCDjcGoDe14Kn1iOVxwV0s5fGsb84aH1Ze5vVIhx*)
(*TC3D7BBeWBb1nQpnHkeEYdbpxt0QTaHCo8YDRv/f376bTTlpKP/iVUQxn9IK*)
(*uofPoIJ4lOwYFg/uyM7uhp8ov+7v/oJ5te+g3sMcKggot93APHDAeGw4jwoG*)
(*f7/OYPGt1vlhZlFAhSzdDz6YPw5Kv/IspUKqx9k07LwcjrjTd5Wj+B/hPotZ*)
(*16DeM7ISnZ/vi8z/n++IxsfsOio4Gm+FYfnAfSyWHxqo8Lf3MR3Ll7Un49H1*)
(*TVQwPh6sh7lm3DZ1qIUKu7f+7MTyK+1EyRnzdirgI4U0sfwLfsacv9pJBZcL*)
(*oTFYfupNf6zi6qfCy4v6fCzI67MZ/VfHqeCR+i+XEXlQgmr4e5IKXLCjGYfl*)
(*o6XMlPoMFXSNcYNYfYRQGldMF9B52LcsYPUksTLJFbFBhan9MsR1ZL6LZ8KO*)
(*baH4RVvkY/VIs3cQ/MmwCo3xgpFYvdavsYjWbVuFbhP8vRVko82j11a4VuGM*)
(*eb0bVv/hzDo2asKrIEePoY4hbxnuPU8WWQVJX0L4KPb5StLyHdFV2MuTd34E*)
(*69ducvYq4qsg6GRqiPWfOvqJ14r/rQLh5u/AXmTq+qKzjPIqJAm4XiYiP1jw*)
(*8T1pswohAklyWL/kGciNY6xahYSWm6IOyNveuKpferkG5h99JrB5k7Mws0aw*)
(*XoPxyBjGWWQjY42Y63Zr4P1k5sAMctUdial7jmvAxXBVYwLZQ6jf2cJzDdb3*)
(*Xu4bxOZX3X+p8Z/XoPaKybEWbH7yzOM4K9aANCPXmYuc9U7re0jVGtCjtx/O*)
(*QTag1Sjvr12DQH2n51nIFaNRn040rsHROmvmDGS3/Fun5TrWYKF+QTkJGafz*)
(*XcNhag2S5S+f/YLMkKyTMcy9DkytpVrY/K2Z6bSI2bMOq3zOpS7IfpJ3xR/z*)
(*r8NI/5ujzsh78+V+tB9Yhw8mV5aw+S1eeyK17sQ69NwZzLRFfjxKSUyXWQfP*)
(*pymfniEXHHkX/fr5OhCXMg7dQ377hEHnovk6qHl0N9xFlk9+dWDpxTr4b6q9*)
(*VkUmSZpFmtuuw+lvUgMqyBNwN0LPZR2OQnK5IjKfzv6w62HrUN0xuSKDbBWR*)
(*4c9dsQ5wynTPaeRftIh3FVXrYDl5hksMmabj5WpVuw7XewTYTyG/O6pt29K4*)
(*DqYSittFkSPTdmh/7FyHJLm1g8eQq0ufiHH/WwednRbpB5B3HLkrUjG3DnLr*)
(*sb8FkVXdrxy2oqzDeHP/2H7knhs8vC3UdfjyxeL4PuRZUtnGB8YNOPmtpIwP*)
(*mf/v/obd/Bsg8cBHDbsfaSuwVpfv24CgTF9/LuS475TSlwc2wCnQsZ4TWczs*)
(*dxZZaAO2nxFU2oksu2wX8eH0BuQEn9TbgfycteXZbrkN6FUvGWVGznhWYliu*)
(*sAGyaykymJcakrRfKm5A0tqTz0zIzkHOd8kqG9AdJa3OiBwicPrSB40NiE27*)
(*V0pH97eSU95suy02IOa9VtAK8oW2vU0fLTeg4oiD3zJyptP3D7zWG2Bn7ui7*)
(*hBzXVLOP32ED+GOOf6Qge7/YdvqQ2wZc1cgsn0VWy3a7ezpsA9LOKCeOIzdo*)
(*cfNlhG8Ao7cgcQxZnjmuR/LLBuilhK+PIl+8X24oFbsBpU88H40gC1LpttIp*)
(*G6DJVHdhCHn88puIm6UbcLwr2bQXWe8vh25D2QakOtqW9CD3+EYeVancALfR*)
(*zN2YG/uK0lTrNkCK/rS8Cznr7Xrpw5YNmFo2v9qBfOr4e/futg0g29+Oa0dO*)
(*IAooanVuQHaKy3bMnw5fIuv0bcChQw0DrchOFfZ/n4yh85EPTCAjLz9nTRqf*)
(*2ICHKiH7MVvwfDJ7No3iwdUUQkLWN8pbNpvfgP2ken8i8nWWZVab9Q2wZwr8*)
(*3IS8W8lKzI2TBnOtG7z1yLHlXOyXd9NgPKkv4Tey5KUfkws8NOi7z3kR8+0T*)
(*E4n6AjQQjrprUIf8nkXniKwwDeJPd7TWIAu83aCvitDAgF5kizlpOXwgU5QG*)
(*53OaBDDXjrR+OSxBg4i3BibVyMyVN/kYLtPgtnTLwSrk4MvjS7+u0ODnr0/k*)
(*SmShLI/WF0CDSy1GXpjxXwlBf+Rp8GaNaaUCmciv/eLTDfT8fU2ZmHWC1lXu*)
(*KKH94fzMMb9xPs9OuEODWxbdM+XIHNSWSTs1Gjw/qpuBOdLCsk78Pg2+0uqs*)
(*MRdqp3pGa9KAGHOREfOK9BGGt09oYMnDer4M2TO7dOC8CQ32SkuxYOY99bjk*)
(*33Ma4OeEugnI5/Z9eq1tSYOja+c8MFd8kHq0x5oGSR0qjzGrbW+52GRLAw1V*)
(*/guYLVc5lq840iDhXvpsKTL9RUrr0hsa8Hm+bMIcMH4jK80FrR/X9gNzWofb*)
(*C0EvGrQeeG6NWfr24dtt3jRolA/RwFxfXSLm50uDyEkVwDyZszq5EUSDG+a6*)
(*uzE7iIXVZQfT4NjFDFoJMlvCue+mYTSYVX87ifnTfrLn0c80YK6t68QsEmxh*)
(*1PuFBuIf39dizt3BIRccjeJFIORjlndLPnLrKw2YtI1SMbeuXWdgSqCBnZNN*)
(*DGaDlyMDhYk0+Cw1E4J5YcK1xCqZBsGxtb6YXfQORZ5MowHPIKMH5l1dxa+H*)
(*02mgvTfiDeaYO5qPIn7SQPGFjwPmUplQvh2/aOAkrW+Fue80qfVVMdrvhOnL*)
(*/6+XfE/Ev5wGPR8P2GKeFO50iK2hwenfwv9/Xtowv9DJ3zR4ulzmhtkyVrMh*)
(*u4EG3e+e+WA+pxNpc7UZva/P/g+YV/YPHKgl0aBwW9tnzIXdh2pVW2lguzMo*)
(*HvObT/qWPe002JajnI4Zfz9ewKiLBiW7WAsxM/OMVvzroUGcTHEN5lqSiJl9*)
(*P1r/8bM2zO8DTPbgBmng4Mf+F/PuHTNPeUdpUEA4yoSdV1vt6d3R4zTQ3BfC*)
(*iznc80Xh8SkanLowK4L5IG6RQ3qOBno+6rcxD5dK5VUt0EDsgpYh5m9Odrq3*)
(*l2iwGoB/hVlsde2n/hoN2jUiv2Oey5XWmt5A62vvLMOcZf2G2XYL5ct5tW7M*)
(*l+YYHnozbYJMhjInlq/Xx1hp6Ts3waSz0BXz9oSb3/7btQk/n7jEYm7U971d*)
(*wb0JjQ3MZZjv9XN+bd+7CUpTgMPqRa9tz3XakU1Qa/nPD7NjxdGgmxc2IWGW*)
(*owKrNxmXJ5da/9sE49TQacyMMt+HH0tvgt7xvD1YPXsXnDxvid+EHsFcU8xh*)
(*mWd6wpSQuU0PYv0hMxqER3Q2wVS9KhLrJ4uvF1n19TeB2FzahPnCw+/T/Yab*)
(*MHtNlI65iIsru9NkE1IUrhti/ajGZUC2wWoT/a58Lon1q16DN7pZXptg55bf*)
(*V4vMcrwowiV9E+zvLls2IGtnXNh9YGsTwv9dlcD6c57mKieJYQv6ax7YYOZk*)
(*KeRwZ9qCcPe3BZjLHl9lm2Tbgoi6Jbk2rL/tkKfn8GwBm6L6I6zfjxnd/acs*)
(*ugXE/1I8OrF+vs/st5P6FojPtLRh88bF86trX8oWpJ7uP4TNp532rEr6P7ZA*)
(*zvyCFuYIE3PusYwt0LYzDMOcfetS/FzOFmgOX+DA5tvYbnIVI2ELeN1zlzHf*)
(*imZgE23dgvzvqQRsPvL90g2w3diCaNFMsWnk+KTqB+ubW8Cwla6PWTLi1CFn*)
(*BjqsytqGYb75hprhvY0O7SwP6Zid5ALJX7jokGpo0oz9/8wwsZS34igdmJMD*)
(*9LF5/GPiwBdOZTqY6H7RXMDOM8pTb+A2HeLVTPwxG6n9O5Z+lw4WJeMEzILF*)
(*xZkqD+kQ4bp5FJvv7wO1avwM6KA+7jqB2eTC54Udr+iw71G/AXYfkJ9m+NXj*)
(*SAf74NogzIdiTZxS3tJBQNqwFHPH9v9YlTzoYNeYzo/dJxT6OwR9AunwXfTT*)
(*b8xCHryKrIl0kHyZjUY/Ebb+c9rZmUSHnbt+KWDu/ve3JTGVDla/zCwwBz3M*)
(*1r7+kw6LDrOlmOkn1aw9i+nwyOX241WsP5EDo5la6VAbGOG9hpzvRTVsbafD*)
(*ckNDGuYQaV3R+C46nGP5ScJ865t4zrUBOgztC9+7jvUbh6bfrpN0YHzf/xVz*)
(*2OnzgXdn0HrW2ZWYXw5Hqh+Zo0Nbhcwo5hPKZn/KluiQct/u+AbWf3BtCUFU*)
(*OvDThRQxD+ZKP9dbp4Nul6cJ5uLn8RKSm3Rw5YvxxvzpEPsynU4HgzmLJMz/*)
(*A7lgmeE=*)
(*"]]}, "Charting`Private`Tag$392830#3"]}, {}}, AspectRatio -> GoldenRatio^(-1), Axes -> {True, True}, AxesLabel -> {"Tt", None}, AxesOrigin -> {Automatic, 0}, DisplayFunction -> Identity, Frame -> {{False, False}, {False, False}}, FrameLabel -> {{None, None}, {None, None}}, FrameTicks -> {{Automatic, Charting`ScaledFrameTicks[{Identity, Identity}]}, {Automatic, Charting`ScaledFrameTicks[{Identity, Identity}]}}, GridLines -> {None, None}, GridLinesStyle -> Directive[GrayLevel[0.5, 0.4]], ImagePadding -> All, ImageSize -> 600, LabelStyle -> {GrayLevel[0], 20}, Method -> {"DefaultBoundaryStyle" -> Automatic, "DefaultGraphicsInteraction" -> {"Version" -> 1.2, "TrackMousePosition" -> {True, False}, "Effects" -> {"Highlight" -> {"ratio" -> 2}, "HighlightPoint" -> {"ratio" -> 2}, "Droplines" -> {"freeformCursorMode" -> True, "placement" -> {"x" -> "All", "y" -> "None"}}}}, "DefaultMeshStyle" -> AbsolutePointSize[6], "ScalingFunctions" -> None, "CoordinatesToolOptions" -> {"DisplayFunction" -> ({(Identity[#]& )[Part[#, 1]], (Identity[#]& )[Part[#, 2]]}& ), "CopiedValueFunction" -> ({(Identity[#]& )[Part[#, 1]], (Identity[#]& )[Part[#, 2]]}& )}}, PlotRange -> {All, All}, PlotRangeClipping -> True, PlotRangePadding -> {{Scaled[0.02], Scaled[0.02]}, {Scaled[0.05], Scaled[0.05]}}, Ticks -> {Automatic, Automatic}]]*)


(* ::Input:: *)
(*SetDirectory[NotebookDirectory[]];*)
(*Export["EMT_homogeous.pdf",plotHolEMT];*)


(* ::Input:: *)
(*(* QNM *)*)
(*plotDeltaP=LogPlot[{Abs[pl[Pi t]-pt[Pi t]]},{t,0,Pi},LabelStyle->{Black,20},AxesLabel->{"Tt","Abs[\!\(\*FormBox[\(\(\\\ \)\(\[Delta]\[ScriptCapitalP](t)\)\(]\)\),*)
(*TraditionalForm]\)"},PlotRange->All,PlotStyle->{{Thick,Black}},ImageSize->500]*)


(* ::Input:: *)
(*SetDirectory[NotebookDirectory[]];*)
(*Export["deltaP.pdf",plotDeltaP];*)
